/**
 *
 */
package abc.parser;

import java.util.LinkedList;

import abc.score.EDecoration;
import abc.score.GraceNote;
import abc.score.IInformationFieldClef;
import abc.score.InformationFieldClef;
import abc.score.Note;

/**
 * 解析中のTrackデータ
 *
 * @author CrapeMyrtle
 *
 */
public class ParseDataTrack implements IInformationFieldClef {
	public static final int			GROUP_NOTE_TOTAL_COUNT	= 0;
	public static final int			GROUP_NOTE_REMAIN		= 1;

	/** 次回追加予定のNote */
	private Note					nextNote				= new Note();

	/** 次回追加予定のNote(Grace) */
	private Note					nextGraceNote			= new Note();

	/** 連音符の情報を保持 */
	private LinkedList<int[]>		groupNoteStartPos		= new LinkedList<int[]>();

	/** トラックの一時的な重複の小節数 */
	private int						voiceOverlay			= 0;

	/** Grace Notes */
	private GraceNote				grace;

	/** 装飾 */
	private LinkedList<EDecoration>	chordDecorations		= new LinkedList<EDecoration>();

	/** clefオプション */
	private InformationFieldClef	clef					= new InformationFieldClef();

	/**
	 * コンストラクタ
	 */
	public ParseDataTrack() {
		this.clef.setClefOctave(0);
		this.clef.setOctave(0);
		this.clef.setTransposition(0);
	}

	/**
	 * 解析中のNoteを返す。
	 *
	 * @return note 解析中のNote
	 */
	public Note getNextNote() {
		return nextNote;
	}

	/**
	 * 解析中のNoteをセットする。
	 *
	 * @param nextNote
	 *            セットする解析中Note
	 */
	public void setNextNote(Note nextNote) {
		this.nextNote = nextNote;
	}

	/**
	 * 解析中のGrace Noteを返す。
	 *
	 * @return note 解析中のGrace Note
	 */
	public Note getNextGraceNote() {
		return nextGraceNote;
	}

	/**
	 * 解析中のGrace Noteをセットする。
	 *
	 * @param nextGraceNote
	 *            セットする解析中Grace Note
	 */
	public void setNextGraceNote(Note nextGraceNote) {
		this.nextGraceNote = nextGraceNote;
	}

	/**
	 * 連音符の開始位置情報を取得する。連音符の開始位置情報({連音符の続く全音符数, 連音符の終わりまでの音符数})のリストを返す。
	 *
	 * @return 連音符の開始位置情報のリスト
	 */
	public LinkedList<int[]> getGroupNoteStartPos() {
		return groupNoteStartPos;
	}

	/**
	 * Voice Overlayの残り小節数を取得する
	 *
	 * @return Voice Overlayの残り小節数
	 */
	public int getVoiceOverlay() {
		return voiceOverlay;
	}

	/**
	 * Voice Overlayの残り小節数をセットする
	 *
	 * @param voiceOverlay
	 *            セットするVoice Overlayの残り小節数
	 */
	public void setVoiceOverlay(int voiceOverlay) {
		this.voiceOverlay = voiceOverlay;
	}

	/**
	 * Grace Notesを取得する
	 *
	 * @return Grace Notes
	 */
	public GraceNote getGraceNote() {
		return grace;
	}

	/**
	 * Grace Notesをセットする
	 *
	 * @param grace
	 *            セットするGrace Notes
	 */
	public void setGraceNote(GraceNote grace) {
		this.grace = grace;
	}

	/**
	 * Grace NotesにNoteを追加する
	 *
	 * @param note
	 *            追加するNote
	 */
	public void addNoteToGrace(Note note) {
		this.grace.getNotes().add(note);
	}

	/**
	 * Grace Notesを初期化する
	 *
	 * @param isShortGrace
	 *            短前打音かどうか。true: 短前打音。false: 長前打音。
	 */
	public void initGrace(boolean isShortGrace) {
		this.grace = new GraceNote();
		this.grace.setShortGrace(isShortGrace);
	}

	/**
	 * 和音に付与する装飾リストを取得する
	 *
	 * @return 装飾リスト
	 */
	public LinkedList<EDecoration> getChordDecorations() {
		return chordDecorations;
	}

	/**
	 * 和音に付与する装飾リストをセットする
	 *
	 * @param chordDecorations
	 *            セットする装飾リスト
	 */
	public void setChordDecorations(LinkedList<EDecoration> chordDecorations) {
		this.chordDecorations = chordDecorations;
	}

	/**
	 * 和音に付与する装飾リストをクリアする
	 */
	public void clearChordDecorations() {
		this.chordDecorations.clear();
	}

	/**
	 * clefオプションを取得する
	 * @return clefオプション
	 */
	public InformationFieldClef getClef() {
		return clef;
	}

	/**
	 * clefオプションをセットする
	 * @param clef セットするclefオプション
	 */
	public void setClef(InformationFieldClef clef) {
		this.clef = clef;
	}

	/* (非 Javadoc)
	 * @see abc.score.IInformationFieldClef#getClefOctave()
	 */
	@Override
	public Integer getClefOctave() {
		return this.clef.getClefOctave();
	}

	/* (非 Javadoc)
	 * @see abc.score.IInformationFieldClef#setClefOctave(java.lang.Integer)
	 */
	@Override
	public void setClefOctave(Integer clefOctave) {
		this.clef.setClefOctave(clefOctave);
	}

	/* (非 Javadoc)
	 * @see abc.score.IInformationFieldClef#getOctave()
	 */
	@Override
	public Integer getOctave() {
		return this.clef.getOctave();
	}

	/* (非 Javadoc)
	 * @see abc.score.IInformationFieldClef#setOctave(java.lang.Integer)
	 */
	@Override
	public void setOctave(Integer octave) {
		this.clef.setOctave(octave);
	}

	/* (非 Javadoc)
	 * @see abc.score.IInformationFieldClef#getTransposition()
	 */
	@Override
	public Integer getTransposition() {
		return this.clef.getTransposition();
	}

	/* (非 Javadoc)
	 * @see abc.score.IInformationFieldClef#setTransposition(java.lang.Integer)
	 */
	@Override
	public void setTransposition(Integer transposition) {
		this.clef.setTransposition(transposition);
	}

	/* (非 Javadoc)
	 * @see abc.score.IInformationFieldClef#getOctaveTotal()
	 */
	@Override
	public Integer getOctaveTotal() {
		return this.clef.getOctaveTotal();
	}
}
