package abc.parser;

import java.io.IOException;
import java.io.LineNumberReader;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Map.Entry;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import util.Debug;
import util.Option;

import abc.score.AInformationField;
import abc.score.InformationFieldKey;
import abc.score.InformationFieldReferenceNumber;
import abc.score.Key;
import abc.score.MapInformationField;
import abc.score.Score;
import abc.score.Track;
import abc.score.Tune;

/**
 * ABC譜のパーサ
 *
 * @author CrapeMyrtle
 *
 */
public class Parser {
	private static final Pattern	PTN_COMMENT		= Pattern.compile("%.*$");
	private static final Pattern	PTN_HEADER_X	= Pattern.compile("^X:");
	private static final Pattern	PTN_HEADER_K	= Pattern.compile("^K:");

	private static final String		STR_ARGS		= "x-args";
	private static final Pattern	PTN_NOSKIP		= Pattern.compile("[^\\s]+");

	private final LineNumberReader	in;
	private String					line			= "";
	private String					comment			= "";
	private EState					state;

	private Score					score			= new Score();
	private ParserInformationField	pInField		= null;
	private ParserTuneBody			pTuneBody		= null;
	private Option					opt				= null;

	/**
	 * コンストラクタ
	 *
	 * @param in
	 *            パースするReader
	 */
	public Parser(LineNumberReader in) {
		this.in = in;
	}

	/**
	 * パースする
	 *
	 * @return パース結果の{@link Score}。パースに失敗した場合はnullを返す。
	 * @throws IOException
	 *             ReaderのReadに失敗した際例外を発行する
	 */
	public Score parse(Option opt) throws IOException {
		this.opt = opt;
		this.pTuneBody = new ParserTuneBody(opt);
		this.pInField = new ParserInformationField(opt);

		this.state = EState.FILE_HEADER;

		while ((this.line = this.in.readLine()) != null) {
			Debug.out("Read line " + this.in.getLineNumber() + ": " + this.line);
			this.parseComment();

			if (this.line.equals("") && this.comment.equals("")) {
				Debug.out("Empty Line");
				// TODO 本来はABC譜の区切り(直後にX:フィールド)だが、取り敢えず無視して
				// ABC譜の区切りはX:フィールドのみで判定
				continue;
			}

			this.nextState();
			Debug.out("Next State: " + this.state);

			MapInformationField mapField;
			String text;
			switch (state) {
			case FILE_HEADER:
				if (this.line.equals("") && this.comment.startsWith("%%")) {
					if (this.comment.startsWith("%%" + STR_ARGS)) {
						// コマンドライン引数を設定できる！
						String line = this.comment.substring(2 + STR_ARGS.length());
						ArrayList<String> args = new ArrayList<String>();

						while (true) {
							Matcher m = PTN_NOSKIP.matcher(line);
							if (!m.find()) {
								break;
							}

							args.add(m.group(0));
							line = line.substring(m.end());
						}

						this.opt.parseOpt(args.toArray(new String[] {}));
					}
				} else {
					mapField = this.pInField.parse(this.line, this.comment);
					if (mapField != null) {
						for (String name : mapField.keySet()) {
							for (AInformationField field : mapField.get(name)) {
								score.addFileHeader(name, field);
							}
						}
					}
				}
				break;
			case TUNE_HEADER:
			case TUNE_HEADER_LAST:
				mapField = this.pInField.parse(this.line, this.comment);
				if (mapField != null) {
					for (String name : mapField.keySet()) {
						if (name.equals("X")) {
							// X: があった場合は、新しいTuneにする
							text = String.valueOf(((InformationFieldReferenceNumber) mapField.getLast(name)).getRefNum());
							this.newTune(text);
						}
						for (AInformationField field : mapField.get(name)) {
							if (field instanceof InformationFieldKey) {
								Key key = ((InformationFieldKey) field).getKey();
								if (key.isNone()) {
									// 何もないKeyは追加しない
									// TODO 追加しておいて、後から探すときにisNone()のをスキップする？
									continue;
								}
							}
							score.getActiveTune().addInformation(name, field);
						}
					}
				}
				break;
			case TUNE_BODY:
				this.pTuneBody.parse(this.line, this.comment);
				break;
			}
		}

		this.nextState();

		// Voice Overlay の合成 → Graceを適用してから → Grace適用はRepeat展開後
		// → Voice Overlayにもリピート情報付与しておかないと…
		for (Tune tune : this.score.getTunes().values()) {
			for (Iterator<Entry<String, Track>> iter = tune.getTracks().entrySet().iterator(); iter.hasNext();) {
				Entry<String, Track> entry = iter.next();
				String trackId = entry.getKey();
				String[] idOverlay = ParserTuneBody.getTrackIdAndOverlay(trackId);
				if (!idOverlay[1].equals("0")) {
					Track track = entry.getValue();
					// iter.remove();
					track.copyRepeatFrom(tune.getTracks().get(idOverlay[0]));
					// tune.getTracks().get(idOverlay[0]).copyInformationFrom(track);
				}

				// 解析に使った一時領域をクリア
				entry.getValue().clearTemporary();
			}
		}

		return this.score;
	}

	/**
	 * 行のうちコメント部分を分離する
	 */
	private void parseComment() {
		this.comment = "";
		Matcher m = PTN_COMMENT.matcher(this.line);
		if (m.find()) {
			this.comment = this.line.substring(m.start());
			this.line = this.line.substring(0, m.start());
		}
	}

	/**
	 * 次状態に移行する
	 */
	private void nextState() {
		switch (this.state) {
		case FILE_HEADER:
			if (this.line == null) {
				this.state = EState.EOF;
			} else if (PTN_HEADER_X.matcher(this.line).find()) {
				this.state = EState.TUNE_HEADER;
			}
			break;
		case TUNE_HEADER:
			if (this.line == null) {
				this.state = EState.EOF;
			} else if (PTN_HEADER_K.matcher(this.line).find()) {
				this.state = EState.TUNE_HEADER_LAST;
			}
			break;
		case TUNE_HEADER_LAST:
			if (this.line == null) {
				this.state = EState.EOF;
			} else if (PTN_HEADER_X.matcher(this.line).find()) {
				this.state = EState.TUNE_HEADER;
			} else {
				this.state = EState.TUNE_BODY;
				this.pTuneBody.parseStart();
			}
			break;
		case TUNE_BODY:
			if (this.line == null) {
				this.state = EState.EOF;
				this.pTuneBody.parseEnd();
			} else if (PTN_HEADER_X.matcher(this.line).find()) {
				this.state = EState.TUNE_HEADER;
				this.pTuneBody.parseEnd();
			}
			break;
		}
	}

	/**
	 * 新しいTuneを追加し、追加したTuneに解析対象を移す。
	 *
	 * @param id
	 *            新しく追加するTuneのid
	 */
	private void newTune(String id) {
		this.score.addTune(id, new Tune());
		this.score.setActiveTuneId(id);
		this.pTuneBody.clear();
		this.pTuneBody.setTune(this.score.getActiveTune());
	}
}
