/**
 *
 */
package abc.parser;

import java.util.regex.Matcher;

import util.Debug;
import util.Option;

import abc.score.*;
import abc.util.Fraction;

/**
 * information fieldのパーサ
 *
 * @author CrapeMyrtle
 *
 */
public class ParserInformationField {

	private Option	opt	= null;

	/**
	 * コンストラクタ
	 * @param opt オプション
	 */
	public ParserInformationField(Option opt) {
		this.opt = opt;
	}

	/**
	 * information field を解析する
	 *
	 * @param text
	 *            解析する文字列
	 * @param comment
	 *            コメント部分
	 * @return 解析結果。解析結果有効なinformation fieldでなかった場合、nullを返す。
	 */
	public MapInformationField parse(String text, String comment) {
		Matcher m = Constant.PTN_INFO_FIELD.matcher(text);

		MapInformationField field = null;

		if (m.find()) {
			field = parseInformationField(m.group(1), m.group(2));
			if (field == null) {
				Debug.out("Parse Failed -> " + text);
			}
		} else {
			if (!text.equals("")) {
				// information field以外はないはず
				Debug.out("Ignored -> " + text);
			}
		}

		// 本来はコメント部分のABCファイルヘッダを解析するが、面倒なので全部無視
		if (!comment.equals("")) {
			Debug.out("Comment -> " + comment);
		}
		return field;
	}

	/**************************************************************************
	 * XXX: ここから実際の解析を行う
	 **************************************************************************/

	/**
	 * information fieldを解析する
	 *
	 * @param name
	 *            フィールド名
	 * @param text
	 *            テキスト
	 */
	private MapInformationField parseInformationField(String name, String text) {
		switch (name) {
		case "K":
			return this.parseInformationFieldKey(name, text);
		case "L":
			return this.parseInformationFieldLength(name, text);
		case "M":
			return this.parseInformationFieldMeter(name, text);
		case "Q":
			return this.parseInformationFieldTempo(name, text);
		case "T":
			return this.parseInformationFieldTitle(name, text);
		case "V":
			return this.parseInformationFieldVoice(name, text);
		case "X":
			return this.parseInformationFieldReferenceNumber(name, text);
		case "U":
			return this.parseInformationFieldDefine(name, text);
		default:
			Debug.out("NoParse -> " + name + ":" + text);
			return null;
		}
	}

	/**
	 * information field (K:)を解析する
	 *
	 * @param text
	 *            テキスト
	 * @return 解析結果
	 */
	private MapInformationField parseInformationFieldKey(String name, String text) {
		InformationFieldKey field = new ParserInformationFieldKey().parse(text);
		if (field == null) {
			return null;
		} else {
			return new MapInformationField(name, field);
			// Key key = field.getKey();
			// if (key.isNone()) {
			// // Key情報は持っていない
			// return new MapInformationField("clef", field.getClef());
			// } else {
			// MapInformationField map = new MapInformationField();
			// LinkedList<AInformationField> list = new
			// LinkedList<AInformationField>();
			// list.add(field);
			// map.put(name, list);
			//
			// list = new LinkedList<AInformationField>();
			// list.add(field.getClef());
			// map.put("clef", list);
			// return map;
			// }
		}
	}

	/**
	 * information field (L:)を解析する
	 *
	 * @param text
	 *            テキスト
	 * @return 解析結果
	 */
	private MapInformationField parseInformationFieldLength(String name, String text) {
		InformationFieldLength field = new InformationFieldLength(Fraction.parseFraction(text));
		field.setLength(field.getLength().mul(this.opt.getLengthScale()));
		return new MapInformationField(name, field);
	}

	/**
	 * information field (M:)を解析する
	 *
	 * @param text
	 *            テキスト
	 * @return 解析結果
	 */
	private MapInformationField parseInformationFieldMeter(String name, String text) {
		Meter meter = new Meter(text);
		meter.setLength(meter.getLength().mul(this.opt.getLengthScale()));
		return new MapInformationField(name, new InformationFieldMeter(meter));
	}

	/**
	 * information field (Q:)を解析する
	 *
	 * @param text
	 *            テキスト
	 * @return 解析結果
	 */
	private MapInformationField parseInformationFieldTempo(String name, String text) {
		InformationFieldTempo field = new ParserInformationFieldTempo().parse(text);
		if (field == null) {
			return null;
		} else {
			Tempo tempo = field.getTempo();
			tempo.setBase(tempo.getBase().mul(this.opt.getLengthScale()));
			return new MapInformationField(name, field);
		}
	}

	/**
	 * information field (T:)を解析する
	 *
	 * @param text
	 *            テキスト
	 * @return 解析結果
	 */
	private MapInformationField parseInformationFieldTitle(String name, String text) {
		return new MapInformationField(name, new InformationFieldTitle(text.trim()));
	}

	/**
	 * information field (V:)を解析する
	 *
	 * @param text
	 *            テキスト
	 * @return 解析結果
	 */
	private MapInformationField parseInformationFieldVoice(String name, String text) {
		InformationFieldVoice field = new ParserInformationFieldVoice().parse(text);
		if (field == null) {
			return null;
		} else {
			return new MapInformationField(name, field);
			// MapInformationField map = new MapInformationField();
			// LinkedList<AInformationField> list = new
			// LinkedList<AInformationField>();
			// list.add(field);
			// map.put(name, list);
			//
			// list = new LinkedList<AInformationField>();
			// list.add(field.getClef());
			// map.put("clef", list);
			// return map;
		}
	}

	/**
	 * information field (X:)を解析する
	 *
	 * @param text
	 *            テキスト
	 * @return 解析結果
	 */
	private MapInformationField parseInformationFieldReferenceNumber(String name, String text) {
		return new MapInformationField(name, new InformationFieldReferenceNumber(
				Integer.parseInt(text.trim())));
	}

	/**
	 * information field (U:)を解析する
	 *
	 * @param text
	 *            テキスト
	 * @return 解析結果
	 */
	private MapInformationField parseInformationFieldDefine(String name, String text) {
		InformationFieldDefine field = new ParserInformationFieldDefine().parse(text);
		if (field == null) {
			return null;
		} else {
			return new MapInformationField(name, field);
		}
	}

}
