/**
 *
 */
package abc.parser;

import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map.Entry;
import java.util.NavigableMap;
import java.util.regex.Matcher;

import util.Debug;
import util.Option;

import abc.score.*;
import abc.util.Fraction;

/**
 * tune bodyのパーサ。おそらく一番のキモであり、大きく複雑になりすぎたクラス。
 *
 * @author CrapeMyrtle
 *
 */
public class ParserTuneBody {

	private ParserInformationField			pInField	= null;
	private Tune							tune;
	private String							activeTrackId;
	private HashMap<String, ParseDataTrack>	trackData;
	private boolean							isParsingChord;
	private boolean							isParsingChordExist;
	private boolean							isParsingGrace;

	private Option							opt;

	public ParserTuneBody(Option opt) {
		this.clear();
		this.opt = opt;
		this.pInField = new ParserInformationField(opt);
	}

	/**
	 * 解析中の情報をクリアする。Tuneが移った時に実施する。
	 */
	public void clear() {
		this.tune = null;
		this.activeTrackId = InformationFieldVoice.DEFAULT_ID;
		this.trackData = new HashMap<String, ParseDataTrack>();
		this.trackData.put(this.activeTrackId, new ParseDataTrack());
	}

	/**
	 * 解析中のTuneをセットする
	 *
	 * @param tune
	 *            解析中のTune
	 */
	public void setTune(Tune tune) {
		this.tune = tune;
	}

	/**
	 * 解析中のTuneを取得する
	 *
	 * @return 解析中のTune
	 */
	public Tune getTune() {
		return this.tune;
	}

	/**
	 * tune body部分の解析を行う
	 *
	 * @param text
	 *            解析を行う文字列
	 * @param comment
	 *            コメント
	 */
	public void parse(String text, String comment) {
		Matcher m = Constant.PTN_INLINE_INFO_FIELD.matcher(text);

		if (m.find()) {
			Debug.out("Information Field -> " + m.group(0));
			this.parseInformationField(text, comment);
		} else {
			boolean isParsingBar = false;
			boolean isParsingInline = false;
			this.isParsingChord = false;
			this.isParsingChordExist = false;
			this.isParsingGrace = false;
			String substr = null;
			for (int idx = 0; idx < text.length();) {
				boolean isPreBar = isParsingBar;
				isParsingBar = false;
				substr = text.substring(idx);

				// Debug.out("SubStr -> " + substr);

				m = Constant.PTN_SKIP.matcher(substr);
				if (m.find()) {
					idx += m.end();
					continue;
				}

				// 縦線or和音orインラインフィールドの開始
				m = Constant.PTN_BAR.matcher(substr);
				if (m.find()) {
					if (!this.isParsingChord) {
						if (m.group(2).equals("[")) {
							// インラインフィールド、和音、繰り返し、縦線のいずれか
							String next = substr.substring(m.end());

							Matcher m2;
							if (!this.isParsingGrace) {
								m2 = Constant.PTN_INLINE_INFO_FIELD.matcher(next);
								if (m2.find()) {
									// インラインフィールド
									this.parseInformationField(m2.group(0), "");
									idx += m.end() + m2.end();
									isParsingInline = true;
									continue;
								}
							}

							if (!this.isParsingGrace) {
								m2 = Constant.PTN_REPEAT.matcher(next);
								if (m2.find()) {
									// 繰り返し
									// "["も縦線として追加しておく
									this.parseBar(m.group(2), isPreBar);
									this.parseRepeatOrder(m2.group(0));
									idx += m.end() + m2.end();
									isParsingBar = true;
									continue;
								}
							}

							m2 = Constant.PTN_SKIP.matcher(next);
							if (m2.find()) {
								// 空白は飛ばす (次回ループの最初で処理)
								idx += m.end();
								continue;
							}

							if (!this.isParsingGrace) {
								m2 = Constant.PTN_BAR.matcher(next);
								if (m2.find()) {
									// 縦線
									// (mの部分だけ。m2の部分は、縦線以外の可能性があるため次回ループ時に解析)
									this.parseBar(m.group(2), isPreBar);
									idx += m.end();
									isParsingBar = true;
									continue;
								}
							}

							// 和音部分だと判断
							if (m.group(1) != null) {
								// ".["の"."はスタッカート
								this.parseDecoration(m.group(1));
							}
							this.parseChordStart();
							idx += m.end();
							continue;
						}
					}

					if (!this.isParsingChord && !this.isParsingGrace) {
						if (m.group(2).equals("|")) {
							// 繰り返し、縦線のいずれか
							isParsingBar = true;
							String next = substr.substring(m.end());

							Matcher m2 = Constant.PTN_REPEAT.matcher(next);
							if (m2.find()) {
								// 繰り返し |[(数字) の省略形なので"|"を追加しておく
								this.parseBar(m.group(2), isPreBar);
								this.parseRepeatOrder(m2.group(0));
								idx += m.end() + m2.end();
								continue;
							}

							// 縦線
							this.parseBar(m.group(2), isPreBar);
							idx += m.end();
							continue;
						}
					}

					if (m.group(2).equals("]")) {
						// インラインフィールド、和音、縦線のいずれか
						if (isParsingInline) {
							// インラインフィールド
							idx += m.end();
							isParsingInline = false;
							continue;
						}

						if (this.isParsingChord) {
							// 和音
							String next = substr.substring(m.end());

							Matcher m2 = Constant.PTN_LENGTH.matcher(next);
							if (m2.find()) {
								this.parseChordEnd(m2.group(1), m2.group(2));
								idx += m.end() + m2.end();
							} else {
								this.parseChordEnd(null, null);
								idx += m.end();
							}
							continue;
						}

						if (!this.isParsingChord && !this.isParsingGrace) {
							// 縦線
							this.parseBar(m.group(2), isPreBar);
							idx += m.end();
							isParsingBar = true;
							continue;
						}
					}

					if (!this.isParsingChord && !this.isParsingGrace) {
						// 縦線
						this.parseBar(m.group(2), isPreBar);
						idx += m.end();
						isParsingBar = true;
						continue;
					}
				}

				m = Constant.PTN_NOTE.matcher(substr);
				if (m.find()) {
					this.parseNote(m.group(1), m.group(2), m.group(3), m.group(4), m.group(5));
					idx += m.end();
					continue;
				}

				if (!this.isParsingChord && !this.isParsingGrace) {
					m = Constant.PTN_REST.matcher(substr);
					if (m.find()) {
						this.parseRest(m.group(1), m.group(2), m.group(3));
						idx += m.end();
						continue;
					}
				}

				if (!this.isParsingChord && !this.isParsingGrace) {
					m = Constant.PTN_MEASURE_REST.matcher(substr);
					if (m.find()) {
						this.parseMeasureRest(m.group(1), m.group(2));
						idx += m.end();
						continue;
					}
				}

				m = Constant.PTN_SYMBOL.matcher(substr);
				if (m.find()) {
					this.parseDecoration(m.group(1));
					idx += m.end();
					continue;
				}

				if (!this.isParsingChord && !this.isParsingGrace) {
					m = Constant.PTN_STRING.matcher(substr);
					if (m.find()) {
						int end = m.end();
						String inside = m.group(1);

						m = Constant.PTN_ANNOTATION.matcher(substr);
						if (m.find()) {
							this.parseAnnotation(m.group(1), m.group(2));
							idx += m.end();
							continue;
						}

						m = Constant.PTN_CHORD_STRING.matcher(substr);
						if (m.find()) {
							this.parseChordString(m.group(1), m.group(2), m.group(3));
							idx += m.end();
							continue;
						}

						// ABCフォーマットから外れている文字列要素→annotaionとしておいておこうか。。。
						this.parseAnnotation(null, inside);
						idx += end;
						continue;
					}
				}

				if (!this.isParsingChord) {
					m = Constant.PTN_BROKEN.matcher(substr);
					if (m.find()) {
						this.parseBroken(m.group(1));
						idx += m.end();
						continue;
					}
				}

				if (!this.isParsingGrace) {
					m = Constant.PTN_TIE.matcher(substr);
					if (m.find()) {
						this.parseTie(m.group(0));
						idx += m.end();
						continue;
					}
				}

				if (!this.isParsingChord && !this.isParsingGrace) {
					m = Constant.PTN_GROUP_NOTES.matcher(substr);
					if (m.find()) {
						this.parseGroupNote(m.group(1), m.group(2), m.group(3));
						idx += m.end();
						continue;
					}
				}

				if (!this.isParsingChord && !this.isParsingGrace) {
					m = Constant.PTN_VOICE_OVERLAY.matcher(substr);
					if (m.find()) {
						this.parseVoiceOverlay(m.group(1));
						idx += m.end();
						continue;
					}
				}

				if (!this.isParsingChord && !this.isParsingGrace) {
					m = Constant.PTN_GRACE_START.matcher(substr);
					if (m.find()) {
						this.parseGraceStart(m.group(1));
						idx += m.end();
						continue;
					}
				}

				if (this.isParsingGrace) {
					m = Constant.PTN_GRACE_END.matcher(substr);
					if (m.find()) {
						this.parseGraceEnd();
						idx += m.end();
						continue;
					}
				}

				if (!this.isParsingChord && !this.isParsingGrace) {
					m = Constant.PTN_SLUR_START.matcher(substr);
					if (m.find()) {
						this.parseDecoration(m.group(1));
						idx += m.end();
						continue;
					}
				}

				if (!this.isParsingChord && !this.isParsingGrace) {
					m = Constant.PTN_SLUR_END.matcher(substr);
					if (m.find()) {
						this.parseDecoration(m.group(1));
						idx += m.end();
						continue;
					}
				}
				m = Constant.PTN_DECORATION.matcher(substr);
				if (m.find()) {
					this.parseDecoration(m.group(0));
					idx += m.end();
					continue;
				}

				if (!this.isParsingChord && !this.isParsingGrace) {
					m = Constant.PTN_SPACE.matcher(substr);
					if (m.find()) {
						this.parseSpace();
						idx += m.end();
						continue;
					}
				}

				// 何も当てはまらなかったら、U:フィールドに定義があるか探す。なければ無視！
				String chr = substr.substring(0, 1);
				LinkedList<AInformationField> field = this.getActiveTrack().getInformation(
						"U", true);
				boolean isFound = false;
				for (Iterator<AInformationField> iter = field.descendingIterator(); iter.hasNext();) {
					InformationFieldDefine def = (InformationFieldDefine) iter.next();
					if (def.getKey().equals(chr)) {
						Matcher m2;
						m2 = Constant.PTN_SYMBOL.matcher(def.getValue());
						if (m2.find()) {
							this.parseDecoration(m2.group(1));
							isFound = true;
							break;
						}

						m2 = Constant.PTN_ANNOTATION.matcher(def.getValue());
						if (m2.find()) {
							this.parseAnnotation(m2.group(1), m2.group(2));
							isFound = true;
							break;
						}
						break;
					}
				}

				if (!isFound) {
					Debug.out("Ignored -> " + chr);
				}

				++idx;
			}

			if (this.isParsingChord) {
				// 和音が閉じてなかったら、強引に閉じておく
				this.parseChordEnd(null, null);
			}

			if (this.isParsingGrace) {
				// Graceが閉じてなかったら、強引に閉じておく
				this.parseGraceEnd();
			}
		}

		this.isParsingChord = false;
		this.isParsingChordExist = false;
		this.isParsingGrace = false;

		// コメント
		if (!comment.equals("")) {
			boolean isSpecial = false;
			if (text.equals("")) {
				// コメントのみ
				m = Constant.PTN_PROGRAM_ID.matcher(comment);
				if (m.find()) {
					this.parseProgramId(m.group(1));
				}
			}
			if (!isSpecial) {
				Debug.out("Comment -> " + comment);
			}
		}
	}

	/**
	 * TuneBodyの解析開始
	 */
	public void parseStart() {
		this.updateTrackDataClefFromTrack(this.activeTrackId);
	}

	/**
	 * TuneBodyの解析終了
	 */
	public void parseEnd() {
		for (Entry<String, Track> entry : this.tune.getTracks().entrySet()) {
			String trackId = entry.getKey();

			ParseDataTrack trackData = this.trackData.get(trackId);
			Note note = trackData.getNextNote();
			if (note != null) {
				if (note.getRealLength() == null) {
					note.setRealLength(Fraction.ZERO.clone());
				}
				this.updateRepeatBar(note);
				this.addNote(entry.getValue(), note);
			}
		}
	}

	/**************************************************************************
	 * XXX ここから実際の解析を行う
	 **************************************************************************/

	/**
	 * information fieldを解析する
	 *
	 * @param text
	 *            解析する文字列
	 * @param comment
	 *            コメント
	 */
	private void parseInformationField(String text, String comment) {
		MapInformationField mapField = this.pInField.parse(text, comment);
		if (mapField != null) {
			for (String name : mapField.keySet()) {
				for (AInformationField field : mapField.get(name)) {
					if (name.equals("V")) {
						// V: があったら、新しいTrackにする
						this.newTrack(((InformationFieldVoice) field).getId());
					}
					if (field instanceof InformationFieldKey
							&& ((InformationFieldKey) field).getKey().isNone()) {
						// 何もないKeyは追加しない
						// TODO 追加しておいて、後から探すときにisNone()をスキップする？
					} else {
						this.getActiveNextNote().addInformation(name, field);
					}

					if (field instanceof IInformationFieldClef) {
						this.getActiveNextNote().addInformation(Constant.INFO_FIELD_KEY_CLEF, field);

						IInformationFieldClef clef = (IInformationFieldClef) field;
						if (clef.getClefOctave() != null) {
							this.getActiveTrackData().setClefOctave(clef.getClefOctave());
						}
						if (clef.getOctave() != null) {
							this.getActiveTrackData().setOctave(clef.getOctave());
						}
						if (clef.getTransposition() != null) {
							this.getActiveTrackData().setTransposition(clef.getTransposition());
						}
					}
				}
			}
		}
	}

	/**
	 * 長さの部分の解析を行う
	 *
	 * @param length1
	 *            長さ ('/'より前の部分)
	 * @param length2
	 *            長さ ('/'以降)
	 * @return
	 */
	private Fraction parseLength(String length1, String length2) {
		int num = 1;
		int den = 1;

		if (length1 != null) {
			num = Integer.parseInt(length1);
		}

		if (length2 != null) {
			Matcher m = Constant.PTN_LENGTH_DEN.matcher(length2);
			if (m.find()) {
				den = Integer.parseInt(m.group(1));
			} else {
				den = (int) Math.pow(2, length2.length());
			}
		}

		return new Fraction(num, den);
	}

	/**
	 * 音符部分の解析を行う
	 *
	 * @param accidental
	 *            臨時記号
	 * @param note
	 *            音階
	 * @param octave
	 *            オクターブ
	 * @param length1
	 *            長さ ('/'より前の部分)
	 * @param length2
	 *            長さ ('/'以降)
	 */
	private void parseNote(String accidental, String note, String octave, String length1,
			String length2) {
		NoteUnit oNoteUnit = new NoteUnit();

		EAccidental oAccidental = null;
		int oOctave = 0;

		if (accidental != null) {
			switch (accidental) {
			case "^^":
				oAccidental = EAccidental.DOUBLE_SHARP;
				break;
			case "^":
				oAccidental = EAccidental.SHARP;
				break;
			case "=":
				oAccidental = EAccidental.NATURAL;
				break;
			case "_":
				oAccidental = EAccidental.FLAT;
				break;
			case "__":
				oAccidental = EAccidental.DOUBLE_FLAT;
				break;
			}
		}

		if (octave != null) {
			oOctave = octave.length() * (octave.startsWith("'") ? 1 : -1);
		}

		// clefオプションのoctaveにより変化するので、その時にまとめて調整
		// if (oOctave > 0 && note.equals(note.toUpperCase())) {
		// --oOctave;
		// note = note.toLowerCase();
		// } else if (oOctave < 0 && note.equals(note.toLowerCase())) {
		// ++oOctave;
		// note = note.toUpperCase();
		// }

		oNoteUnit.setAcci(oAccidental);
		oNoteUnit.setLength(this.parseLength(length1, length2));
		oNoteUnit.setNote(note);
		oNoteUnit.setOctave(oOctave);

		this.getActiveNextNote().addNote(oNoteUnit);

		this.addActiveNextNote();
	}

	/**
	 * 和音開始時の解析を行う
	 */
	private void parseChordStart() {
		this.isParsingChord = true;
		this.isParsingChordExist = false;

		Note nextNote = this.getActiveNextNote();
		ParseDataTrack trackData = this.getActiveTrackData();
		trackData.getChordDecorations().addAll(nextNote.getUnitDecorations());
		nextNote.getUnitDecorations().clear();
	}

	/**
	 * 和音終了時の解析を行う
	 *
	 * @param length1
	 *            和音に続く長さ ('/'より前の部分)
	 * @param length2
	 *            和音に続く長さ ('/'以降)
	 */
	private void parseChordEnd(String length1, String length2) {
		Fraction oLength = this.parseLength(length1, length2);

		Track track = this.getActiveTrack();
		String trackId = this.activeTrackId;

		if (this.isParsingChordExist) {
			// 直前のNoteの取得
			Note nextNote = this.isParsingGrace
					? this.getActiveTrackData().getGraceNote().getNotes().getLast()
					: this.getTailMelody(track, 1).getFirst().getValue();

			// 長さのかけ合わせ
			for (NoteUnit unit : nextNote.getNotes()) {
				unit.setRealLength(unit.getRealLength().mul(oLength));
			}

			// 実際の音符の長さの再計算
			Fraction unitNoteLength = track.getUnitNoteLength();
			if (unitNoteLength == null) {
				unitNoteLength = InformationFieldLength.getDefaultLength(track.getMeterLength());
			}
			nextNote.calcRealLength(unitNoteLength);

			// Broken Rhythmや連音符の長さの再計算
			this.calcPosition(trackId);
		}

		// 和音解析終了
		this.isParsingChord = false;
		this.isParsingChordExist = false;
		this.getActiveTrackData().clearChordDecorations();
	}

	/**
	 * 休符部分の解析を行う
	 *
	 * @param rest
	 *            休符の種類
	 * @param length1
	 *            長さ ('/'より前の部分)
	 * @param length2
	 *            長さ ('/'以降)
	 */
	private void parseRest(String rest, String length1, String length2) {
		NoteUnit oNoteUnit = new NoteUnit();

		oNoteUnit.setLength(this.parseLength(length1, length2));
		oNoteUnit.setNote(rest);

		this.getActiveNextNote().addNote(oNoteUnit);

		this.addActiveNextNote();
	}

	/**
	 * 小節単位の休符部分の解析を行う
	 *
	 * @param rest
	 *            休符の種類
	 * @param length
	 *            長さ
	 */
	private void parseMeasureRest(String rest, String length) {
		NoteUnit oNoteUnit = new NoteUnit();

		int oLength = 1;
		if (length != null) {
			oLength = Integer.parseInt(length);
		}

		oNoteUnit.setLength(new Fraction(oLength));
		oNoteUnit.setNote(rest);

		this.getActiveNextNote().addNote(oNoteUnit);

		this.addActiveNextNote();
	}

	/**
	 * 縦線の解析を行う
	 *
	 * @param bar
	 *            縦線
	 * @param isPreBar
	 *            直前も縦線であったか。true: 直前も縦線。
	 */
	private void parseBar(String bar, boolean isPreBar) {
		ParseDataTrack trackData = this.getActiveTrackData();

		// Voice Overlay
		int overlay = trackData.getVoiceOverlay();
		if (overlay > 0) {
			trackData.setVoiceOverlay(--overlay);

			// Voice Overlay 終了
			if (overlay == 0) {
				this.getActiveTrack().clearAccidental();
				String[] idOverlay = getTrackIdAndOverlay(this.activeTrackId);
				this.newTrack(idOverlay[0]);
			}
		}

		this.getActiveNextNote().addBar(bar, isPreBar);

		// 新しい小節になっていれば臨時記号を初期化
		this.getActiveTrack().clearAccidental();
	}

	/**
	 * 繰り返し順番の解析を行う
	 *
	 * @param order
	 *            繰り返し順番を表す文字列
	 */
	private void parseRepeatOrder(String order) {
		String[] orders = order.split(",");
		LinkedList<Integer> list = new LinkedList<Integer>();

		for (String o : orders) {
			String[] oo = o.split("-");
			if (oo.length == 1) {
				list.add(Integer.parseInt(oo[0]));
			} else {
				int a = Integer.parseInt(oo[0]);
				int b = Integer.parseInt(oo[1]);
				if (b < a) {
					int temp = a;
					a = b;
					b = temp;
				}
				for (int i = a; i <= b; ++i) {
					list.add(i);
				}
			}
		}

		Collections.sort(list);
		this.getActiveNextNote().setOrders(list);
	}

	/**
	 * 装飾部分の解析を行う
	 *
	 * @param decoraton
	 *            装飾を表す文字列 (!～!の!を除いた部分、もしくは単一文字(~、Hなど)
	 */
	private void parseDecoration(String decoraton) {
		String deco = decoraton;

		if (deco.startsWith("fff")) {
			deco = "fff";
		}
		if (deco.startsWith("ppp")) {
			deco = "ppp";
		}

		Note nextNote = this.getActiveNextNote();
		LinkedList<EDecoration> noteDeco = nextNote.getNoteDecorations();
		LinkedList<EDecoration> unitDeco = nextNote.getUnitDecorations();
		switch (deco) {
		case "ppp":
			nextNote.setDynamics(EDynamics.PPP);
			break;
		case "pp":
			nextNote.setDynamics(EDynamics.PP);
			break;
		case "p":
			nextNote.setDynamics(EDynamics.P);
			break;
		case "mp":
			nextNote.setDynamics(EDynamics.MP);
			break;
		case "mf":
			nextNote.setDynamics(EDynamics.MF);
			break;
		case "f":
			nextNote.setDynamics(EDynamics.F);
			break;
		case "ff":
			nextNote.setDynamics(EDynamics.FF);
			break;
		case "fff":
			nextNote.setDynamics(EDynamics.FFF);
			break;
		case ">":
		case "accent":
		case "emphasis":
			unitDeco.add(EDecoration.ACCENT);
			break;
		case ".":
			unitDeco.add(EDecoration.STACCATO);
			break;
		case "<(":
		case "crescendo(":
			if (this.opt.isDynamicsVolume()) {
				noteDeco.add(EDecoration.CRES_START);
			} else {
				noteDeco.add(EDecoration.TRACK_CRES_START);
			}
			break;
		case "<)":
		case "crescendo)":
			if (this.opt.isDynamicsVolume()) {
				noteDeco.add(EDecoration.CRES_END);
			} else {
				noteDeco.add(EDecoration.TRACK_CRES_END);
			}
			break;
		case ">(":
		case "diminuendo(":
			if (this.opt.isDynamicsVolume()) {
				noteDeco.add(EDecoration.DIMI_START);
			} else {
				noteDeco.add(EDecoration.TRACK_DIMI_START);
			}
			break;
		case ">)":
		case "diminuendo)":
			if (this.opt.isDynamicsVolume()) {
				noteDeco.add(EDecoration.DIMI_END);
			} else {
				noteDeco.add(EDecoration.TRACK_DIMI_END);
			}
			break;
		case "arpeggio":
			noteDeco.add(EDecoration.ARPEGGIO);
			break;
		case "(":
			noteDeco.add(EDecoration.SLUR_START);
			break;
		case ")":
			// スラー終了は直前のNoteに付与する(面倒…)
			this.getTailMelody(this.getActiveTrack(), 1).getFirst().getValue().getNoteDecorations().add(
					EDecoration.SLUR_END);
			break;
		case "rit(":
			noteDeco.add(EDecoration.RITA_START);
			break;
		case "rit)":
			noteDeco.add(EDecoration.RITA_END);
			break;
		case "accel(":
			noteDeco.add(EDecoration.ACCE_START);
			break;
		case "accel)":
			noteDeco.add(EDecoration.ACCE_END);
			break;
		case "t<(":
			if (this.opt.isDynamicsVolume()) {
				noteDeco.add(EDecoration.TRACK_CRES_START);
			} else {
				noteDeco.add(EDecoration.CRES_START);
			}
			break;
		case "t<)":
			if (this.opt.isDynamicsVolume()) {
				noteDeco.add(EDecoration.TRACK_CRES_END);
			} else {
				noteDeco.add(EDecoration.CRES_END);
			}
			break;
		case "t>(":
			if (this.opt.isDynamicsVolume()) {
				noteDeco.add(EDecoration.TRACK_DIMI_START);
			} else {
				noteDeco.add(EDecoration.DIMI_START);
			}
			break;
		case "t>)":
			if (this.opt.isDynamicsVolume()) {
				noteDeco.add(EDecoration.TRACK_DIMI_END);
			} else {
				noteDeco.add(EDecoration.DIMI_END);
			}
			break;
		case "trill":
			noteDeco.add(EDecoration.TRILL);
			break;
		case "trill(":
			noteDeco.add(EDecoration.TRILL_START);
			break;
		case "trill)":
			noteDeco.add(EDecoration.TRILL_END);
			break;
		case "lowermordent":
		case "mordent":
			noteDeco.add(EDecoration.LOWER_MORDENT);
			break;
		case "uppermordent":
		case "pralltriller":
			noteDeco.add(EDecoration.UPPER_MORDENT);
			break;
		case "turn":
		case "invertedturnx":
			noteDeco.add(EDecoration.TURN);
			break;
		case "turnx":
		case "invertedturn":
			noteDeco.add(EDecoration.TURNX);
			break;

		case "^^":
			noteDeco.add(EDecoration.DOUBLE_SHARP);
			break;
		case "^":
			noteDeco.add(EDecoration.SHARP);
			break;
		case "=":
			noteDeco.add(EDecoration.NATURAL);
			break;
		case "_":
			noteDeco.add(EDecoration.FLAT);
			break;
		case "__":
			noteDeco.add(EDecoration.DOUBLE_FLAT);
			break;

		default:
			// TODO 他のも実装？
			Debug.out("Ignored Decoration -> " + deco);
			noteDeco.add(EDecoration.NONE);
			break;
		}

	}

	/**
	 * アノテーションの解析を行う。(未実装)
	 *
	 * @param position
	 *            ポジション
	 * @param text
	 *            テキスト
	 */
	private void parseAnnotation(String position, String text) {
		// 独自仕様
		Matcher m = Constant.PTN_VELOCITY.matcher(text);
		if (m.find()) {
			this.getActiveNextNote().setVelocity(Integer.parseInt(m.group(1)));
			return;
		}

		m = Constant.PTN_VOLUME.matcher(text);
		if (m.find()) {
			this.getActiveNextNote().setVolume(Integer.parseInt(m.group(1)));
			return;
		}
		switch (text) {
		case "<(":
		case "<)":
		case ">(":
		case ">)":
			this.parseDecoration("t" + text);
			break;
		case "rit(":
		case "rit)":
		case "accel(":
		case "accel)":
			this.parseDecoration(text);
			break;
		default:
			// TODO 無視で良い？
			Debug.out("Ignored Annotation -> " + position + text);
			break;
		}
	}

	/**
	 * 和音文字列の部分を解析を行う。(未実装)
	 *
	 * @param chord
	 *            音階+臨時記号
	 * @param type
	 *            maj、min、など
	 * @param base
	 *            ベース音
	 */
	private void parseChordString(String chord, String type, String base) {
		// TODO 実装？
		// 和音の音を作ったほうが良い？ 面倒なのでアノテーションとして保持するのみ
		Debug.out("Chord -> " + chord + type + base);
	}

	/**
	 * Broken Rhythmの解析を行う。
	 *
	 * @param broken
	 *            Broken Rhythm
	 */
	private void parseBroken(String broken) {
		int num = broken.length();
		if (broken.startsWith("<")) {
			num = -num;
		}
		this.getActiveNextNote().setBroken(num);
	}

	/**
	 * タイの解析を行う。
	 *
	 * @param tie
	 *            タイ
	 */
	private void parseTie(String tie) {
		Note note;

		if (this.isParsingGrace) {
			note = this.getActiveTrackData().getGraceNote().getNotes().getLast();
		} else {
			Track track = this.getActiveTrack();
			note = this.getTailMelody(track, 1).getFirst().getValue();
		}
		if (this.isParsingChord) {
			// 和音解析中だったら直前のNoteUnitに追加
			note.getNotes().getLast().setWithTie(true);
		} else {
			// 和音解析中でなければ、直前のNoteの全てのNoteUnitに追加
			for (NoteUnit unit : note.getNotes()) {
				unit.setWithTie(true);
			}
		}
	}

	/**
	 * 連音符 (noteCount:wholeTime:nextNoteCount の解析を行う。
	 *
	 * @param noteCount
	 *            nextCount
	 * @param wholeTime
	 *            wholeTime
	 * @param nextNoteCount
	 *            nextNoteCount
	 */
	private void parseGroupNote(String noteCount, String wholeTime, String nextNoteCount) {
		int nNoteCount;
		int nWholeTime;
		int nNextNoteCount;

		nNoteCount = Integer.parseInt(noteCount);

		if (wholeTime != null) {
			nWholeTime = Integer.parseInt(wholeTime);
		} else {
			switch (nNoteCount) {
			case 3:
			case 6:
				nWholeTime = 2;
				break;
			case 2:
			case 4:
			case 8:
				nWholeTime = 3;
				break;
			case 5:
			case 7:
			case 9:
				Fraction meterLength = this.getActiveTrack().getMeterLength();
				if (meterLength == null) {
					meterLength = InformationFieldMeter.getDefaultLength();
				}

				if (meterLength.equals(new Fraction(6, 8))
						|| meterLength.equals(new Fraction(9, 8))
						|| meterLength.equals(new Fraction(12, 8))) {
					nWholeTime = 3;
				} else {
					nWholeTime = 2;
				}
				break;
			default:
				throw new RuntimeException("Invalid Note Group NoteCount->" + nNoteCount);
			}
		}

		if (nextNoteCount != null) {
			nNextNoteCount = Integer.parseInt(nextNoteCount);
		} else {
			nNextNoteCount = nNoteCount;
		}

		this.getActiveNextNote().setGroupNote(new GroupNote(nNoteCount, nWholeTime, nNextNoteCount));
	}

	/**
	 * Voice Overlay部分の解析を行う
	 *
	 * @param op
	 *            Voice Overlay開始を意味するオペレータ
	 */
	private void parseVoiceOverlay(String op) {
		int count = op.length();
		String[] idOverlay = getTrackIdAndOverlay(this.activeTrackId);
		String mainId = idOverlay[0];
		Track trackMain = this.tune.getTracks().get(mainId);
		ParseDataTrack trackDataMain = this.trackData.get(mainId);
		Fraction startPos = this.lastBarPos(trackMain, count);

		int overlay = getEmptyOverlayNo(mainId, 1 + Integer.parseInt(idOverlay[1]), startPos);
		String newTrackId = mainId + Constant.STR_OVERLAY_SEP + overlay;

		this.newTrack(newTrackId);

		this.getActiveTrackData().setVoiceOverlay(count);

		// information fieldなど必要な物をコピー
		Track trackNew = this.getActiveTrack();
		trackNew.setInformationField(trackMain.getInformationField());
		trackNew.setParent(trackMain.getParent());

		ParseDataTrack trackDataNew = this.getActiveTrackData();
		trackDataNew.setClef(trackDataMain.getClef());

		this.getActiveTrack().setEmptyNoteAt(startPos);
	}

	/**
	 * 前打音の部分の解析を行う(開始)
	 *
	 * @param shortGrace
	 *            短前打音なら"/"を、そうでなければnullをセットする。
	 */
	private void parseGraceStart(String shortGrace) {
		this.getActiveTrackData().initGrace(shortGrace != null);
		this.isParsingGrace = true;
	}

	/**
	 * 前打音の部分の解析を行う(終了)
	 */
	private void parseGraceEnd() {
		this.isParsingGrace = false;
		ParseDataTrack trackData = this.getActiveTrackData();
		this.getActiveNextNote().setGrace(trackData.getGraceNote());
		this.getActiveTrackData().setGraceNote(null);
	}

	/**
	 * 装飾記号を音符間に記載するためのスペース部分の解析を行う
	 */
	private void parseSpace() {
		Note prevNote = this.getTailMelody(this.getActiveTrack(), 1).getFirst().getValue();
		LinkedList<EDecoration> prevDeco = prevNote.getNoteDecorations();
		for (EDecoration deco : this.getActiveNextNote().getNoteDecorations()) {
			switch (deco) {
			case TURN:
				prevDeco.add(EDecoration.AFTER_TURN);
				break;
			case TURNX:
				prevDeco.add(EDecoration.AFTER_TURNX);
				break;
			default:
				prevDeco.add(deco);
				break;
			}
		}

		// 次回解析Noteを新規で作成する
		this.newNote();
	}

	/**
	 * Program ID部分の解析を行う
	 *
	 * @param id
	 *            Program ID
	 */
	private void parseProgramId(String id) {
		Note note = this.getActiveTrackData().getNextNote();
		note.setProgramId(Integer.parseInt(id));
	}

	/**************************************************************************
	 * XXX Track関連の関数群
	 **************************************************************************/

	/**
	 * 次回解析Noteを新規で作成する
	 */
	private void newNote() {
		if (this.isParsingGrace) {
			this.getActiveTrackData().setNextGraceNote(new Note());
		} else {
			this.getActiveTrackData().setNextNote(new Note());
		}
	}

	/**
	 * 新しいTrackに解析対象を移す。解析対象がなければ新しくTrackを追加する。
	 *
	 * @param id
	 *            新しく追加するTrackのid
	 */
	private void newTrack(String id) {
		Track track;
		ParseDataTrack trackData;
		String defaultId = InformationFieldVoice.DEFAULT_ID;
		String[] idOverlay = getTrackIdAndOverlay(id);

		// メイントラック追加時、デフォルトのTrackがあれば、それを先頭に取り込む
		if (idOverlay[1].equals("0")
				&& getTrackIdAndOverlay(this.activeTrackId)[0].equals(defaultId)) {
			String mainId = idOverlay[0];
			LinkedList<String> list = new LinkedList<String>();

			for (String trackId : this.tune.getTracks().keySet()) {
				idOverlay = getTrackIdAndOverlay(trackId);
				if (idOverlay[0].equals(defaultId)) {
					list.add(trackId);
				}
			}

			for (String trackId : list) {
				idOverlay = getTrackIdAndOverlay(trackId);
				String newId = mainId;
				if (!idOverlay[1].equals("0")) {
					newId += Constant.STR_OVERLAY_SEP + idOverlay[1];
				}

				track = this.tune.getTracks().remove(trackId);
				track.setTrackId(newId);
				this.tune.addTrack(newId, track);

				trackData = this.trackData.remove(trackId);
				this.trackData.put(newId, trackData);
			}
		}

		this.activeTrackId = id;
		track = this.getActiveTrack();
		if (track == null) {
			Track newTrack = new Track();
			newTrack.setTrackId(id);
			this.tune.addTrack(id, newTrack);
		}
		trackData = this.trackData.get(id);
		if (trackData == null) {
			this.trackData.put(id, new ParseDataTrack());
			this.updateTrackDataClefFromTrack(id);
		}

	}

	/**
	 * 解析中のTrackを取得する
	 *
	 * @return 解析中のTrack
	 */
	private Track getActiveTrack() {
		return this.tune.getTracks().get(this.activeTrackId);
	}

	/**
	 * 解析中のTrackデータを取得する
	 *
	 * @return 解析中のTrackデータ
	 */
	private ParseDataTrack getActiveTrackData() {
		return this.trackData.get(this.activeTrackId);
	}

	/**
	 * 解析中Trackの次回追加予定Noteを取得する
	 *
	 * @return 次回追加予定のNote
	 */
	private Note getActiveNextNote() {
		ParseDataTrack trackData = this.getActiveTrackData();
		return this.isParsingGrace ? trackData.getNextGraceNote() : trackData.getNextNote();
	}

	/**
	 * 次回追加するNoteを解析中のTrackへ追加する
	 *
	 */
	private void addActiveNextNote() {
		// キモとなるメソッド (気合を入れて！)

		// 必要な項目取得
		Track track = this.getActiveTrack();
		String trackId = this.activeTrackId;
		Note nextNote = this.getActiveNextNote();

		// 縦線のくり返し部分の解析を行う
		this.updateRepeatBar(nextNote);

		// information fieldのコピー
		this.copyInformationField(nextNote, track);

		// フィールドの値を取得
		Fraction unitNoteLength = track.getUnitNoteLength();
		Fraction meterLength = track.getMeterLength();
		if (unitNoteLength == null) {
			unitNoteLength = InformationFieldLength.getDefaultLength(meterLength);
		}
		if (meterLength == null) {
			meterLength = InformationFieldMeter.getDefaultLength();
		}

		// parseBarに移動
		// // 新しい小節になっていれば臨時記号を初期化
		// if (nextNote.getBar().length() > 0) {
		// track.clearAccidental();
		// }

		// Noteの長さを「L:に対する相対的な長さ」から「実際の音符の長さ」にする
		this.calcRealLength(nextNote, unitNoteLength, meterLength);

		// octave変化
		this.applyOctave(trackId, nextNote);

		// octave情報を最適化 (臨時記号の更新/Transposition変化のため)
		this.optimizeOctave(nextNote);

		// 臨時記号を更新
		this.updateAccidental(track, nextNote);

		// キー情報を適用
		this.applyKey(track, nextNote);

		// transposition変化
		this.applyTransposition(trackId, nextNote);

		// octave情報を最適化
		this.optimizeOctave(nextNote);

		// 装飾を更新
		this.updateUnitDecoration(trackId, nextNote);

		// Noteの追加
		if (this.isParsingGrace) {
			this.addNoteInGrace(trackId, nextNote);
		} else {
			this.addNote(track, nextNote);
		}

		if (this.isParsingChord) {
			// 和音解析中の場合、和音の音符の存在フラグを立てる
			this.isParsingChordExist = true;
		} else {
			// 和音解析中でない場合
			// 音符の位置の再計算
			this.calcPosition(trackId);
		}

		// 次回解析Noteを新規で作成する
		this.newNote();
	}

	/**
	 * TrackにNoteを追加する
	 *
	 * @param track
	 *            このTrackにNoteを追加する
	 * @param note
	 *            追加するNote
	 */
	private void addNote(Track track, Note note) {
		Entry<Fraction, Note> entry = this.getTailMelody(track, 1).getFirst();
		Note prevNote = entry.getValue();

		if (this.isParsingChord && this.isParsingChordExist) {
			// 和音解析中
			// Information Fieldのコピー
			this.copyInformationField(note, prevNote);

			// Noteの追加
			for (NoteUnit unit : note.getNotes()) {
				prevNote.addNote(unit);
			}

			// 繰り返し情報などはないはず(和音中に書かないよね？)なのでコピー不要？
		} else {
			// 和音解析中ではないか、解析中でも1つの音符の解析も済んでいない場合
			Fraction nextPos = Fraction.add(entry.getKey(), prevNote.getRealLength());
			track.getMelody().put(nextPos, note);
		}
	}

	/**
	 * GraceNotesにNoteを追加する
	 *
	 * @param trackId
	 *            このTrackIDを持つTrackのGraceNotesにNoteを追加する
	 * @param note
	 *            追加するNote
	 */
	private void addNoteInGrace(String trackId, Note note) {
		ParseDataTrack trackData = this.trackData.get(trackId);

		if (this.isParsingChord && this.isParsingChordExist) {
			// 和音解析中
			Note prevNote = trackData.getGraceNote().getNotes().getLast();

			// Information Fieldのコピー
			this.copyInformationField(note, prevNote);

			// Noteの追加
			for (NoteUnit unit : note.getNotes()) {
				prevNote.addNote(unit);
			}

			// 繰り返し情報などはないはず(和音中に書かないよね？)なのでコピー不要？
		} else {
			// 和音解析中ではないか、解析中でも1つの音符の解析も済んでいない場合
			trackData.addNoteToGrace(note);
		}
	}

	/**
	 * Broken Rhythmや連音符の位置・長さ再計算を行う。
	 * <ol>
	 * <li>Broken Rhythmの再計算</li>
	 * <li>連符の再計算</li>
	 * </ol>
	 *
	 * @param trackId
	 *            位置の再計算を行うTrackId
	 */
	private void calcPosition(String trackId) {
		if (this.isParsingGrace) {
			LinkedList<Note> notes = this.getActiveTrackData().getGraceNote().getNotes();
			Note prevNote = notes.getLast();
			if (prevNote.getBroken() != 0) {
				this.updateBrokenRhythmInGrace(notes);
			}
		} else {
			Track track = this.tune.getTracks().get(trackId);
			Note prevNote = this.getTailMelody(track, 1).getFirst().getValue();

			// ">"の長さ再計算 (Note追加後なので位置がずれている)
			if (prevNote.getBroken() != 0) {
				this.updateBrokenRhythm(track);
			}

			// 連音符
			LinkedList<int[]> groupNoteStartPos = this.trackData.get(trackId).getGroupNoteStartPos();

			// 新規連音符を追加
			GroupNote groupNote = prevNote.getGroupNote();
			if (groupNote != null) {
				groupNoteStartPos.add(new int[] { groupNote.getNextNoteCount(),
						groupNote.getNextNoteCount() });
			}

			// 連音符の終わりがあれば、長さ・位置の更新を行う
			for (Iterator<int[]> iter = groupNoteStartPos.iterator(); iter.hasNext();) {
				int[] nn = iter.next();
				if (--nn[ParseDataTrack.GROUP_NOTE_REMAIN] == 0) {
					this.updateGroupNote(track, nn[ParseDataTrack.GROUP_NOTE_TOTAL_COUNT]);
					iter.remove();
				}
			}
		}
	}

	/**
	 * 縦線のくり返し部分を解析して、繰り返しStart/End情報を更新する
	 *
	 * @param note
	 *            解析するNote
	 */
	private void updateRepeatBar(Note note) {
		String bar = note.getBar();
		Matcher m;

		m = Constant.PTN_REPEAT_START.matcher(bar);
		if (m.find()) {
			note.setRepeatStart(m.group(1).length());
		}

		m = Constant.PTN_REPEAT_END.matcher(bar);
		if (m.find()) {
			note.setRepeatEnd(m.group(1).length());
		}

		m = Constant.PTN_REPEAT_START_END.matcher(bar);
		if (m.find()) {
			// TODO Start:1、End:それ以外として後から変更する？
			int start = 1;
			note.setRepeatStart(start);
			note.setRepeatEnd(m.group(1).length() - start);
		}
	}

	/**
	 * Information FieldをNoteからTrackへコピーする
	 *
	 * @param note
	 *            コピー元のNote
	 * @param track
	 *            コピー先のTrack
	 */
	private void copyInformationField(Note note, Track track) {
		MapInformationField mapField = note.getInformationField();
		for (String name : mapField.keySet()) {
			for (AInformationField field : mapField.get(name)) {
				track.addInformation(name, field);
			}
		}
	}

	/**
	 * Information FieldをNoteからNoteへコピーする
	 *
	 * @param noteSrc
	 *            コピー元のNote
	 * @param noteDest
	 *            コピー先のNote
	 */
	private void copyInformationField(Note noteSrc, Note noteDest) {
		MapInformationField mapField = noteSrc.getInformationField();
		for (String name : mapField.keySet()) {
			for (AInformationField field : mapField.get(name)) {
				noteDest.addInformation(name, field);
			}
		}
	}

	/**
	 * Noteの長さを「L:に対する相対的な長さ」から「実際の音符の長さ」にする。
	 *
	 * @param note
	 *            長さを変更するNote
	 * @param unitNoteLength
	 *            現在のNote Unit Length (L:フィールドの値)
	 * @param meterLength
	 *            現在のMeter Length (M:フィールドの値)
	 */
	private void calcRealLength(Note note, Fraction unitNoteLength, Fraction meterLength) {
		// NoteUnitの長さを計算する
		for (NoteUnit unit : note.getNotes()) {
			// Note長さを実際の音符の長さにする
			if (!unit.isRealLength()) {
				if (unit.getNote().equals("X") || unit.getNote().equals("Z")) {
					// X,Zは小節単位の休符のため別計算
					// 1小節分の実際の音符の長さはMeter Length(M:フィールドの値)にする
					unit.setRealLength(unit.getLength().mul(meterLength));
					// プログラム内部では小節単位の休符は扱わないため、小文字に変換
					unit.setNote(unit.getNote().toLowerCase());
				} else {
					// 小節単位の休符以外の場合
					// unit note length(L:フィールド)の値とABC譜表記上の長さをかけ
					// 実際の音符の長さにする
					unit.setRealLength(unit.getLength().mul(unitNoteLength));
				}
			}
		}

		// Note自体の長さを計算する
		note.calcRealLength(unitNoteLength);
	}

	/**
	 * オクターブ情報を音符に適用する
	 *
	 * @param trackId
	 *            適用するTrackID
	 * @param nextNote
	 *            適用するNote
	 */
	private void applyOctave(String trackId, Note nextNote) {
		ParseDataTrack trackData = this.trackData.get(trackId);
		if (trackData.getOctaveTotal() != null && trackData.getOctaveTotal() != 0) {
			for (NoteUnit unit : nextNote.getNotes()) {
				if ("xzXZ".indexOf(unit.getNote()) >= 0) {
					continue;
				}
				unit.setOctave(unit.getOctave() + trackData.getOctaveTotal());
			}
		}
	}

	/**
	 * 移調情報を音符に適用する
	 *
	 * @param trackId
	 *            適用するTrackID
	 * @param nextNote
	 *            適用するNote
	 */
	private void applyTransposition(String trackId, Note nextNote) {
		Track track = this.tune.getTracks().get(trackId);
		ParseDataTrack trackData = this.trackData.get(trackId);
		String stepString = "C_D_EF_G_A_Bc_d_ef_g_a_b";

		if (trackData.getTransposition() != null && trackData.getTransposition() != 0) {
			for (NoteUnit unit : nextNote.getNotes()) {
				int step = stepString.indexOf(unit.getNote());
				if (step < 0) {
					// 休符
					continue;
				}

				// 音の高さの計算
				step += unit.getOctave() * 12;
				// 臨時記号
				EAccidental acci = unit.getAcci();
				if (acci != null) {
					switch (acci) {
					case DOUBLE_FLAT:
						step -= 2;
						break;
					case FLAT:
						step -= 1;
						break;
					case SHARP:
						step += 1;
						break;
					case DOUBLE_SHARP:
						step += 2;
						break;
					default:
						break;
					}
				}
				// 移調
				step += trackData.getTransposition();

				// 音の高さ->音符へ戻す
				int octave = 0;
				while (step < 0) {
					--octave;
					step += 12;
				}
				while (step >= stepString.length()) {
					++octave;
					step -= 12;
				}
				if (octave > 0 && step < 12) {
					// オクターブ高いのに大文字
					--octave;
					step += 12;
				}
				if (octave < 0 && step >= 12) {
					// オクターブ低いのに小文字
					++octave;
					step -= 12;
				}
				unit.setOctave(octave);

				String noteString = stepString.substring(step, step + 1);
				if (noteString.equals("_")) {
					// 半音。下の音より半音上げる(#)か上の音より半音下げる(b)か
					noteString = stepString.substring(step - 1, step);
					acci = track.getAccidental(noteString + octave);
					if (acci == null) {
						// 臨時記号がなければKeyより取得
						LinkedList<AInformationField> field = track.getInformation("K");
						if (field != null) {
							Key key = ((InformationFieldKey) field.getLast()).getKey();
							acci = key.getAccidental(unit.getNote());
						}
					}
					if (acci == EAccidental.SHARP) {
						// 臨時記号と合っているし、これに決定！
						unit.setNote(noteString);
						unit.setAcci(EAccidental.SHARP);
					} else {
						// XXX 半音下げるしかない？ (ダブルシャープ、ダブルフラットは無視で良いよね？
						noteString = stepString.substring(step + 1, step + 2);
						unit.setNote(noteString);
						unit.setAcci(EAccidental.FLAT);
					}
				} else {
					unit.setNote(noteString);
					unit.setAcci(EAccidental.NATURAL);
				}
			}
		}
	}

	/**
	 * キー情報を音符に適用する
	 */
	private void applyKey(Track track, Note nextNote) {
		LinkedList<AInformationField> list = track.getInformation("K");
		if (list != null) {
			Key key = ((InformationFieldKey) list.getLast()).getKey();
			for (NoteUnit unit : nextNote.getNotes()) {
				if ("xzXZ".indexOf(unit.getNote()) >= 0) {
					continue;
				}
				if (unit.getAcci() == null) {
					unit.setAcci(key.getAccidental(unit.getNote()));
				}
			}
		}
	}

	/**
	 * オクターブ情報を最適化する。C'やc,という表現を最適化する
	 *
	 * @param nextNote
	 *            適用するNote
	 */
	private void optimizeOctave(Note nextNote) {
		for (NoteUnit unit : nextNote.getNotes()) {
			if (unit.getOctave() > 0 && unit.getNote().equals(unit.getNote().toUpperCase())) {
				// オクターブ高いのに、大文字が使われている場合
				unit.setOctave(unit.getOctave() - 1);
				unit.setNote(unit.getNote().toLowerCase());
			}
			if (unit.getOctave() < 0 && unit.getNote().equals(unit.getNote().toLowerCase())) {
				// オクターブ低いのに、小文字が使われている場合
				unit.setOctave(unit.getOctave() + 1);
				unit.setNote(unit.getNote().toUpperCase());
			}
		}
	}

	/**
	 * 臨時記号を更新する。更新はNote->Track、Track->Noteの順番で行う。
	 *
	 * @param track
	 *            臨時記号を更新するTrack
	 * @param note
	 *            このNoteの臨時記号を元にTrackの臨時記号情報を更新する
	 */
	private void updateAccidental(Track track, Note note) {
		for (NoteUnit unit : note.getNotes()) {
			// 小節毎の臨時記号を反映or設定
			EAccidental acci = unit.getAcci();
			String noteOctave = unit.getNote() + unit.getOctave();
			if (acci != null) {
				track.putAccidental(noteOctave, acci);
			} else {
				if ((acci = track.getAccidental(noteOctave)) != null) {
					unit.setAcci(acci);
				}
			}
		}
	}

	/**
	 * 解析中Trackに追加されている音符の位置と音符を後ろから指定の個数分取得する。位置は昇順にソートされている。
	 *
	 * @param track
	 *            位置・音符を取得するTrack
	 * @param count
	 *            取得する個数
	 * @return 位置・音符のリスト
	 */
	private LinkedList<Entry<Fraction, Note>> getTailMelody(Track track, int count) {
		LinkedList<Entry<Fraction, Note>> list = new LinkedList<Entry<Fraction, Note>>();

		NavigableMap<Fraction, Note> map = track.getMelody().descendingMap();
		for (Iterator<Entry<Fraction, Note>> iter = map.entrySet().iterator(); count > 0
				&& iter.hasNext(); --count) {
			list.addFirst(iter.next());
		}
		return list;
	}

	/**
	 * Broken Rhythmを、実際の長さ、演奏位置にする。前のNoteの長さと後ろのNoteの長さが違うと想定通りの結果とならない。
	 * (ABCフォーマットでもそのように規定されてるはず)
	 *
	 * @param track
	 *            Track
	 */
	private void updateBrokenRhythm(Track track) {
		// 前後の音符を取得
		LinkedList<Entry<Fraction, Note>> entryList = this.getTailMelody(track, 2);

		Note prevNote = entryList.get(0).getValue();
		Fraction nextPos = entryList.get(1).getKey();
		Note nextNote = entryList.get(1).getValue();

		int broken = nextNote.getBroken();

		if (nextNote == null || prevNote == null || broken == 0) {
			return;
		}

		// ">"や"<"の個数によってかける倍率の計算
		Fraction scale = new Fraction(1, (int) Math.pow(2, Math.abs(broken)));
		if (broken > 0) {
			// ">"の時

			// 後ろのNoteの位置を (1 - 1/2^(">"の個数)) * (前のNoteの長さ) だけ後ろにずらす
			track.getMelody().remove(nextPos);
			track.getMelody().put(
					nextPos.clone().add(new Fraction(1).sub(scale).mul(nextNote.getRealLength())),
					nextNote);

			// 前のNoteの長さを (2 - 1/2^(">"の個数)) 倍する
			Fraction prevLength = new Fraction(2).sub(scale);
			for (NoteUnit unit : prevNote.getNotes()) {
				unit.setRealLength(unit.getRealLength().mul(prevLength));
			}
			prevNote.setRealLength(prevNote.getRealLength().mul(prevLength));

			// 後ろのNoteの長さを (1/2^(">"の個数)) 倍する
			for (NoteUnit unit : nextNote.getNotes()) {
				unit.setRealLength(unit.getRealLength().mul(scale));
			}
			nextNote.setRealLength(nextNote.getRealLength().mul(scale));
		} else {
			// "<"の時

			// 後ろのNoteの位置を (1 - 1/2^("<"の個数)) * (前のNoteの長さ) だけ前にずらす
			track.getMelody().remove(nextPos);
			track.getMelody().put(
					nextPos.clone().sub(new Fraction(1).sub(scale).mul(nextNote.getRealLength())),
					nextNote);

			// 前のNoteの長さを (1/2^("<"の個数)) 倍する
			for (NoteUnit unit : prevNote.getNotes()) {
				unit.setRealLength(unit.getRealLength().mul(scale));
			}
			prevNote.setRealLength(prevNote.getRealLength().mul(scale));

			// 後ろのNoteの長さを (2 - 1/2^("<"の個数)) 倍する
			Fraction nextLength = new Fraction(2).sub(scale);
			for (NoteUnit unit : nextNote.getNotes()) {
				unit.setRealLength(unit.getRealLength().mul(nextLength));
			}
			nextNote.setRealLength(nextNote.getRealLength().mul(nextLength));
		}
	}

	/**
	 * Broken Rhythmを、実際の長さ、演奏位置にする。前のNoteの長さと後ろのNoteの長さが違うと想定通りの結果とならない。
	 * (ABCフォーマットでもそのように規定されてるはず)
	 *
	 * @param track
	 *            Track
	 */
	private void updateBrokenRhythmInGrace(LinkedList<Note> notes) {
		// 前後の音符を取得
		Note prevNote = notes.get(notes.size() - 2);
		Note nextNote = notes.getLast();

		int broken = nextNote.getBroken();

		if (nextNote == null || prevNote == null || broken == 0) {
			return;
		}

		// ">"や"<"の個数によってかける倍率の計算
		Fraction scale = new Fraction(1, (int) Math.pow(2, Math.abs(broken)));
		if (broken > 0) {
			// ">"の時

			// 前のNoteの長さを (2 - 1/2^(">"の個数)) 倍する
			Fraction prevLength = new Fraction(2).sub(scale);
			for (NoteUnit unit : prevNote.getNotes()) {
				unit.setRealLength(unit.getRealLength().mul(prevLength));
			}
			prevNote.setRealLength(prevNote.getRealLength().mul(prevLength));

			// 後ろのNoteの長さを (1/2^(">"の個数)) 倍する
			for (NoteUnit unit : nextNote.getNotes()) {
				unit.setRealLength(unit.getRealLength().mul(scale));
			}
			nextNote.setRealLength(nextNote.getRealLength().mul(scale));
		} else {
			// "<"の時

			// 前のNoteの長さを (1/2^("<"の個数)) 倍する
			for (NoteUnit unit : prevNote.getNotes()) {
				unit.setRealLength(unit.getRealLength().mul(scale));
			}
			prevNote.setRealLength(prevNote.getRealLength().mul(scale));

			// 後ろのNoteの長さを (2 - 1/2^("<"の個数)) 倍する
			Fraction nextLength = new Fraction(2).sub(scale);
			for (NoteUnit unit : nextNote.getNotes()) {
				unit.setRealLength(unit.getRealLength().mul(nextLength));
			}
			nextNote.setRealLength(nextNote.getRealLength().mul(nextLength));
		}
	}

	/**
	 * 連音符を、実際の長さ、演奏位置にする。適用範囲のNoteの長さが異なると想定通りの結果とならない。
	 * (ABCフォーマットでもそのように規定されてるはず)
	 *
	 * @param track
	 *            Track
	 * @param count
	 *            連音符の適用される音符の個数
	 */
	private void updateGroupNote(Track track, int count) {
		LinkedList<Entry<Fraction, Note>> entryList = this.getTailMelody(track, count);

		Fraction posNow = entryList.get(0).getKey().clone();
		Note startNote = track.getMelody().get(posNow);
		GroupNote groupNote = startNote.getGroupNote(); // ミスがなければ取得できるはず！
														// (ミスればぬるぽ)
		Fraction scale = new Fraction(groupNote.getWholeTime(), groupNote.getNoteCount());

		for (int i = 0; i < count; ++i) {
			Fraction pos = entryList.get(i).getKey();
			Note note = entryList.get(i).getValue();

			if (i != 0) {
				// 位置をずらすため、一旦音符を取り除く
				track.getMelody().remove(pos);
				track.getMelody().put(posNow.clone(), note);
			}

			// 長さ変更
			Fraction length = note.getRealLength().clone().mul(scale);
			note.setRealLength(length);
			for (NoteUnit unit : note.getNotes()) {
				unit.setRealLength(unit.getRealLength().mul(scale));
			}

			// 次の位置を計算
			posNow.add(length);
		}
	}

	/**
	 * 装飾を更新する。解析中NoteからNoteUnit用の装飾を取得し、NoteUnitへ移動させる。
	 * また、和音中なら和音開始前の装飾をNoteUnitへコピーする。
	 *
	 * @param trackId
	 *            解析中のTrackID
	 * @param note
	 *            対象の音符
	 */
	private void updateUnitDecoration(String trackId, Note note) {
		ParseDataTrack trackData = this.trackData.get(trackId);
		for (NoteUnit unit : note.getNotes()) {
			unit.getDecorations().addAll(trackData.getChordDecorations());
			unit.getDecorations().addAll(note.getUnitDecorations());
		}
		note.getUnitDecorations().clear();
	}

	/**
	 * TrackIDからメイントラックのIDとVoice Overlayの重なりの順番に分割する。
	 *
	 * @param trackId
	 *            分割するTrackID
	 * @return length=2の配列。[0]: メインのトラックID。[1]:
	 *         Overlayの重なりの順番。メイントラック=0、1つめのOverlay=1、、、となる。
	 */
	public static String[] getTrackIdAndOverlay(String trackId) {
		String[] idOverlay = { "", "" };
		String[] split = trackId.split(Constant.STR_OVERLAY_SEP, 2);

		idOverlay[0] = split[0];
		idOverlay[1] = split.length == 1 ? "0" : split[1];

		return idOverlay;
	}

	/**
	 * 最後からn番目の縦線のある位置を返す。
	 *
	 * @param track
	 *            検索対象のTrack
	 * @param count
	 *            縦線を検索する個数。
	 * @return
	 */
	private Fraction lastBarPos(Track track, int count) {
		Fraction barPos = Fraction.ZERO.clone();

		Iterator<Entry<Fraction, Note>> iter = track.getMelody().descendingMap().entrySet().iterator();
		for (int remain = count; iter.hasNext();) {
			Entry<Fraction, Note> entry = iter.next();
			Fraction pos = entry.getKey();
			Note note = entry.getValue();

			if (note.isNoteAfterBar()) {
				if (--remain == 0) {
					barPos = pos.clone();
					break;
				}
			}
		}

		return barPos;
	}

	/**
	 * Voice Overlayの空番号を検索。空とは、startPosの位置に何もNoteがない状態を言う。
	 *
	 * @param trackId
	 *            検索対象のTrackのID
	 * @param startOverlay
	 *            検索を開始するOverlayの順番
	 * @param startPos
	 *            Overlayの追加音符の開始位置
	 * @return 空のOverlayの順番
	 */
	private int getEmptyOverlayNo(String trackId, int startOverlay, Fraction startPos) {
		int overlay = startOverlay;
		while (true) {
			Track track = this.tune.getTracks().get(trackId + Constant.STR_OVERLAY_SEP + overlay);
			if (track == null) {
				return overlay;
			}

			Note note = track.getMelody().get(startPos);
			if (note == null) {
				return overlay;
			}
			++overlay;
		}
	}

	/**
	 * TrackDataのclefオプションを、Tuneから取得し更新する
	 *
	 * @param trackId
	 *            更新するTrackID
	 */
	private void updateTrackDataClefFromTrack(String trackId) {
		ParseDataTrack trackData = this.trackData.get(trackId);
		for (AInformationField field : this.tune.getTracks().get(trackId).getInformation(
				Constant.INFO_FIELD_KEY_CLEF, true)) {
			IInformationFieldClef clef = (IInformationFieldClef) field;
			if (clef.getClefOctave() != null) {
				trackData.setClefOctave(clef.getClefOctave());
			}
			if (clef.getOctave() != null) {
				trackData.setOctave(clef.getOctave());
			}
			if (clef.getTransposition() != null) {
				trackData.setTransposition(clef.getTransposition());
			}
		}
	}
}
