package abc.score;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * 調(Key)を表すクラス。
 *
 * @author CrapeMyrtle
 */
public class Key implements Cloneable {

	private static final List<String>			KEY_SHARP		= new ArrayList<String>();
	private static final List<String>			KEY_SHARP_ALT	= new ArrayList<String>();
	private static final List<String>			KEY_FLAT		= new ArrayList<String>();
	private static final List<String>			KEY_FLAT_ALT	= new ArrayList<String>();
	private static final Map<String, String>	START_KEY		= new HashMap<String, String>();

	private static final int					STEP_COUNT		= 12;

	/**
	 * どの音階に変化記号を設定しているのか表すMap。<br />
	 * 基本的なModeの変化記号をセット。
	 */
	private Map<String, EKeySignature>			keySignature;
	/**
	 * どの音階に変化記号を設定しているのか表すMap。<br />
	 * K: exp ～ で指定されていた際に使用する
	 */
	private Map<String, EKeySignature>			keySignatureExp;

	/**
	 * modeにおける中央の音 (Major = C、Minor = A、など)
	 */
	private String								base;

	/**
	 * 調におけるmode (Major、Minerなど。最長3文字保有。例: Major = "" or "maj")
	 */
	private String								mode;

	static {
		String sn = "G#,G,F#,F,E,D#,D,C#,C,B,A#,A";
		String sa = "G#,G,F#,E#,E,D#,D,C#,B#,B,A#,A";
		String fn = "Ab,A,Bb,B,C,Db,D,Eb,E,F,Gb,G";
		String fa = "Ab,A,Bb,Cb,C,Db,D,Eb,Fb,F,Gb,G";

		sn += "," + sn;
		sa += "," + sa;
		fn += "," + fn;
		fa += "," + fa;

		KEY_SHARP.addAll(Arrays.asList(sn.split(",")));
		KEY_SHARP_ALT.addAll(Arrays.asList(sa.split(",")));
		KEY_FLAT.addAll(Arrays.asList(fn.split(",")));
		KEY_FLAT_ALT.addAll(Arrays.asList(fa.split(",")));

		START_KEY.put("", "C");
		START_KEY.put("maj", "C");
		START_KEY.put("ion", "C");
		START_KEY.put("m", "A");
		START_KEY.put("min", "A");
		START_KEY.put("aeo", "A");
		START_KEY.put("mix", "G");
		START_KEY.put("dor", "D");
		START_KEY.put("phr", "E");
		START_KEY.put("lyd", "F");
		START_KEY.put("loc", "B");

		sn = null;
		sa = null;
		fn = null;
		fa = null;
	}

	/**
	 * コンストラクタ
	 */
	public Key() {
		this.keySignature = new HashMap<String, EKeySignature>();
		this.keySignatureExp = new HashMap<String, EKeySignature>();
		this.clear();
	}

	/*
	 * (非 Javadoc)
	 *
	 * @see java.lang.Object#clone()
	 */
	public Key clone() {
		Key obj = new Key();
		obj.keySignature.putAll(this.keySignature);
		obj.keySignatureExp.putAll(this.keySignatureExp);
		obj.base = this.base;
		obj.mode = this.mode;
		return obj;
	}

	/*
	 * (非 Javadoc)
	 *
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		StringBuffer sb = new StringBuffer();
		for (String key : this.keySignature.keySet()) {
			EKeySignature sign = (this.keySignatureExp.containsKey(key)
					? this.keySignatureExp.get(key)
					: this.keySignature.get(key));
			sb.append(key + "=" + sign.toString() + ", ");
		}
		return this.getKeyString() + ":{" + sb.substring(0, sb.length() - 2).toString() + "}";
	}

	/**
	 * 調のベース(CmajやFminなどの部分)の文字列表現を取得する。exp以降は考慮しない。
	 *
	 * @return 文字列表現
	 */
	public String toBaseString() {
		StringBuffer sb = new StringBuffer();
		for (String key : this.keySignature.keySet()) {
			EKeySignature sign = this.keySignature.get(key);
			sb.append(key + "=" + sign.toString() + ", ");
		}
		return this.getKeyString() + ":{" + sb.substring(0, sb.length() - 2).toString() + "}";
	}

	/**
	 * 調を表す文字列を返す
	 *
	 * @return 調を表す文字列 (B, Gmなど)
	 */
	public String getKeyString() {
		return this.isNone() ? "null" : this.base + this.mode;
	}

	/**
	 * 調の設定内容をクリアする
	 */
	public void clear() {
		String abc = "ABCDEFG";
		for (int i = 0; i < abc.length(); ++i) {
			this.keySignature.put(abc.substring(i, i + 1), EKeySignature.NATURAL);
		}
		this.keySignatureExp.clear();

		this.base = null;
		this.mode = null;
	}

	/**
	 * 調が設定されているかどうかを返す。
	 *
	 * @return true: 調が設定されていない
	 */
	public boolean isNone() {
		return this.base == null;
	}

	/**
	 * 音階key に 変化記号sign を設定する
	 *
	 * @param key
	 *            音階 (A～G)
	 * @param sign
	 *            変化記号
	 */
	public void setKeySign(String key, EKeySignature sign) {
		String k = key.toUpperCase();
		if ("ABCDEFG".contains(k)) {
			this.keySignatureExp.put(k, sign);
			return;
		}

		throw new RuntimeException("No such key: " + key);
	}

	/**
	 * 調を設定する。 例) B: setKey("B", ""), Gm: setKey("G", "m")
	 *
	 * @param base
	 *            中央の音 (A～G)
	 * @param mode
	 *            設定可能値: "Major" ("", "Ionian"), "Minor" ("m"), "Aeolian",
	 *            "Mixolydian", "Dorian", "Phrygian", "Lydian", "Locrian"
	 */
	public void setKey(String base, String mode) {
		String newBase = base.substring(0, 1).toUpperCase() + base.substring(1).toLowerCase();
		String keyMode = (mode.length() > 3 ? mode.substring(0, 3) : mode).toLowerCase();
		if (!Key.START_KEY.containsKey(keyMode)) {
			throw new RuntimeException("No such mode: " + mode);
		}

		String keyBase = START_KEY.get(keyMode);

		int signCount = 0xff;

		List<List<String>> aryKeys = new ArrayList<List<String>>();
		aryKeys.add(Key.KEY_SHARP);
		aryKeys.add(Key.KEY_SHARP_ALT);
		aryKeys.add(Key.KEY_FLAT);
		aryKeys.add(Key.KEY_FLAT_ALT);

		List<EKeySignature> aryKeySignature = new ArrayList<EKeySignature>();
		aryKeySignature.add(EKeySignature.SHARP);
		aryKeySignature.add(EKeySignature.FLAT);

		List<String> keys = null;
		List<String> keysAlt = null;
		EKeySignature keySign = null;
		for (int i = 1; i >= 0; --i) {
			keys = aryKeys.get(i * 2);
			keysAlt = aryKeys.get(i * 2 + 1);
			keySign = aryKeySignature.get(i);

			int baseStart = keys.indexOf(keyBase);
			int newBaseStart = keys.indexOf(newBase);

			if (newBaseStart < 0) {
				newBaseStart = keysAlt.indexOf(newBase);
			}
			if (newBaseStart < 0) {
				// do nothing
			} else {
				ArrayList<String> usedKey = new ArrayList<String>();
				ArrayList<String> signedKey = new ArrayList<String>();
				usedKey.add(newBase.substring(0, 1));
				if (newBase.length() > 1) {
					signedKey.add(newBase.substring(0, 1));
				}
				for (int j = 1;; j++) {
					String key1 = keys.get(baseStart + j);
					if (key1.equals(keyBase)) {
						break;
					}
					if (key1.length() > 1) {
						continue;
					}

					String key2 = keys.get(newBaseStart + j);
					if (usedKey.contains(key2.substring(0, 1))) {
						key2 = keysAlt.get(newBaseStart + j);
						if (usedKey.contains(key2.substring(0, 1))) {
							signedKey = null;
							break;
						}
					}

					usedKey.add(key2.substring(0, 1));
					if (key2.length() > 1) {
						signedKey.add(key2.substring(0, 1));
					}
				}

				if (signedKey != null && signedKey.size() < signCount) {
					signCount = signedKey.size();
					this.clear();
					for (String key : signedKey) {
						this.keySignature.put(key, keySign);
					}
				}
			}
		}

		if (signCount > 12) {
			throw new RuntimeException("No such base&mode: " + base + mode);
		}

		this.base = newBase;
		this.mode = keyMode;
		return;

	}

	/**
	 * 調は変えずにモード(major、minorなど)を変える
	 *
	 * @param mode
	 *            変えるモード
	 */
	public void convertMode(String mode) {
		if (this.base == null) {
			throw new RuntimeException("Key is not set");
		}

		String keyMode = (mode.length() > 3 ? mode.substring(0, 3) : mode).toLowerCase();
		if (!Key.START_KEY.containsKey(keyMode)) {
			throw new RuntimeException("No such mode: " + mode);
		}

		List<List<String>> aryKeys = new ArrayList<List<String>>();
		aryKeys.add(Key.KEY_SHARP);
		aryKeys.add(Key.KEY_FLAT);

		Map<String, EKeySignature> signExp = new HashMap<String, EKeySignature>(
				this.keySignatureExp);

		String oldBase = this.base;
		String oldMode = this.mode;
		int oldCount = this.getCountBaseSign();
		boolean bConvert = false;
		LOOP: for (int i = 1; i >= 0; --i) {
			List<String> keys1 = aryKeys.get(i);

			int idxOldBase = keys1.indexOf(oldBase);
			int idxOldStart = keys1.indexOf(Key.START_KEY.get(oldMode));
			if (idxOldBase < 0) {
				continue;
			}
			if (idxOldStart < 0) {
				continue;
			}

			int diff = idxOldBase - idxOldStart;
			if (diff < 0) {
				diff += Key.STEP_COUNT;
			}

			for (int d = 0; d < 1; ++d) {
				int j = (i + d) % 2;
				List<String> keys2 = aryKeys.get(j);

				int idxNewStart = keys2.indexOf(Key.START_KEY.get(mode));
				if (idxNewStart < 0) {
					continue;
				}

				int sign = (i == j ? 1 : -1);
				int idxNewBase = idxNewStart + sign * diff;
				if (idxNewBase < 0) {
					idxNewBase += Key.STEP_COUNT;
				}
				if (idxNewBase >= Key.STEP_COUNT) {
					idxNewBase -= Key.STEP_COUNT;
				}

				String newBase = keys2.get(idxNewBase);
				try {
					this.setKey(newBase, mode);
					if (oldCount != this.getCountBaseSign()) {
						continue;
					}
					this.keySignatureExp.putAll(signExp);
					bConvert = true;
					break LOOP;
				} catch (RuntimeException e) {
				}
			}
		}

		if (!bConvert) {
			throw new RuntimeException("failed convert");
		}
	}

	/**
	 * 音階で表される臨時記号を取得する
	 *
	 * @param note
	 *            音階
	 * @return 臨時記号。設定されていなければnullを返す。
	 */
	public EAccidental getAccidental(String note) {
		EKeySignature sign = this.keySignatureExp.get(note.toUpperCase());
		if (sign == null) {
			sign = this.keySignature.get(note.toUpperCase());
		}

		switch (sign) {
		case FLAT:
			return EAccidental.FLAT;
		case SHARP:
			return EAccidental.SHARP;
		case NATURAL:
			return EAccidental.NATURAL;
		default:
			return null;
		}
	}

	/**
	 * 音階で表される臨時記号を取得する(base部分(exp以外)のみ)
	 *
	 * @param note
	 *            音階
	 * @return 臨時記号。設定されていなければnullを返す。
	 */
	public EAccidental getAccidentalBase(String note) {
		EKeySignature sign = this.keySignature.get(note.toUpperCase());
		if (sign == null) {
			return null;
		}

		switch (sign) {
		case FLAT:
			return EAccidental.FLAT;
		case SHARP:
			return EAccidental.SHARP;
		case NATURAL:
			return EAccidental.NATURAL;
		default:
			return null;
		}
	}

	/**
	 * 音階で表される臨時記号を取得する(expで指定された部分のみ)
	 *
	 * @param note
	 *            音階
	 * @return 臨時記号。設定されていなければnullを返す。
	 */
	public EAccidental getAccidentalExp(String note) {
		EKeySignature sign = this.keySignatureExp.get(note.toUpperCase());
		if (sign == null) {
			return null;
		}

		switch (sign) {
		case FLAT:
			return EAccidental.FLAT;
		case SHARP:
			return EAccidental.SHARP;
		case NATURAL:
			return EAccidental.NATURAL;
		default:
			return null;
		}
	}

	/**
	 * 調のベースの臨時記号の合計値を返す。シャープを+1、フラットを-1として合計を計算する。
	 *
	 * @return 合計値
	 */
	private int getCountBaseSign() {
		int count = 0;
		for (EKeySignature sign : this.keySignature.values()) {
			switch (sign) {
			case SHARP:
				++count;
				break;
			case FLAT:
				--count;
				break;
			default:
				break;
			}
		}
		return count;
	}
}
