/**
 *
 */
package abc.score;

import java.util.LinkedList;

import abc.util.Fraction;

/**
 * <p>
 * 主に、楽譜上の音符や休符を表す。
 * </p>
 * <p>
 * 音符・休符以外にも、インラインのinformation fieldや前打音、mfなど音の強さ、スラーなどの情報を持つ。
 * </p>
 *
 * @author CrapeMyrtle
 *
 */
public class Note implements Cloneable {

	private LinkedList<NoteUnit>	notes				= new LinkedList<NoteUnit>();
	private MapInformationField		informationField	= new MapInformationField();
	private Fraction				realLength			= null;
	private String					bar					= "";
	private LinkedList<Integer>		orders				= null;
	private int						repeatStart			= 0;
	private int						repeatEnd			= 0;
	private int						broken				= 0;
	private GroupNote				groupNote			= null;
	private GraceNote				grace				= null;

	private EDynamics				dynamics			= null;
	private Integer					velocity			= null;
	private Integer					volume			= null;
	private LinkedList<EDecoration>	noteDecorations		= new LinkedList<EDecoration>();
	private LinkedList<EDecoration>	unitDecorations		= new LinkedList<EDecoration>();

	private Integer					programId			= null;

	/**
	 * コンストラクタ
	 */
	public Note() {

	}

	/**
	 * コンストラクタ
	 *
	 * @param notes
	 *            同時に鳴らすNoteUnit
	 * @param informationField
	 *            Information Fieldのリスト
	 * @param realLength
	 *            実際の長さ
	 * @param bar
	 *            縦線
	 * @param orders
	 *            リピートで実際に演奏する順番
	 * @param repeatStart
	 *            リピート開始位置か
	 * @param repeatEnd
	 *            リピート終了位置か
	 * @param broken
	 *            Broken Rhythmを表す
	 * @param groupNote
	 *            連音符
	 * @param grace
	 *            Grace Note
	 * @param dynamics
	 *            強弱法
	 * @param velocity
	 *            トラックごとの強さ
	 * @param volume
	 *            Tuneの強さ
	 * @param noteDecorations
	 *            音符に付与される装飾リスト
	 * @param unitDecorations
	 *            音符を構成する単音に付与される装飾リスト
	 * @param programId
	 *            楽器ID (MIDI準拠)
	 */
	public Note(LinkedList<NoteUnit> notes, MapInformationField informationField,
			Fraction realLength, String bar, LinkedList<Integer> orders, int repeatStart,
			int repeatEnd, int broken, GroupNote groupNote, GraceNote grace, EDynamics dynamics,
			Integer velocity, Integer volume, LinkedList<EDecoration> noteDecorations,
			LinkedList<EDecoration> unitDecorations, Integer programId) {
		this.notes = notes;
		this.informationField = informationField;
		this.realLength = realLength;
		this.bar = bar;
		this.orders = orders;
		this.repeatStart = repeatStart;
		this.repeatEnd = repeatEnd;
		this.broken = broken;
		this.groupNote = groupNote;
		this.grace = grace;
		this.dynamics = dynamics;
		this.velocity = velocity;
		this.volume = volume;
		this.noteDecorations = noteDecorations;
		this.unitDecorations = unitDecorations;
		this.programId = programId;
	}

	/*
	 * (非 Javadoc)
	 *
	 * @see java.lang.Object#clone()
	 */
	public Note clone() {
		MapInformationField informationField = new MapInformationField();
		if (this.informationField != null) {
			informationField = this.informationField.clone();
		}

		Fraction realLength = null;
		if (this.realLength != null) {
			realLength = this.realLength.clone();
		}

		LinkedList<NoteUnit> notes = new LinkedList<NoteUnit>();
		if (this.notes != null) {
			for (NoteUnit unit : this.notes) {
				notes.add(unit.clone());
			}
		}

		LinkedList<Integer> orders = null;
		if (this.orders != null) {
			orders = new LinkedList<Integer>();
			for (Integer order : this.orders) {
				orders.add(order);
			}
		}

		GroupNote groupNote = null;
		if (this.groupNote != null) {
			groupNote = this.groupNote.clone();
		}

		GraceNote graceNote = null;
		if (this.grace != null) {
			graceNote = this.grace.clone();
		}

		LinkedList<EDecoration> noteDecorations = new LinkedList<EDecoration>();
		if (this.noteDecorations != null) {
			for (EDecoration deco : this.noteDecorations) {
				noteDecorations.add(deco);
			}
		}

		LinkedList<EDecoration> unitDecorations = new LinkedList<EDecoration>();
		if (this.unitDecorations != null) {
			for (EDecoration deco : this.unitDecorations) {
				unitDecorations.add(deco);
			}
		}

		return new Note(notes, informationField, realLength, this.bar, orders, this.repeatStart,
				this.repeatEnd, this.broken, groupNote, graceNote, dynamics, this.velocity,
				this.volume, noteDecorations, unitDecorations, this.programId);
	}

	/*
	 * (非 Javadoc)
	 *
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		StringBuffer sb = new StringBuffer();

		if (this.groupNote != null) {
			sb.append("(" + this.groupNote.getNoteCount() + ":" + this.groupNote.getWholeTime()
					+ ":" + this.groupNote.getNextNoteCount() + " ");
		}
		if (this.grace != null) {
			sb.append("{" + this.grace.toString() + "}");
		}
		sb.append("[");
		sb.append(bar + " ");
		sb.append("(" + repeatStart + "," + repeatEnd + ") ");
		sb.append(orders + " ");
		sb.append(":" + this.velocity + ":" + this.volume + ":" + this.programId + ":");
		sb.append(this.informationField.toString() + " ");
		sb.append("<");
		sb.append(this.dynamics);
		if (this.noteDecorations != null) {
			for (EDecoration deco : this.noteDecorations) {
				sb.append("," + deco);
			}
		}
		if (this.unitDecorations != null && this.unitDecorations.size() > 0) {
			sb.append("<");
			for (EDecoration deco : this.unitDecorations) {
				sb.append("," + deco);
			}
			sb.append(">");
		}
		sb.append(">");
		sb.append(notes);
		sb.append(realLength);
		sb.append(String.format("%+d", broken));
		sb.append("]");

		return sb.toString();
	}

	/**
	 * 音符のリストを返す
	 *
	 * @return notes 音符リスト
	 */
	public LinkedList<NoteUnit> getNotes() {
		return notes;
	}

	/**
	 * 音符のリストをセットする
	 *
	 * @param notes
	 *            セットする notes
	 */
	public void setNotes(LinkedList<NoteUnit> notes) {
		this.notes = notes;
	}

	/**
	 * 音符のリストに音符を追加する
	 *
	 * @param note
	 *            追加する音符
	 * @return 追加に成功した場合trueを返す
	 */
	public boolean addNote(NoteUnit note) {
		return this.notes.add(note);
	}

	/**
	 * information fieldを取得する
	 *
	 * @return informations information field
	 */
	public MapInformationField getInformationField() {
		return informationField;
	}

	/**
	 * information fieldをセットする
	 *
	 * @param informations
	 *            セットするinformation field
	 */
	public void setInformationField(MapInformationField informations) {
		this.informationField = informations;
	}

	/**
	 * information fieldを追加する
	 *
	 * @param name
	 *            フィールド名
	 * @param field
	 *            フィールドの情報
	 */
	public void addInformation(String name, AInformationField field) {
		LinkedList<AInformationField> list = this.informationField.get(name);
		if (list != null) {
			list.add(field);
		} else {
			list = new LinkedList<AInformationField>();
			list.add(field);
			this.informationField.put(name, list);
		}
	}

	/**
	 * information fieldを取得する
	 *
	 * @param name
	 *            フィールド名
	 * @return 取得できたinformation field。取得できなかった場合nullを返す。
	 */
	public LinkedList<AInformationField> getInformation(String name) {
		return this.informationField.get(name);
	}

	/**
	 * 音符の長さを返す。長さはinformation field(L:)で指定した長さの単位からの倍率で返す。 (ABC譜の長さ表記をそのまま返す)
	 *
	 * @return 音符の長さ。どれか一つでも実際の音符の長さが設定されていたらnullを返す。
	 */
	public Fraction getLength() {
		for (NoteUnit note : this.notes) {
			if (note.isRealLength()) {
				return null;
			}
		}
		return this.notes.size() == 0 ? Fraction.ZERO.clone() : this.notes.get(0).getLength();
	}

	/**
	 * 音符の長さを返す。長さは実際の音符の長さで返す。
	 *
	 * @return 音符の長さ。
	 */
	public Fraction getRealLength() {
		return this.realLength;
	}

	/**
	 * 音符の長さをセットする。長さは実際の音符の長さでセットする。
	 *
	 * @param realLength
	 *            音符の長さ。
	 */
	public void setRealLength(Fraction realLength) {
		this.realLength = realLength;
		;
	}

	/**
	 * 音符の長さを計算する。
	 *
	 * @param unitNoteLength
	 *            計算の時に使用するunit note length(L:フィールドで設定される値)
	 */
	public void calcRealLength(Fraction unitNoteLength) {
		if (this.notes.size() == 0) {
			this.realLength = Fraction.ZERO.clone();
			return;
		}

		NoteUnit unit = this.notes.get(0);
		if (unit == null) {
			this.realLength = Fraction.ZERO.clone();
			return;
		}

		if (unit.isRealLength()) {
			this.realLength = unit.getRealLength().clone();
		} else {
			this.realLength = Fraction.mul(unit.getLength(), unitNoteLength);
		}
	}

	/**
	 * 縦線を取得する
	 *
	 * @return 縦線
	 */
	public String getBar() {
		return this.bar;
	}

	/**
	 * 縦線を追加する
	 *
	 * @param bar
	 *            追加する縦線
	 * @param isContinuing
	 *            縦線が連続しているかどうか。連続している(true)なら、前の縦線と繋げる。
	 */
	public void addBar(String bar, boolean isContinuing) {
		if (this.bar.length() > 0 && !isContinuing) {
			this.bar += " ";
		}
		this.bar += bar;
	}

	/**
	 * 音符の前に縦線があるのかどうかを取得する
	 *
	 * @return true: 音符の前に縦線がある
	 */
	public boolean isNoteAfterBar() {
		return this.bar.length() > 0;
	}

	/**
	 * 繰り返しの順番のリストをセットする
	 *
	 * @param orders
	 *            セットする順番リスト
	 */
	public void setOrders(LinkedList<Integer> orders) {
		this.orders = orders;
	}

	/**
	 * 繰り返しの順番のリストを取得する
	 *
	 * @return 順番リスト
	 */
	public LinkedList<Integer> getOrders() {
		return this.orders;
	}

	/**
	 * 繰り返し開始の個数を返す。 (|:: となっている場合、2を返す)
	 *
	 * @return isRepeatStart 繰り返し個数個数
	 */
	public int getRepeatStart() {
		return repeatStart;
	}

	/**
	 * 繰り返し開始の個数をセットする。 (|:: となっている場合、2をセットする)
	 *
	 * @param repeatStart
	 *            isRepeatStart 繰り返し開始個数
	 */
	public void setRepeatStart(int repeatStart) {
		this.repeatStart = repeatStart;
	}

	/**
	 * 繰り返し終了の個数を返す。 (::| となっている場合、2を返す)
	 *
	 * @return isRepeatEnd 繰り返し終了個数
	 */
	public int getRepeatEnd() {
		return repeatEnd;
	}

	/**
	 * 繰り返し終了の個数をセットする。 (::| となっている場合、2をセットする)
	 *
	 * @param repeatEnd
	 *            isRepeatEnd 繰り返し終了個数
	 */
	public void setRepeatEnd(int repeatEnd) {
		this.repeatEnd = repeatEnd;
	}

/**
	 * Broken Rhythm ("<"や">>")を取得する。正負で">"(正)、"<"(負)を、絶対値で">"や"<"の個数を表す。
	 * <p>例:<dl>
	 * <dt>broken = 1</dt>
	 * <dd>">"</dd>
	 * <dt>broken = -3</dt>
	 * <dd>"<<<"</dd>
	 * </dl></p>
	 *
	 * @return broken Broken Rhythmの値
	 */
	public int getBroken() {
		return broken;
	}

	/**
	 * Broken Rhythm ("<"や">>")を設定する。設定する値の詳細は{@link Note#getBroken()}参照。
	 *
	 * @param broken
	 *            セットするBroken Rhythmの値
	 */
	public void setBroken(int broken) {
		this.broken = broken;
	}

	/**
	 * 連音符を取得する。
	 *
	 * @return 取得する連音符。設定されていなければnullを返す。
	 */
	public GroupNote getGroupNote() {
		return groupNote;
	}

	/**
	 * 連音符をセットする。
	 *
	 * @param groupNote
	 *            セットする連音符。
	 */
	public void setGroupNote(GroupNote groupNote) {
		this.groupNote = groupNote;
	}

	/**
	 * Grace Notesを取得する
	 *
	 * @return Grace Notes
	 */
	public GraceNote getGrace() {
		return grace;
	}

	/**
	 * Grace Notesをセットする
	 *
	 * @param grace
	 *            セットするGrace Notes
	 */
	public void setGrace(GraceNote grace) {
		this.grace = grace;
	}

	/**
	 * 強弱法を取得する
	 *
	 * @return 強弱法。設定されていなければnullを返す。
	 */
	public EDynamics getDynamics() {
		return dynamics;
	}

	/**
	 * 強弱法をセットする
	 *
	 * @param dynamics
	 *            セットする強弱法
	 */
	public void setDynamics(EDynamics dynamics) {
		this.dynamics = dynamics;
	}

	/**
	 * トラックごとの強さを取得する
	 *
	 * @return 強さ
	 */
	public Integer getVelocity() {
		return velocity;
	}

	/**
	 * トラックごとの強さをセットする
	 *
	 * @param velocity
	 *            セットする強さ
	 */
	public void setVelocity(Integer velocity) {
		this.velocity = velocity;
	}

	/**
	 * Tuneの強さを取得する
	 *
	 * @return 強さ
	 */
	public Integer getVolume() {
		return volume;
	}

	/**
	 * Tuneの強さをセットする
	 *
	 * @param volume
	 *            セットする強さ
	 */
	public void setVolume(Integer volume) {
		this.volume = volume;
	}

	/**
	 * 音符に付与される装飾リストを取得する
	 *
	 * @return 装飾リスト
	 */
	public LinkedList<EDecoration> getNoteDecorations() {
		return noteDecorations;
	}

	/**
	 * 音符を構成する音に付与される装飾リストを取得する
	 *
	 * @return 装飾リスト
	 */
	public LinkedList<EDecoration> getUnitDecorations() {
		return unitDecorations;
	}

	/**
	 * 楽器IDを取得する
	 * @return 楽器ID。設定されていなければnullを返す。
	 */
	public Integer getProgramId() {
		return programId;
	}

	/**
	 * 楽器IDをセットする
	 * @param programId 楽器ID。楽器IDをセットしない場合nullを指定する。
	 *            セットする programId
	 */
	public void setProgramId(Integer programId) {
		this.programId = programId;
	}
}
