/**
 *
 */
package abc.score;

import java.util.LinkedList;

import abc.util.Fraction;

/**
 * <p>
 * 単音を表す。
 * </p>
 * <p>
 * 臨時記号(シャープやフラット)、音階、音の長さの情報を持つ。
 * </p>
 *
 * @author CrapeMyrtle
 *
 */
public class NoteUnit implements Comparable<NoteUnit>, Cloneable {

	private EAccidental				acci			= null;
	private String					note			= "c";
	private int						octave			= 0;
	private Fraction				length			= new Fraction(1);
	private boolean					isRealLength	= false;
	private boolean					isWithTie		= false;

	private LinkedList<EDecoration>	decorations		= new LinkedList<EDecoration>();

	/**
	 * コンストラクタ
	 */
	public NoteUnit() {
	}

	/**
	 * コンストラクタ
	 *
	 * @param note
	 *            音符の基本音階(C-GAB,c-gab)
	 * @param acci
	 *            臨時記号
	 * @param octave
	 *            オクターブ
	 * @param length
	 *            長さ
	 * @param isRealLength
	 *            実際の音符の長さか、ABC譜上の長さか
	 * @param isWithTie
	 *            タイ記号があるか
	 */
	public NoteUnit(String note, EAccidental acci, int octave, Fraction length,
			boolean isRealLength, boolean isWithTie,
			LinkedList<EDecoration> decorations) {
		this.note = note;
		this.acci = acci;
		this.octave = octave;
		this.length = length;
		this.isRealLength = isRealLength;
		this.isWithTie = isWithTie;
		this.decorations = decorations;
	}

	/*
	 * (非 Javadoc)
	 *
	 * @see java.lang.Object#clone()
	 */
	public NoteUnit clone() {
		LinkedList<EDecoration> decorations = null;
		if (this.decorations != null) {
			decorations = new LinkedList<EDecoration>();
			for (EDecoration deco : this.decorations) {
				decorations.add(deco);
			}
		}

		return new NoteUnit(this.note, this.acci, this.octave, this.length.clone(),
				this.isRealLength, this.isWithTie, decorations);
	}

	/*
	 * (非 Javadoc)
	 *
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		StringBuffer sb = new StringBuffer();

		sb.append("<");
		if (this.decorations != null) {
			for (EDecoration deco : this.decorations) {
				sb.append(deco + ",");
			}
		}
		sb.append(">");
		sb.append(acci + ",");
		sb.append(note);
		sb.append(String.format("%+d,", octave));
		sb.append(length);
		sb.append(isRealLength ? "(R)" : "(A)");
		if (this.isWithTie) {
			sb.append("-");
		}

		return sb.toString();
	}

	/*
	 * (非 Javadoc)
	 *
	 * @see java.lang.Comparable#compareTo(java.lang.Object)
	 */
	public int compareTo(NoteUnit unit) {
		String stepString = "C_D_EF_G_A_B";
		stepString = stepString + stepString.toLowerCase();

		int stepA = this.calcStep();
		int stepB = unit.calcStep();

		return stepA == stepB ? 0 : (stepA > stepB ? 1 : -1);
	}

	/**
	 * 音階の値を数値で返す。比較用に使用するためだけに使用可能。
	 *
	 * @return 音階の値
	 */
	private int calcStep() {
		String stepString = "C_D_EF_G_A_Bc_d_ef_g_a_b";
		int step = stepString.indexOf(this.note);
		step += 12 * this.octave;
		if (this.acci != null) {
			switch (this.acci) {
			case DOUBLE_FLAT:
				step -= 2;
				break;
			case FLAT:
				step -= 1;
				break;
			case SHARP:
				step += 1;
				break;
			case DOUBLE_SHARP:
				step += 2;
				break;
			default:
				break;
			}
		}

		return step;
	}

	/**
	 * 臨時記号を返す
	 *
	 * @return acci 臨時記号
	 */
	public EAccidental getAcci() {
		return acci;
	}

	/**
	 * 臨時記号をセットする
	 *
	 * @param acci
	 *            セットする acci
	 */
	public void setAcci(EAccidental acci) {
		this.acci = acci;
	}

	/**
	 * 音階(C～Bc～b)を返す
	 *
	 * @return note 音階
	 */
	public String getNote() {
		return note;
	}

	/**
	 * 音階(C～Bc～b)をセットする
	 *
	 * @param note
	 *            セットする note
	 */
	public void setNote(String note) {
		this.note = note;
	}

	/**
	 * オクターブ情報を返す。+でオクターブ高く、-でオクターブ低くなる。
	 *
	 * @return octave オクターブ情報
	 */
	public int getOctave() {
		return octave;
	}

	/**
	 * オクターブ情報をセットする
	 *
	 * @param octave
	 *            セットする octave
	 */
	public void setOctave(int octave) {
		this.octave = octave;
	}

	/**
	 * 長さ(L:フィールドに対する倍率)を返す
	 *
	 * @return length 長さ。実際の音符の長さが設定されていればnullを返す。
	 */
	public Fraction getLength() {
		if (this.isRealLength) {
			return null;
		}
		return length;
	}

	/**
	 * 長さ(L:フィールドに対する倍率)をセットする
	 *
	 * @param length
	 *            セットする length
	 */
	public void setLength(Fraction length) {
		this.isRealLength = false;
		this.length = length;
	}

	/**
	 * 長さ(実際の音符の長さ)を返す。
	 *
	 * @return length 長さ。実際の音符の長さが設定されていなければnullを返す。
	 */
	public Fraction getRealLength() {
		if (this.isRealLength) {
			return length;
		}
		return null;
	}

	/**
	 * 長さをセットする
	 *
	 * @param length
	 *            セットする length
	 */
	public void setRealLength(Fraction length) {
		this.isRealLength = true;
		this.length = length;
	}

	/**
	 * 実際の音符の長さかどうかを取得する。
	 *
	 * @return isRealLength true: 実際の音符の長さ。false: フィールド(L:)に対する倍率。
	 */
	public boolean isRealLength() {
		return isRealLength;
	}

	/**
	 * タイ("-")を後ろに持っているかどうかを返す。
	 *
	 * @return タイを持っているかどうか。true: タイを持っている。ABC譜で"a-a"となっていたら、1つめの"a"がtrueとなる。
	 */
	public boolean isWithTie() {
		return isWithTie;
	}

	/**
	 * タイ("-")を後ろに持っているかどうかをセットする。
	 *
	 * @param isWithTie
	 *            タイを持っているかどうか。true: タイを持っている。参考: {@link NoteUnit#isWithTie()}。
	 */
	public void setWithTie(boolean isWithTie) {
		this.isWithTie = isWithTie;
	}

	/**
	 * 装飾リストを取得する
	 *
	 * @return 装飾リスト
	 */
	public LinkedList<EDecoration> getDecorations() {
		return decorations;
	}
}
