/**
 *
 */
package abc.score;

import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedList;

/**
 * <p>
 * ABC譜を表す。
 * <p>
 * <p>
 * ABC譜は以下で構成される
 * <ul>
 * <li>{@link Score}
 * <ul>
 * <li>File Header ({@link MapInformationField})</li>
 * <li>{@link Tune}* (フィールド X: で区切られる)
 * <ul>
 * <li>Tune Header ({@link MapInformationField})</li>
 * <li>{@link Track}* (フィールド V: で区切られる)
 * <ul>
 * <li>{@link Note}*</li>
 * </ul>
 * </li>
 * </ul>
 * </li>
 * </ul>
 * </li>
 * </ul>
 * </p>
 *
 * @author CrapeMyrtle
 *
 */
public class Score {

	private MapInformationField			fileHeader		= new MapInformationField();
	private LinkedHashMap<String, Tune>	tunes			= new LinkedHashMap<String, Tune>();
	private String						activeTuneId	= null;

	/**
	 * file headerを取得する
	 *
	 * @return fileHeader file header
	 */
	public MapInformationField getFileHeader() {
		return fileHeader;
	}

	/**
	 * file headerをセットする
	 *
	 * @param fileHeader
	 *            セットする file header
	 */
	public void setFileHeader(MapInformationField fileHeader) {
		this.fileHeader = fileHeader;
	}

	/**
	 * Tuneデータを取得する
	 *
	 * @return tunes Tuneデータ
	 */
	public LinkedHashMap<String, Tune> getTunes() {
		return tunes;
	}

	/**
	 * Tuneデータをセットする
	 *
	 * @param tunes
	 *            セットする Tuneデータ
	 */
	public void setTunes(LinkedHashMap<String, Tune> tunes) {
		this.tunes = tunes;
		for (Tune tune : tunes.values()) {
			tune.setParent(this);
		}
	}

	/**
	 * Tuneデータを追加する
	 *
	 * @param id
	 *            追加するTuneデータのID
	 * @param tune
	 *            追加するTune
	 */
	public void addTune(String id, Tune tune) {
		this.tunes.put(id, tune);
		tune.setParent(this);
	}

	/**
	 * アクティブ(解析中)なTuneのIDを返す
	 *
	 * @return activeTune アクティブなTuneのID
	 */
	public String getActiveTuneId() {
		return this.activeTuneId;
	}

	/**
	 * アクティブ(解析中)なTuneのIDをセットする
	 *
	 * @param activeTune
	 *            セットするアクティブTuneのID
	 */
	public void setActiveTuneId(String activeTune) {
		this.activeTuneId = activeTune;
	}

	/**
	 * アクティブ(解析中)なTuneを返す
	 *
	 * @return activeTune アクティブなTune
	 */
	public Tune getActiveTune() {
		return this.tunes.get(this.activeTuneId);
	}

	/**
	 * file headerにinformation fieldを追加する
	 *
	 * @param name
	 *            フィールド名
	 * @param field
	 *            フィールドの情報
	 */
	public void addFileHeader(String name, AInformationField field) {
		LinkedList<AInformationField> list = this.fileHeader.get(name);
		if (list != null) {
			list.add(field);
		} else {
			list = new LinkedList<AInformationField>();
			list.add(field);
			this.fileHeader.put(name, list);
		}
	}

	/*
	 * (非 Javadoc)
	 *
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		StringBuffer sb = new StringBuffer();

		sb.append("File Headers:\n");
		for (Iterator<String> it = this.fileHeader.keySet().iterator(); it.hasNext();) {
			String key = it.next();
			sb.append("  " + key + ": " + this.fileHeader.get(key).toString()
					+ "\n");
		}

		sb.append("Tunes:\n");
		for (Iterator<String> it = this.tunes.keySet().iterator(); it.hasNext();) {
			String key = it.next();
			sb.append("  " + key + ":\n");
			sb.append(this.tunes.get(key).toString());
		}

		return sb.toString();
	}

	/**
	 * information fieldから情報のリストを取得する。
	 *
	 * @param name
	 *            取得する情報の名前。information fieldのキー(KやLなど)を指定する。
	 * @param isSize0
	 *            リストのサイズが0でもリストを取得するかどうか。 true: サイズが0でもリストを取得する。
	 * @return 取得した情報リスト。なければnullを返す。
	 */
	public LinkedList<AInformationField> getInformation(String name,
			boolean isSize0) {
		LinkedList<AInformationField> list = this.fileHeader.get(name);
		if (list != null && (list.size() > 0 || isSize0)) {
			return list;
		}

		if (list == null && isSize0) {
			return new LinkedList<AInformationField>();
		}

		return null;
	}

	/**
	 * information fieldから情報のリストを取得する。 サイズが0の情報リストは取得しない。
	 *
	 * @param name
	 *            取得する情報の名前。information fieldのキー(KやLなど)を指定する。
	 * @return 取得した情報リスト。なければnullを返す。
	 */
	public LinkedList<AInformationField> getInformation(String name) {
		return this.getInformation(name, false);
	}
}
