/**
 *
 */
package abc.score;

import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map.Entry;
import java.util.TreeMap;

import abc.parser.Constant;
import abc.util.Fraction;

/**
 * トラックを表す。
 *
 * @author CrapeMyrtle
 * @see Score
 *
 */
public class Track {

	/** トラックごとの楽曲情報(L: Q: など)を保持 */
	private MapInformationField				informationField	= new MapInformationField();
	/** 小節ごとの臨時記号を保持 */
	private HashMap<String, EAccidental>	barAccidental		= new HashMap<String, EAccidental>();

	/** 音符の並び。keyのFractionで音を鳴らす位置、valのNoteで鳴らす音と長さを表現 */
	private TreeMap<Fraction, Note>			melody				= new TreeMap<Fraction, Note>();

	/** 親の{@link Tune} */
	private Tune							parent;

	/** TrackID */
	private String							trackId = InformationFieldVoice.DEFAULT_ID;

	/**
	 * コンストラクタ
	 */
	public Track() {
		this.setEmptyNoteAt(Fraction.ZERO);

		// U:の定義を予め作成しておく
		LinkedList<AInformationField> list = new LinkedList<AInformationField>();
		list.add(new InformationFieldDefine("~", "!roll!"));
		list.add(new InformationFieldDefine("H", "!fermata!"));
		list.add(new InformationFieldDefine("L", "!accent!"));
		list.add(new InformationFieldDefine("M", "!lowermordent!"));
		list.add(new InformationFieldDefine("O", "!coda!"));
		list.add(new InformationFieldDefine("P", "!uppermordent!"));
		list.add(new InformationFieldDefine("S", "!segno!"));
		list.add(new InformationFieldDefine("T", "!trill!"));
		list.add(new InformationFieldDefine("u", "!upbow!"));
		list.add(new InformationFieldDefine("v", "!downbow!"));
		this.informationField.put("U", list);
	}

	/**
	 * 解析中など一時的に使用するデータをクリアする
	 */
	public void clearTemporary() {
		this.informationField.clear();
		this.barAccidental.clear();
	}

	/**
	 * トラックIDを取得する
	 * @return トラックID
	 */
	public String getTrackId() {
		return this.trackId;
	}

	/**
	 * トラックIDをセットする
	 * @param trackId セットするトラックID
	 */
	public void setTrackId(String trackId) {
		this.trackId = trackId;
	}

	/**
	 * トラックごとの楽曲情報(information field)を取得する
	 *
	 * @return field トラックごとの楽曲情報(information field)
	 */
	public MapInformationField getInformationField() {
		return informationField;
	}

	/**
	 * トラックごとの楽曲情報(information field)をセットする
	 *
	 * @param field
	 *            セットするトラックごとの楽曲情報(information field)
	 */
	public void setInformationField(MapInformationField field) {
		this.informationField = field;
	}

	/**
	 * 音符の並びを取得する
	 *
	 * @return melody 音符の並び
	 */
	public TreeMap<Fraction, Note> getMelody() {
		return melody;
	}

	/**
	 * 音符の並び
	 *
	 * @param melody
	 *            セットする音符の並び
	 */
	public void setMelody(TreeMap<Fraction, Note> melody) {
		this.melody = melody;
	}

	/**
	 * 親を取得する
	 *
	 * @return parent 親
	 */
	public Tune getParent() {
		return parent;
	}

	/**
	 * 親をセットする
	 *
	 * @param parent
	 *            セットする親
	 */
	public void setParent(Tune parent) {
		this.parent = parent;
	}

	/**
	 * information fieldから情報のリストを取得する。情報は{@link Tune} -> {@link Track}
	 * という順番で前からリストに格納する。
	 *
	 * @param name
	 *            取得する情報の名前。information fieldのキー(KやLなど)を指定する。
	 * @param isSize0
	 *            リストのサイズが0でもリストを取得するかどうか。 true: サイズが0でもリストを取得する。
	 * @return 取得した情報リスト。なければnullを返す。
	 */
	public LinkedList<AInformationField> getInformation(String name, boolean isSize0) {
		LinkedList<AInformationField> list = new LinkedList<AInformationField>();

		list.addAll(this.parent.getInformation(name, true));

		if (name.equals(Constant.INFO_FIELD_KEY_CLEF)) {
			// clefオプションはトラック毎の物を親のTuneが保持していることがある
			list.addAll(this.parent.getTrackInformation(name, this.trackId, true));
		}

		LinkedList<AInformationField> listTrack = null;

		listTrack = this.informationField.get(name);
		if (listTrack != null) {
			list.addAll(listTrack);
		}

		return (!isSize0 && list.size() == 0) ? null : list;
	}

	/**
	 * information fieldから情報のリストを取得する。情報は{@link Tune} -> {@link Track}->
	 * 小節で保持する情報(内部保有) という順番で前からリストに格納する。 サイズが0の情報リストは取得しない。
	 *
	 * @param name
	 *            取得する情報の名前。information fieldのキー(KやLなど)を指定する。
	 * @return 取得した情報リスト。なければnullを返す。
	 */
	public LinkedList<AInformationField> getInformation(String name) {
		return this.getInformation(name, false);
	}

	/**
	 * information fieldを追加する
	 *
	 * @param name
	 *            フィールド名
	 * @param field
	 *            フィールドの情報
	 */
	public void addInformation(String name, AInformationField field) {
		LinkedList<AInformationField> list = this.informationField.get(name);
		if (list == null) {
			list = new LinkedList<AInformationField>();
			this.informationField.put(name, list);
		}
		list.add(field);
	}

	/**
	 * unit note length (L:フィールドの値)を取得する
	 *
	 * @return unit note length。値が設定されていなければnullを返す。
	 */
	public Fraction getUnitNoteLength() {
		Fraction unitNoteLength = null;
		LinkedList<AInformationField> fieldL = this.getInformation("L", true);
		if (fieldL.size() > 0) {
			unitNoteLength = ((InformationFieldLength) fieldL.getLast()).getLength();
		}

		return unitNoteLength;
	}

	/**
	 * meter length (M:フィールドの値)を取得する
	 *
	 * @return meter length。値が設定されていなければnullを返す。
	 */
	public Fraction getMeterLength() {
		Fraction meterLength = null;
		LinkedList<AInformationField> fieldM = this.getInformation("M", true);
		if (fieldM.size() > 0) {
			meterLength = ((InformationFieldMeter) fieldM.getLast()).getMeter().getLength();
		}

		return meterLength;
	}

	/**
	 * 臨時記号情報をクリアする
	 */
	public void clearAccidental() {
		this.barAccidental.clear();
	}

	/**
	 * 臨時記号をセットする
	 *
	 * @param step
	 *            この音階の臨時記号をセットする
	 * @param acci
	 *            セットする臨時記号
	 */
	public void putAccidental(String step, EAccidental acci) {
		this.barAccidental.put(step, acci);
	}

	/**
	 * 臨時記号を取得する
	 *
	 * @param step
	 *            この音階の臨時記号を取得する
	 * @return 臨時記号
	 */
	public EAccidental getAccidental(String step) {
		return this.barAccidental.get(step);
	}

	/*
	 * (非 Javadoc)
	 *
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		StringBuffer sb = new StringBuffer();

		for (Iterator<Entry<Fraction, Note>> iter = this.melody.entrySet().iterator(); iter.hasNext();) {
			sb.append("        ");
			Entry<Fraction, Note> entry = iter.next();
			Fraction pos = entry.getKey();
			sb.append(String.format(
					"(%5d/%5d)%s\n", pos.getNumerator(), pos.getDenominator(), entry.getValue()));
		}
		sb.append("\n");

		return sb.toString();
	}

	public void setEmptyNoteAt(Fraction pos) {
		Note note = new Note();
		note.setRealLength(Fraction.ZERO.clone());
		this.melody.put(pos.clone(), note);
	}

	/**
	 * トラックを合成する。合成をしたら自身を更新する。 合成するのは音符情報のみで、Information
	 * 繰り返し情報、縦線情報などは合成しない。 そのため、合成元と合成先で合わせておくこと。
	 * また合成することで、前の音符の位置 + 前の音符の長さ = 次の音符の位置とならなくなるため注意をすること。
	 *
	 * @param track
	 *            合成するTrack
	 */
	public void combineWith(Track track) {
		for (Iterator<Entry<Fraction, Note>> iter = track.getMelody().entrySet().iterator(); iter.hasNext();) {
			Entry<Fraction, Note> entry = iter.next();
			Fraction pos = entry.getKey();
			Note note = entry.getValue();

			Note thisNote = this.melody.get(pos);
			if (thisNote == null) {
				this.melody.put(pos.clone(), note.clone());
			} else {
				for (NoteUnit unit : note.getNotes()) {
					thisNote.addNote(unit.clone());
				}
			}
		}
	}

	/**
	 * Trackから繰り返し情報をコピーする。 元からある繰り返し情報は消さない。
	 *
	 * @param track
	 *            コピー元のTrack
	 */
	public void copyRepeatFrom(Track track) {
		for (Iterator<Entry<Fraction, Note>> iter = track.getMelody().entrySet().iterator(); iter.hasNext();) {
			Entry<Fraction, Note> entry = iter.next();
			Fraction pos = entry.getKey();
			Note srcNote = entry.getValue();

			if (srcNote.getRepeatStart() > 0 || srcNote.getRepeatEnd() > 0
					|| (srcNote.getOrders() != null && srcNote.getOrders().size() > 0)) {
				Note destNote = this.melody.get(pos);
				if (destNote == null) {
					destNote = new Note();
					destNote.setRealLength(Fraction.ZERO.clone());
					this.melody.put(pos.clone(), destNote);
				}
				destNote.setRepeatStart(srcNote.getRepeatStart());
				destNote.setRepeatEnd(srcNote.getRepeatEnd());
				destNote.setOrders(srcNote.getOrders()); // TODO cloneしないとダメか？
			}
		}
	}

	/**
	 * TrackからInformation Fieldをコピーする。
	 *
	 * @param track
	 *            コピー元のTrack
	 */
	public void copyInformationFrom(Track track) {
		for (Iterator<Entry<Fraction, Note>> iter = track.getMelody().entrySet().iterator(); iter.hasNext();) {
			Entry<Fraction, Note> entry = iter.next();
			Fraction pos = entry.getKey();
			Note srcNote = entry.getValue();

			if (!this.melody.containsKey(pos)) {
				this.setEmptyNoteAt(pos);
			}
			Note destNote = this.melody.get(pos);

			// information field
			MapInformationField map = srcNote.getInformationField();
			for (Iterator<Entry<String, LinkedList<AInformationField>>> it = map.entrySet().iterator(); it.hasNext(); ) {
				Entry<String, LinkedList<AInformationField>> e = it.next();
				String key = e.getKey();
				if (destNote.getInformationField().get(key) != null) {
					destNote.getInformationField().get(key).addAll(e.getValue());
				} else {
					destNote.getInformationField().put(key, e.getValue());
				}
			}

			if (srcNote.getDynamics() != null) {
				destNote.setDynamics(srcNote.getDynamics());
			}
			destNote.getNoteDecorations().addAll(srcNote.getNoteDecorations());

			if (srcNote.getVolume() != null) {
				destNote.setVolume(srcNote.getVolume());
			}

			if (srcNote.getVelocity() != null) {
				destNote.setVelocity(srcNote.getVelocity());
			}
		}
	}
}
