/**
 *
 */
package abc.score;

import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedList;

import abc.parser.Constant;

/**
 * tune (X:で区切られた一曲) を表す。
 *
 * @author CrapeMyrtle
 * @see Score
 */
public class Tune {

	private MapInformationField				tuneHeader	= new MapInformationField();
	private LinkedHashMap<String, Track>	tracks		= new LinkedHashMap<String, Track>();

	private Score							parent;

	/**
	 * コンストラクタ。予めデフォルトトラックIDを持つTrackを用意しておく。
	 */
	public Tune() {
		this.addTrack(InformationFieldVoice.DEFAULT_ID, new Track());
	}

	/**
	 * tune headerを取得する
	 *
	 * @return tuneHeader tune header
	 */
	public MapInformationField getTuneHeader() {
		return tuneHeader;
	}

	/**
	 * tune headerをセットする
	 *
	 * @param tuneHeader
	 *            セットするtune header
	 */
	public void setTuneHeader(MapInformationField tuneHeader) {
		this.tuneHeader = tuneHeader;
	}

	/**
	 * トラックデータを取得する
	 *
	 * @return tracks トラックデータ
	 */
	public LinkedHashMap<String, Track> getTracks() {
		return tracks;
	}

	/**
	 * トラックデータをセットする
	 *
	 * @param tracks
	 *            セットする トラックデータ
	 */
	public void setTracks(LinkedHashMap<String, Track> tracks) {
		this.tracks = tracks;
		for (Track track : tracks.values()) {
			track.setParent(this);
		}
	}

	/**
	 * Trackデータを追加する。
	 *
	 * @param id
	 *            追加するTrackデータのID
	 * @param track
	 *            追加するTrackデータ
	 */
	public void addTrack(String id, Track track) {
		this.tracks.put(id, track);
		track.setParent(this);
	}

	/**
	 * @return parent
	 */
	public Score getParent() {
		return parent;
	}

	/**
	 * @param parent
	 *            セットする parent
	 */
	public void setParent(Score parent) {
		this.parent = parent;
	}

	/*
	 * (非 Javadoc)
	 *
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		StringBuffer sb = new StringBuffer();

		sb.append("    Tune Headers:\n");

		for (Iterator<String> it = this.tuneHeader.keySet().iterator(); it.hasNext();) {
			String key = it.next();
			sb.append("      " + key + ": " + this.tuneHeader.get(key) + "\n");
		}

		sb.append("    Tracks:\n");
		for (Iterator<String> it = this.tracks.keySet().iterator(); it.hasNext();) {
			String key = it.next();
			sb.append("      " + key + ":\n");
			sb.append(this.tracks.get(key).toString());
		}

		return sb.toString();
	}

	/**
	 * information fieldから情報のリストを取得する。 情報は{@link Score} -> {@link Tune}
	 * という順番で前からリストに格納する。
	 *
	 * @param name
	 *            取得する情報の名前。information fieldのキー(KやLなど)を指定する。
	 * @param isSize0
	 *            リストのサイズが0でもリストを取得するかどうか。 true: サイズが0でもリストを取得する。
	 * @return 取得した情報リスト。なければnullを返す。
	 */
	public LinkedList<AInformationField> getInformation(String name, boolean isSize0) {
		LinkedList<AInformationField> list = new LinkedList<AInformationField>();

		list.addAll(this.parent.getInformation(name, true));

		LinkedList<AInformationField> listTune = this.tuneHeader.get(name);
		if (listTune != null) {
			list.addAll(listTune);
		}

		return (!isSize0 && list.size() == 0) ? null : list;
	}

	/**
	 * information fieldから情報のリストを取得する。 情報は{@link Track} -> {@link Tune} ->
	 * {@link Score}という順番で 一番最初に見つかったものを取得する。 サイズが0の情報リストは取得しない。
	 *
	 * @param name
	 *            取得する情報の名前。information fieldのキー(KやLなど)を指定する。
	 * @return 取得した情報リスト。なければnullを返す。
	 */
	public LinkedList<AInformationField> getInformation(String name) {
		return this.getInformation(name, false);
	}

	/**
	 * information fieldからTrack毎の情報リストを取得する。
	 *
	 * @param name
	 *            取得する情報の名前
	 * @param trackId
	 *            取得するTrackのID
	 * @return 情報リスト。なければnullを返す。
	 */
	public LinkedList<AInformationField> getTrackInformation(String name, String trackId) {
		return this.getInformation(name + Constant.INFO_FIELD_KEY_SEP + trackId);
	}

	/**
	 * information fieldからTrack毎の情報リストを取得する。
	 *
	 * @param name
	 *            取得する情報の名前
	 * @param trackId
	 *            取得するTrackのID
	 * @param isSize0
	 *            リストのサイズが0でもリストを取得するかどうか。 true: サイズが0でもリストを取得する。
	 * @return 情報リスト。なければnullを返す。
	 */
	public LinkedList<AInformationField> getTrackInformation(String name, String trackId,
			boolean isSize0) {
		return this.getInformation(name + Constant.INFO_FIELD_KEY_SEP + trackId, isSize0);
	}

	/**
	 * tune headerにinformation fieldを追加する
	 *
	 * @param name
	 *            フィールド名
	 * @param field
	 *            フィールドの情報
	 */
	public void addInformation(String name, AInformationField field) {
		LinkedList<AInformationField> list = this.tuneHeader.get(name);
		if (list == null) {
			list = new LinkedList<AInformationField>();
			this.tuneHeader.put(name, list);
		}
		list.add(field);

		// clefはTuneヘッダにありながらTrack毎に影響を及ぼす
		if (field instanceof InformationFieldKey) {
			// K:にあるclefは全トラックに影響
			String key = Constant.INFO_FIELD_KEY_CLEF;
			list = this.tuneHeader.get(key);
			if (list == null) {
				list = new LinkedList<AInformationField>();
				this.tuneHeader.put(key, list);
			}
			list.add(((InformationFieldKey) field).getClef());
		}
		if (field instanceof InformationFieldVoice) {
			// V:にあるclefはトラックごとに影響
			InformationFieldVoice infVoice = (InformationFieldVoice) field;
			String key = Constant.INFO_FIELD_KEY_CLEF + Constant.INFO_FIELD_KEY_SEP
					+ infVoice.getId();
			list = this.tuneHeader.get(key);
			if (list == null) {
				list = new LinkedList<AInformationField>();
				this.tuneHeader.put(key, list);
			}
			list.add(infVoice.getClef());
		}
	}
}
