package abc.translator;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map.Entry;
import java.util.TreeMap;

import util.Option;
import util.StringUtil;
import abc.parser.ParserTuneBody;
import abc.score.AInformationField;
import abc.score.EAccidental;
import abc.score.EDecoration;
import abc.score.GraceNote;
import abc.score.InformationFieldKey;
import abc.score.InformationFieldLength;
import abc.score.InformationFieldMeter;
import abc.score.InformationFieldTempo;
import abc.score.InformationFieldTitle;
import abc.score.Key;
import abc.score.Note;
import abc.score.NoteUnit;
import abc.score.Score;
import abc.score.Tempo;
import abc.score.Track;
import abc.score.Tune;
import abc.util.Fraction;
import abc.util.TrackID;
import abc.util.TrackSrcDstID;

/**
 * ABCフォーマットからMoE-ABCフォーマットへの変換を行う。
 *
 * @author CrapeMyrtle
 *
 */
public class Translator {

	private Key						key;
	private Tempo					tempo;
	private Fraction				unitNoteLength;
	private Fraction				meterLength;
	private Fraction				barPos;
	private Integer					programId;

	private Option					opt;

	private static final int		DEFAULT_VELOCITY	= 100;
	// private static final int MIN_VELOCITY = 0;
	private static final int		DEFAULT_VOLUME		= 100;
	private static final int		MIN_VOLUME			= 0;
	private static final int		DEFAULT_TEMPO		= 120;
	private static final int		MIN_TEMPO			= 32;
	private static final Fraction	TEMPO_BASE			= new Fraction(1, 4);
	private static final Fraction	DEFAULT_LENGTH		= new Fraction(1, 4);
	private static final Fraction	MIN_LENGTH			= new Fraction(1, 64);
	private static final Fraction	MAX_LENGTH			= new Fraction(16, 1);

	public Translator(Option opt) {
		this.opt = opt;
	}

	/**
	 * 文字列を出力する(改行なし)
	 *
	 * @param ojb
	 *            出力するオブジェクト
	 * @throws IOException
	 */
	private void print(Object obj) throws IOException {
		this.opt.getOut().print(obj);
	}

	/**
	 * 文字列を出力する(改行あり)
	 *
	 * @param ojb
	 *            出力するオブジェクト
	 * @throws IOException
	 */
	private void println(Object obj) throws IOException {
		this.opt.getOut().println(obj);
	}

	/**
	 * リピート記号を展開し、繰り返し情報を取り除く
	 * <p>
	 * <em>注)</em> 最適化({@link Translator#optimize(Score)})されているものとする。
	 * </p>
	 *
	 * @param score
	 *            展開するScore
	 */
	public void expandRepeat(Score score) {
		for (Tune tune : score.getTunes().values()) {
			for (Track track : tune.getTracks().values()) {
				this.expandRepeatTrack(track);
			}
		}
	}

	/**
	 * Trackのリピート記号を展開し、繰り返し情報を取り除く。
	 * <p>
	 * <em>注)</em> Trackは最適化({@link Translator#optimizeTrack(Track)})されているものとする。
	 * </p>
	 *
	 * @param track
	 *            展開するTrack
	 */
	public void expandRepeatTrack(Track track) {
		int preStart = -1;
		int preEnd = 0;
		int repeatCount = 1;
		Fraction nextPos = Fraction.ZERO.clone();

		LinkedList<Entry<Fraction, Note>> list = new LinkedList<Entry<Fraction, Note>>(
				track.getMelody().entrySet());
		TreeMap<Fraction, Note> melody = new TreeMap<>();
		boolean isInOrder = false;

		for (int idx = 0; idx < list.size();) {
			Entry<Fraction, Note> entry = list.get(idx);
			Note note = entry.getValue();

			// 次の処理が決定しているかどうか
			boolean isProc = false;

			if (note.getRepeatEnd() > 0) {
				// 繰り返しの終了(:|)

				if (!isProc) {
					if (repeatCount <= note.getRepeatEnd() || isInOrder) {
						isProc = true;
						++repeatCount;
						idx = (preStart >= 0 ? preStart : preEnd);
					}
				}

				if (!isProc) {
					repeatCount = 1;
					preStart = idx;
					preEnd = idx;
				}
			}
			entry = list.get(idx);
			note = entry.getValue();

			if (note.getRepeatStart() > 0) {
				// 繰り返しの開始(|:)

				preStart = idx;
				isInOrder = false;

				if (!isProc) {
					repeatCount = 1;
				}
			}

			if (note.getBar().indexOf("|]") >= 0) {
				preStart = idx;
				isInOrder = false;
				if (!isProc) {
					repeatCount = 1;
				}
			}

			if (note.getOrders() != null) {
				// 繰り返し開始順([1、[2、など)

				isInOrder = true;

				// 次の開始箇所を検索
				// なければ最後の繰り返し終了(:|)を探す
				// それでもなければ演奏終了
				int lastEnd = idx; // 見つかった最後の繰り返し終了(:|)
				boolean isFound = false;
				for (int nextIdx = idx; nextIdx < list.size(); ++nextIdx) {
					Note nextNote = list.get(nextIdx).getValue();

					if (nextNote.getOrders() != null && nextNote.getOrders().contains(repeatCount)) {
						isFound = true;
						isProc = true;
						idx = nextIdx;
						break;
					}

					if (nextNote.getRepeatEnd() > 0) {
						lastEnd = nextIdx;
						if (nextNote.getOrders() == null) {
							// 繰り返し終了(:|)の次に、演奏順が設定されていなければ
							// 繰り返し終了とみなす
							break;
						}
					}

					if (nextNote.getRepeatStart() > 0) {
						// 繰り返し開始があれば、次の繰り返しになっているとして検索終了
						break;
					}
				}

				if (isFound) {
					isProc = true;
				} else {
					// 次の順番が見つからなかった
					if (lastEnd == idx) {
						// 繰り返し終了すらみつからなかった場合、演奏終了
						break;
					}
					isInOrder = false;
					idx = lastEnd;
					repeatCount = 1;
					preStart = idx;
				}
			}
			entry = list.get(idx);
			note = entry.getValue();

			// 音符の追加
			note = note.clone();
			note.setRepeatStart(0);
			note.setRepeatEnd(0);
			note.setOrders(null);
			melody.put(nextPos.clone(), note);
			nextPos.add(note.getRealLength());

			++idx;
		}

		track.setMelody(melody);
	}

	/**
	 * Trackのリピート記号を展開し、繰り返し情報を取り除く。
	 * <p>
	 * <em>注)</em> Trackは最適化({@link Translator#optimizeTrack(Track)})されているものとする。
	 * </p>
	 *
	 * @param track
	 *            展開するTrack
	 */
	public void expandRepeatTrackOld(Track track) {
		int preStart = 0;
		int repeatCount = 1;
		Fraction nextPos = Fraction.ZERO.clone();

		LinkedList<Entry<Fraction, Note>> list = new LinkedList<Entry<Fraction, Note>>(
				track.getMelody().entrySet());
		TreeMap<Fraction, Note> melody = new TreeMap<>();
		boolean isInOrder = false;
		boolean isEnd = false;

		LoopMain: for (int idx = 0; idx < list.size();) {
			Entry<Fraction, Note> entry = list.get(idx);
			Note note = entry.getValue();

			boolean isPreEnd = isEnd;
			isEnd = false;

			// 繰り返しの終わり(:|)に来た場合で、まだ繰り返す必要のある場合
			if (note.getRepeatEnd() >= repeatCount || (isInOrder && note.getRepeatEnd() > 0)) {
				idx = preStart;
				++repeatCount;
				isEnd = true;
				continue LoopMain;
			}

			// 繰り返し開始の順番([1 [2など)に来た場合、現在の繰り返し順番と等しいものに当たるまで探す
			// なければ最後の繰り返し終了(:|)まで探す
			LinkedList<Integer> orders = note.getOrders();
			if (orders != null) {
				isInOrder = true;
				int lastEnd = idx; // 見つかった最後のリピート終了(:|)

				for (int nextIdx = idx; nextIdx < list.size(); ++nextIdx) {
					Note nextNote = list.get(nextIdx).getValue();
					orders = nextNote.getOrders();

					// リピート順番にあればその音符を追加して次から再開
					if (orders != null && orders.contains(repeatCount)) {
						// 追加しておかないと無限ループへ突入。フラグを持たせておいて追加は一箇所で行う？
						nextNote = nextNote.clone();
						nextNote.setRepeatStart(0);
						nextNote.setRepeatEnd(0);
						nextNote.setOrders(null);
						melody.put(nextPos.clone(), nextNote);
						nextPos.add(nextNote.getRealLength());
						idx = nextIdx + 1;
						continue LoopMain;
					}

					if (nextNote.getRepeatEnd() > 0) {
						lastEnd = nextIdx;
					}

					if (nextNote.getRepeatStart() > 0) {
						// 繰り返し開始があれば、次の繰り返しになっているとして検索終了
						break;
					}
				}
				// 見つからなかった場合
				isInOrder = false;
				if (lastEnd == idx) {
					// リピート終了(:|)すら見つからない場合演奏終了
					break;
				}

				idx = lastEnd;
				continue LoopMain;
			}

			// リピート開始位置
			if (!isPreEnd) {
				if (note.getRepeatStart() > 0) {
					preStart = idx;
					repeatCount = 1;
				}
			}

			// 音符の追加
			note = note.clone();
			note.setRepeatStart(0);
			note.setRepeatEnd(0);
			note.setOrders(null);
			melody.put(nextPos.clone(), note);
			nextPos.add(note.getRealLength());

			++idx;
		}

		track.setMelody(melody);
	}

	/**
	 * 全てのTrackに対し最適化({@link Translator#optimizeTrack(Track)})を行う。
	 *
	 * @param score
	 *            最適化を行うScore
	 * @param eraseRest
	 *            休符を削除するかどうか
	 */
	public void optimize(Score score, boolean eraseRest) {
		for (Tune tune : score.getTunes().values()) {
			for (Track track : tune.getTracks().values()) {
				this.optimizeTrack(track, eraseRest);
			}
		}
	}

	/**
	 * Trackの最適化のため以下を行う。
	 * <ol>
	 * <li>休符や意味のないを全て取り除く</li>
	 * <li>次の音符までの長さを再計算する</li>
	 *
	 * @param track
	 *            対象Track
	 * @param eraseRest
	 *            休符を削除するかどうか
	 */
	public void optimizeTrack(Track track, boolean eraseRest) {
		Fraction prevPos = null;
		for (Iterator<Entry<Fraction, Note>> iter = track.getMelody().descendingMap().entrySet().iterator(); iter.hasNext();) {
			Entry<Fraction, Note> entry = iter.next();
			Fraction pos = entry.getKey();
			Note note = entry.getValue();

			if (prevPos == null) {
				prevPos = pos;
				continue;
			}

			// 休符を取り除く
			for (Iterator<NoteUnit> iterNoteUnit = note.getNotes().iterator(); iterNoteUnit.hasNext();) {
				NoteUnit unit = iterNoteUnit.next();
				if ("xz".indexOf(unit.getNote().toLowerCase()) >= 0) {
					iterNoteUnit.remove();
				}
			}

			// 不要なNoteを取り除く
			if (eraseRest && note.getNotes().size() == 0) {
				// if (note.getInformationField().size() == 0 &&
				// note.getRepeatStart() == 0
				if (note.getInformation("K") == null && note.getInformation("L") == null
						&& note.getInformation("Q") == null && note.getInformation("M") == null
						&& note.getRepeatStart() == 0 && note.getRepeatEnd() == 0
						&& (note.getOrders() == null || note.getOrders().size() == 0)
						&& note.getVolume() == null && note.getVelocity() == null
						&& note.getDynamics() == null && note.getProgramId() == null) {
					iter.remove();
					continue;
				}
			}

			note.setRealLength(Fraction.sub(prevPos, pos));

			prevPos = pos;
		}

		// 一番初め(位置=0)に何もなければ、何もない音符を設定する
		if (track.getMelody().get(Fraction.ZERO) == null) {
			Note note = new Note();
			note.setRealLength(prevPos.clone());
			track.getMelody().put(Fraction.ZERO, note);
		}
	}

	/**
	 * 全てのTrackに対し、Grace Noteの実際に演奏する音符への変換(
	 * {@link Translator#updateGraceTrack(Track)})を行う。
	 *
	 * @param score
	 *            変換を行うScore
	 */
	public void updateGrace(Score score) {
		for (Tune tune : score.getTunes().values()) {
			for (Track track : tune.getTracks().values()) {
				this.updateGraceTrack(track);
			}
		}
	}

	/**
	 * Grace Noteを実際に演奏する音符に変換する
	 *
	 * @param track
	 *            変換対象のTrack
	 */
	public void updateGraceTrack(Track track) {
		// 長前打音の際に使用する長さにかける倍率
		Fraction appoggiaturaLength = new Fraction(1, this.opt.getAppoggiaturaNum());
		// 短前打音の際に使用する長さにかける倍率
		Fraction acciaccaturaLength = new Fraction(1, this.opt.getAcciaccaturaNum());
		// 短前打音の際、前の音符の末尾に追加するかどうか
		boolean isAcciBeforeNote = this.opt.isAcciaccaturaBeforNote();
		// 前打音の長さを親音に対する倍率にするか(演奏法としてはtrueが正しい？)
		boolean isRelative = this.opt.isRelativeAppoggiatura();

		TreeMap<Fraction, Note> newMelody = new TreeMap<Fraction, Note>();

		Note prevNote = null;
		Fraction prevPos = null;
		for (Iterator<Entry<Fraction, Note>> iter = track.getMelody().entrySet().iterator(); iter.hasNext();) {
			Entry<Fraction, Note> entry = iter.next();
			Fraction pos = entry.getKey();
			Note note = entry.getValue();

			GraceNote graceNote = note.getGrace();
			if (graceNote != null && graceNote.getNotes().size() > 0) {
				// 解析結果のGrace Note内の音符の長さの合計
				Fraction total = Fraction.ZERO.clone();
				for (Note n : graceNote.getNotes()) {
					total.add(n.getRealLength());
				}
				// 音符の長さにかける倍率
				Fraction scale = graceNote.isShortGrace() ? acciaccaturaLength : appoggiaturaLength;

				// 実際に鳴らすGrace Noteの音符の長さの合計
				Fraction length;
				if (note.getNotes().size() > 0) {
					length = Fraction.mul(note.getNotes().getFirst().getRealLength(), scale);
				} else {
					length = Fraction.mul(note.getRealLength(), scale);
				}
				if (length.getNumerator() % 3 == 0) {
					// TODO 3で割り切れたら4/3倍する？
					length.mul(new Fraction(4, 3));
				}

				if (!isRelative) {
					Fraction length2 = Fraction.mul(total, scale);
					if (length2.compareTo(length) < 0) {
						length = length2;
					}
				}

				// Grace Noteの長さにかける倍率を計算 TODO ←あやしいかも！
				scale = Fraction.div(length, total);

				if (isAcciBeforeNote && graceNote.isShortGrace() && prevPos != null) {
					// 前の音符の後ろにつける

					// 前の音符の長さを減らす
					for (NoteUnit unit : prevNote.getNotes()) {
						if (Fraction.add(prevPos, unit.getRealLength()).equals(pos)) {
							unit.setRealLength(unit.getRealLength().sub(length));
						}
						if (Fraction.add(prevPos, prevNote.getRealLength()).equals(pos)) {
							prevNote.setRealLength(prevNote.getRealLength().sub(length));
						}
					}

					// 音符を追加する
					Fraction posGrace = Fraction.sub(pos, length);
					boolean isFirst = true;
					for (Note n : graceNote.getNotes()) {
						for (NoteUnit unit : n.getNotes()) {
							unit.setRealLength(unit.getRealLength().mul(scale));
						}
						n.setRealLength(n.getRealLength().mul(scale));

						if (isFirst && note.getNoteDecorations().contains(EDecoration.TRILL_END)) {
							// XXX トリルの特殊対応(あんまりこんなことやりたくないなぁ)
							n.getNoteDecorations().add(EDecoration.TRILL_END);
						}

						newMelody.put(posGrace.clone(), n);
						posGrace.add(n.getRealLength());
						isFirst = false;
					}
				} else {
					// 次の音符の前につける
					Fraction posGrace = pos.clone();
					// 音符を追加する
					boolean isMove = false;
					boolean isFirst = true;
					for (Note n : graceNote.getNotes()) {
						for (NoteUnit unit : n.getNotes()) {
							unit.setRealLength(unit.getRealLength().mul(scale));
						}
						n.setRealLength(n.getRealLength().mul(scale));

						if (isFirst && note.getNoteDecorations().contains(EDecoration.TRILL_END)) {
							// XXX トリルの特殊対応(あんまりこんなことやりたくないなぁ)
							n.getNoteDecorations().add(EDecoration.TRILL_END);
						}

						if (!isMove) {
							if (note.getVelocity() != null) {
								n.setVelocity(note.getVelocity());
								note.setVelocity(null);
							}
							if (note.getVolume() != null) {
								n.setVolume(note.getVolume());
								note.setVolume(null);
							}
							isMove = true;
						}

						newMelody.put(posGrace.clone(), n);
						posGrace.add(n.getRealLength());
						isFirst = false;
					}

					// 後の音符の長さを減らす
					for (NoteUnit unit : note.getNotes()) {
						unit.setRealLength(unit.getRealLength().sub(length));
					}
					note.setRealLength(note.getRealLength().sub(length));

					// 位置をずらす
					pos = posGrace;
				}

				// Grace Noteの削除
				note.setGrace(null);
			}

			newMelody.put(pos.clone(), note);
			prevNote = note;
			prevPos = pos;
		}

		track.setMelody(newMelody);
	}

	/**
	 * 全てのTrackに対し、装飾の変換を行う。
	 *
	 * @param score
	 *            変換を行うScore
	 */
	public void updateDecoration(Score score) {
		for (Tune tune : score.getTunes().values()) {
			for (Track track : tune.getTracks().values()) {
				this.updateDecorationTrack(track);
			}
		}
	}

	/**
	 * 装飾を変換する
	 *
	 * @param track
	 *            変換対象のTrack
	 */
	public void updateDecorationTrack(Track track) {
		for (Iterator<Entry<Fraction, Note>> iter = track.getMelody().entrySet().iterator(); iter.hasNext();) {
			Entry<Fraction, Note> entry = iter.next();
			Note note = entry.getValue();

			// スタッカート
			for (NoteUnit unit : note.getNotes()) {
				while (unit.getDecorations().contains(EDecoration.STACCATO)) {
					unit.setRealLength(unit.getRealLength().div(new Fraction(2)));
					unit.getDecorations().remove(EDecoration.STACCATO);
				}
			}

			// 強弱法
			if (note.getDynamics() != null) {
				int value = this.opt.getDynamicsValue(note.getDynamics());
				if (this.opt.isDynamicsVolume()) {
					note.setVolume(value);
				} else {
					note.setVelocity(value);
				}
			}
		}

		int oldVelocity = DEFAULT_VELOCITY;
		int oldVolume = DEFAULT_VOLUME;
		int oldTempo = DEFAULT_TEMPO;
		Fraction lengthSlur = this.opt.getSlurLength();
		LinkedList<AInformationField> field = track.getParent().getInformation("Q");
		if (field != null && field.size() > 0) {
			Tempo tempo = ((InformationFieldTempo) field.getLast()).getTempo().clone();
			tempo.convertBase(TEMPO_BASE);
			oldTempo = tempo.getCount();
		}
		int slurCount = 0;

		this.updateDecorationTrack2(track);

		LinkedList<Fraction> posList = new LinkedList<Fraction>(track.getMelody().keySet());
		for (int idx = 0; idx < posList.size(); ++idx) {
			Fraction pos = posList.get(idx);
			Note note = track.getMelody().get(pos);
			LinkedList<EDecoration> decorations = note.getNoteDecorations();

			// information field
			{
				field = note.getInformation("Q");
				if (field != null && field.size() > 0) {
					Tempo tempo = ((InformationFieldTempo) field.getLast()).getTempo().clone();
					tempo.convertBase(TEMPO_BASE);
					oldTempo = tempo.getCount();
				}

				if (note.getVelocity() != null) {
					oldVelocity = note.getVelocity();
				}

				if (note.getVolume() != null) {
					oldVolume = note.getVolume();
				}
			}

			// アルペッジョ
			if (decorations.contains(EDecoration.ARPEGGIO) && note.getNotes().size() >= 2) {
				LinkedList<NoteUnit> notes = new LinkedList<NoteUnit>();
				for (NoteUnit u : note.getNotes()) {
					notes.add(u.clone());
				}
				Collections.sort(notes);

				Fraction length = note.getRealLength().clone();
				Fraction elapsedLength = Fraction.ZERO.clone();
				Fraction diff = this.opt.getArpeggioLength();
				{
					Fraction diff2 = Fraction.div(note.getRealLength(), new Fraction(notes.size()));
					if (diff.compareTo(diff2) > 0) {
						diff = diff2;
					}
				}

				note.getNotes().clear();
				note.addNote(notes.getFirst().clone());
				note.setRealLength(diff.clone());

				Note prevNote = null;
				for (int i = 1; i < notes.size(); ++i) {
					prevNote = new Note();
					elapsedLength.add(diff);
					NoteUnit unit = notes.get(i).clone();
					unit.setRealLength(unit.getRealLength().sub(elapsedLength));
					if (unit.getRealLength().compareTo(Fraction.ZERO) > 0) {
						prevNote.addNote(unit);
					}
					prevNote.setRealLength(diff.clone());
					track.getMelody().put(Fraction.add(pos, elapsedLength), prevNote);
				}
				prevNote.setRealLength(Fraction.sub(length, elapsedLength));
			}

			// スラー
			{
				while (decorations.remove(EDecoration.SLUR_START)) {
					++slurCount;
				}
				while (decorations.remove(EDecoration.SLUR_END)) {
					--slurCount;
				}

				if (this.opt.getSlurOutputType() == 2 && slurCount <= 0) {
					// TODO スラー外の音を短くする
					for (NoteUnit unit : note.getNotes()) {
						if (unit.isWithTie()) {
							// タイが付与されていれば何もしない
							continue;
						}

						Fraction length = lengthSlur;

						// 最大で自身の音の半分までしか縮めない
						Fraction length2 = Fraction.div(unit.getRealLength(), new Fraction(2));
						if (length2.compareTo(length) < 0) {
							length = length2;
						}

						unit.setRealLength(unit.getRealLength().sub(length));
					}
				}

				if (this.opt.getSlurOutputType() == 1 && slurCount > 0) {
					// スラー内の音を長くする
					for (NoteUnit unit : note.getNotes()) {
						if (unit.isWithTie()) {
							// タイが付与されていれば何もしない
							continue;
						}

						if (unit.getRealLength().compareTo(note.getRealLength()) < 0) {
							// 次の音符まで音の長さが足りてなければ、足りない分の半分だけ伸ばす
							Fraction length = Fraction.sub(
									note.getRealLength(), unit.getRealLength()).div(new Fraction(2));
							if (length.compareTo(lengthSlur) > 0) {
								length = lengthSlur;
							}
							unit.setRealLength(unit.getRealLength().add(length));

						} else {
							// スラー伸縮の長さを足すが、音が終わる位置に次の音があった場合、
							// 次の音の半分の位置を超えない長さにする
							Fraction length = lengthSlur;

							Fraction nextPos = Fraction.add(pos, unit.getRealLength());
							Note nextNote = track.getMelody().get(nextPos);
							if (nextNote != null) {
								// 音符があれば、その音の長さの半分を超えないようにする

								boolean isFound = false;
								for (NoteUnit u : nextNote.getNotes()) {
									if (unit.compareTo(u) == 0) {
										// 同じ高さの音が続いてたら音を長くしない
										// XXX MIDIの仕様？回避
										length = Fraction.ZERO;
										isFound = true;
										break;
									}
								}

								if (!isFound) {
									// 最大で次の音の半分までしか伸ばさない
									Fraction length2 = Fraction.div(
											nextNote.getRealLength(), new Fraction(2));
									if (length2.compareTo(length) < 0) {
										length = length2;
									}
								}
							}
							unit.setRealLength(unit.getRealLength().add(length));
						}
					}
				}
			}

			// クレシェンド、ディミニュエンド
			{
				Integer lastValue = null;
				Fraction endPos = null;
				boolean isFound = false;
				boolean isIncrese = false;

				// Tune全体
				if (decorations.contains(EDecoration.CRES_START)) {
					isFound = true;
					isIncrese = true;
				} else if (decorations.contains(EDecoration.DIMI_START)) {
					isFound = true;
					isIncrese = false;
				}
				if (isFound) {
					endPos = this.searchEnd(track, posList, idx + 1, isIncrese
							? EDecoration.CRES_END
							: EDecoration.DIMI_END);
					Note n = track.getMelody().get(endPos);
					if (n.getVolume() != null) {
						lastValue = n.getVolume();
					} else {
						int amount = this.opt.getVolumeChangeAmount();

						lastValue = oldVolume + (isIncrese ? amount : -amount);
					}

					this.updateVolumeChange(track, posList, pos, endPos, oldVolume, lastValue);
					oldVolume = lastValue;
				}

				// トラックごと
				isFound = false;
				if (decorations.contains(EDecoration.TRACK_CRES_START)) {
					isFound = true;
					isIncrese = true;
				} else if (decorations.contains(EDecoration.TRACK_DIMI_START)) {
					isFound = true;
					isIncrese = false;
				}
				if (isFound) {
					endPos = this.searchEnd(track, posList, idx + 1, isIncrese
							? EDecoration.TRACK_CRES_END
							: EDecoration.TRACK_DIMI_END);
					Note n = track.getMelody().get(endPos);
					if (n.getVelocity() != null) {
						lastValue = n.getVelocity();
					} else {
						int amount = this.opt.getVelocityChangeAmount();
						lastValue = oldVelocity + (isIncrese ? amount : -amount);
					}

					this.updateVelocityChange(track, posList, pos, endPos, oldVelocity, lastValue);
					oldVelocity = lastValue;
				}
			}

			// テンポチェンジ
			{
				Integer lastValue = null;
				Fraction endPos = null;
				boolean isFound = false;
				boolean isIncrease = false;

				if (decorations.contains(EDecoration.ACCE_START)) {
					isFound = true;
					isIncrease = true;
				} else if (decorations.contains(EDecoration.RITA_START)) {
					isFound = true;
					isIncrease = false;
				}
				if (isFound) {
					endPos = this.searchEnd(track, posList, idx + 1, isIncrease
							? EDecoration.ACCE_END
							: EDecoration.RITA_END);
					Note n = track.getMelody().get(endPos);
					field = n.getInformation("Q");
					if (field != null && field.size() > 0) {
						Tempo tempo = ((InformationFieldTempo) field.getLast()).getTempo();
						lastValue = (int) new Fraction(tempo.getCount()).mul(tempo.getBase()).div(
								TEMPO_BASE).getDouble();
					} else {
						lastValue = oldTempo
								+ (isIncrease
										? this.opt.getTempoChangeAmount()
										: -this.opt.getTempoChangeAmount());
					}

					this.updateTempoChange(track, posList, pos, endPos, oldTempo, lastValue);
					oldTempo = lastValue;
				}
			}

		}
	}

	/**
	 * 装飾(トリル、モルデント、ターンなど)を変換する
	 *
	 * @param track
	 *            変換対象のTrack
	 */
	private void updateDecorationTrack2(Track track) {
		Key key = new Key();
		LinkedList<AInformationField> field = track.getParent().getInformation("K");
		if (field != null && field.size() > 0) {
			key = ((InformationFieldKey) field.getLast()).getKey();
		}

		LinkedList<EDecoration> decos = new LinkedList<EDecoration>();
		decos.add(EDecoration.TRILL);
		decos.add(EDecoration.TRILL_START);
		decos.add(EDecoration.TURN);
		decos.add(EDecoration.TURNX);
		decos.add(EDecoration.AFTER_TURN);
		decos.add(EDecoration.AFTER_TURNX);
		decos.add(EDecoration.LOWER_MORDENT);
		decos.add(EDecoration.UPPER_MORDENT);

		String step = "CDEFGABcdefgab";

		TreeMap<Fraction, Note> newMelody = new TreeMap<Fraction, Note>();
		LinkedList<Fraction> posList = new LinkedList<Fraction>(track.getMelody().keySet());
		for (int idx = 0; idx < posList.size(); ++idx) {
			Fraction pos = posList.get(idx);
			Note note = track.getMelody().get(pos);
			Note newNote = note.clone();
			{
				// 既に音符が追加されてる？ (トリルとかで追加される可能性がある)
				Note n = newMelody.get(pos);
				if (n != null) {
					newNote.getNotes().addAll(n.getNotes());
				}
				newMelody.put(pos.clone(), newNote);
			}

			LinkedList<EDecoration> decorations = note.getNoteDecorations();

			field = note.getInformation("K");
			if (field != null && field.size() > 0) {
				key = ((InformationFieldKey) field.getLast()).getKey();
			}

			// トリル、ターン、モルデント等
			if (decorations.contains(EDecoration.TRILL_END)) {
				while (decos.remove(EDecoration.IN_TRILL))
					;
			}
			LoopDeco: for (int iDecos = 0; iDecos < decos.size(); ++iDecos) {
				EDecoration deco = decos.get(iDecos);
				int idxDeco = decorations.indexOf(deco);
				if (idxDeco >= 0 || deco == EDecoration.IN_TRILL) {
					if (deco == EDecoration.TRILL_START) {
						decos.add(EDecoration.IN_TRILL);
					}
					if (deco == EDecoration.TRILL_START && decorations.contains(EDecoration.TRILL)) {
						// トリル開始でありtrがあれば、どちらかで処理をする
						continue LoopDeco;
					}
					if (deco == EDecoration.IN_TRILL
							&& (decorations.contains(EDecoration.TRILL_START) || decorations.contains(EDecoration.TRILL))) {
						continue LoopDeco;
					}

					for (int idxUnit = note.getNotes().size() - 1; idxUnit >= 0; --idxUnit) {
						NoteUnit unit = note.getNotes().get(idxUnit);
						LinkedList<Integer> degrees = new LinkedList<Integer>();
						LinkedList<Fraction> lengths = new LinkedList<Fraction>();
						EAccidental upperAcci = null;
						EAccidental lowerAcci = null;
						if (idxDeco >= 1) {
							switch (decorations.get(idxDeco - 1)) {
							case DOUBLE_SHARP:
								upperAcci = EAccidental.DOUBLE_SHARP;
								break;
							case SHARP:
								upperAcci = EAccidental.SHARP;
								break;
							case NATURAL:
								upperAcci = EAccidental.NATURAL;
								break;
							case FLAT:
								upperAcci = EAccidental.FLAT;
								break;
							case DOUBLE_FLAT:
								upperAcci = EAccidental.DOUBLE_FLAT;
								break;
							default:
								break;
							}
						}
						if (idxDeco + 1 < decorations.size()) {
							switch (decorations.get(idxDeco + 1)) {
							case DOUBLE_SHARP:
								lowerAcci = EAccidental.DOUBLE_SHARP;
								break;
							case SHARP:
								lowerAcci = EAccidental.SHARP;
								break;
							case NATURAL:
								lowerAcci = EAccidental.NATURAL;
								break;
							case FLAT:
								lowerAcci = EAccidental.FLAT;
								break;
							case DOUBLE_FLAT:
								lowerAcci = EAccidental.DOUBLE_FLAT;
								break;
							default:
								break;
							}
						}

						/*
						 * if (false && deco == EDecoration.TRILL) { //
						 * トリル。1度->2度上->1度->2度上->1度と素早く // →TRILL_STARTと同列にする？
						 * Fraction length =
						 * this.opt.getDecorationLength().clone(); if
						 * (Fraction.mul(length, new Fraction(4)).compareTo(
						 * unit.getRealLength()) > 0) { length =
						 * Fraction.div(unit.getRealLength(), new Fraction(5));
						 * } degrees.add(0); lengths.add(length.clone());
						 * degrees.add(1); lengths.add(length.clone());
						 * degrees.add(0); lengths.add(length.clone());
						 * degrees.add(1); lengths.add(length.clone());
						 * degrees.add(0); lengths.add(Fraction.sub(
						 * unit.getRealLength(), Fraction.mul(length, new
						 * Fraction(lengths.size())))); } else
						 */
						if (deco == EDecoration.TRILL_START || deco == EDecoration.TRILL
								|| deco == EDecoration.IN_TRILL) {
							// 長いトリル。1度->2度上->1度->2度上->…と素早く
							// 最後の4つ分は5連符にして …->2度上->1度で終わる
							// トリルの終了は探さない
							Fraction lengthPre = this.opt.getDecorationLength().clone();
							Fraction lengthPost = null;
							Fraction count = Fraction.div(unit.getRealLength(), lengthPre);
							int cntPre = -1;
							int cntPost = -1;
							cntPre = (int) count.getDouble() - 4;
							if (!this.opt.isTrillFromUpper() && cntPre % 2 == 1) {
								cntPost = 4;
							} else if (this.opt.isTrillFromUpper() && cntPre % 2 == 0) {
								cntPost = 4;
							} else {
								cntPost = 5;
							}
							lengthPost = Fraction.sub(
									unit.getRealLength(),
									Fraction.mul(lengthPre, new Fraction(cntPre))).div(
									new Fraction(cntPost));
							if (lengthPost.compareTo(MIN_LENGTH) < 0) {
								cntPost -= 2;
								lengthPost = Fraction.sub(
										unit.getRealLength(),
										Fraction.mul(lengthPre, new Fraction(cntPre))).div(
										new Fraction(cntPost));
							}

							int degree = this.opt.isTrillFromUpper() ? 1 : 0;
							for (int i = cntPre; i > 0; --i) {
								degrees.add(degree);
								lengths.add(lengthPre.clone());
								degree = 1 - degree;
							}
							for (int i = cntPost; i > 0; --i) {
								degrees.add(degree);
								lengths.add(lengthPost.clone());
								degree = 1 - degree;
							}
						} else if (deco == EDecoration.UPPER_MORDENT) {
							// プラルトリラー。1度->2度上->1度と素早く
							Fraction length = this.opt.getDecorationLength().clone();
							if (Fraction.mul(length, new Fraction(3)).compareTo(
									unit.getRealLength()) > 0) {
								length = Fraction.div(unit.getRealLength(), new Fraction(3));
							}
							degrees.add(0);
							lengths.add(length.clone());
							degrees.add(1);
							lengths.add(length.clone());
							degrees.add(0);
							lengths.add(Fraction.sub(
									unit.getRealLength(),
									Fraction.mul(length, new Fraction(lengths.size()))));
						} else if (deco == EDecoration.LOWER_MORDENT) {
							// モルデント。1度->2度下->1度と素早く
							Fraction length = this.opt.getDecorationLength().clone();
							if (Fraction.mul(length, new Fraction(3)).compareTo(
									unit.getRealLength()) > 0) {
								length = Fraction.div(unit.getRealLength(), new Fraction(3));
							}
							degrees.add(0);
							lengths.add(length.clone());
							degrees.add(-1);
							lengths.add(length.clone());
							degrees.add(0);
							lengths.add(Fraction.sub(
									unit.getRealLength(),
									Fraction.mul(length, new Fraction(lengths.size()))));
						} else if (deco == EDecoration.TURN) {
							// ターン。2度上->1度->2度下->1度と素早く
							Fraction length = this.opt.getDecorationLength().clone();
							if (Fraction.mul(length, new Fraction(4)).compareTo(
									unit.getRealLength()) > 0) {
								length = Fraction.div(unit.getRealLength(), new Fraction(4));
							}
							degrees.add(1);
							lengths.add(length.clone());
							degrees.add(0);
							lengths.add(length.clone());
							degrees.add(-1);
							lengths.add(length.clone());
							degrees.add(0);
							lengths.add(Fraction.sub(
									unit.getRealLength(),
									Fraction.mul(length, new Fraction(lengths.size()))));
						} else if (deco == EDecoration.TURNX) {
							// 転回ターン。2度下->1度->2度上->1度と素早く
							Fraction length = this.opt.getDecorationLength().clone();
							if (Fraction.mul(length, new Fraction(4)).compareTo(
									unit.getRealLength()) > 0) {
								length = Fraction.div(unit.getRealLength(), new Fraction(4));
							}
							degrees.add(-1);
							lengths.add(length.clone());
							degrees.add(0);
							lengths.add(length.clone());
							degrees.add(1);
							lengths.add(length.clone());
							degrees.add(0);
							lengths.add(Fraction.sub(
									unit.getRealLength(),
									Fraction.mul(length, new Fraction(lengths.size()))));
						} else if (deco == EDecoration.AFTER_TURN) {
							// ターン(音符間)。1度(長く)->2度上->1度->2度下->1度と素早く
							Fraction length = this.opt.getDecorationLength().clone();
							if (Fraction.mul(length, new Fraction(5)).compareTo(
									unit.getRealLength()) > 0) {
								length = Fraction.div(unit.getRealLength(), new Fraction(5));
							}
							// 最初の音は最後に先頭に挿入
							degrees.add(1);
							lengths.add(length.clone());
							degrees.add(0);
							lengths.add(length.clone());
							degrees.add(-1);
							lengths.add(length.clone());
							degrees.add(0);
							lengths.add(length.clone());
							degrees.addFirst(0);
							lengths.addFirst(Fraction.sub(
									unit.getRealLength(),
									Fraction.mul(length, new Fraction(lengths.size()))));
						} else if (deco == EDecoration.AFTER_TURNX) {
							// ターン(音符間)。1度(長く)->2度下->1度->2度上->1度と素早く
							Fraction length = this.opt.getDecorationLength().clone();
							if (Fraction.mul(length, new Fraction(5)).compareTo(
									unit.getRealLength()) > 0) {
								length = Fraction.div(unit.getRealLength(), new Fraction(5));
							}
							// 最初の音は最後に先頭に挿入
							degrees.add(-1);
							lengths.add(length.clone());
							degrees.add(0);
							lengths.add(length.clone());
							degrees.add(1);
							lengths.add(length.clone());
							degrees.add(0);
							lengths.add(length.clone());
							degrees.addFirst(0);
							lengths.addFirst(Fraction.sub(
									unit.getRealLength(),
									Fraction.mul(length, new Fraction(lengths.size()))));
						}

						if (degrees.size() > 0) {
							// ここには絶対来るはずだが…
							newNote.getNotes().remove(idxUnit);
							Fraction nextPos = pos.clone();

							int base = step.indexOf(unit.getNote());
							for (int i = 0; i < degrees.size(); ++i) {
								int degree = degrees.get(i);
								Fraction length = lengths.get(i);

								NoteUnit newUnit = unit.clone();

								// 音程調整
								if (degree == 0) {
									// 何も必要なし
								} else {
									int newBase = base + degree;
									if (newBase >= step.length()) {
										newBase -= 7;
										newUnit.setOctave(newUnit.getOctave() + 1);
									}
									if (newBase < 0) {
										newBase += 7;
										newUnit.setOctave(newUnit.getOctave() - 1);
									}
									newUnit.setNote(step.substring(newBase, newBase + 1));
									if (newUnit.getNote().toLowerCase().equals(newUnit.getNote())
											&& newUnit.getOctave() < 0) {
										newUnit.setNote(newUnit.getNote().toUpperCase());
										newUnit.setOctave(newUnit.getOctave() + 1);
									}
									if (newUnit.getNote().toUpperCase().equals(newUnit.getNote())
											&& newUnit.getOctave() > 0) {
										newUnit.setNote(newUnit.getNote().toLowerCase());
										newUnit.setOctave(newUnit.getOctave() - 1);
									}

									EAccidental acci = degree > 0 ? upperAcci : lowerAcci;
									if (acci != null) {
										newUnit.setAcci(acci);
									} else if (key.getAccidental(newUnit.getNote()) != null) {
										newUnit.setAcci(key.getAccidental(newUnit.getNote()));
									} else {
										newUnit.setAcci(EAccidental.NATURAL);
									}
								}

								// 長さ
								newUnit.setRealLength(length.clone());

								Note nextNote = newMelody.get(nextPos);
								if (nextNote == null) {
									nextNote = new Note();
									nextNote.setRealLength(length.clone());
									newMelody.put(nextPos.clone(), nextNote);
								}
								nextNote.getNotes().add(newUnit);

								nextPos.add(length);
							}
						}
					}
				}
			}
		}
		track.setMelody(newMelody);
	}

	/**
	 * クレッシェンドなどの終わりを探す
	 *
	 * @param track
	 *            検索するトラック
	 * @param posList
	 *            メロディの開始位置リスト
	 * @param fromIndex
	 *            検索を開始する位置
	 * @param deco
	 *            終わり
	 * @return 終了位置
	 */
	private Fraction searchEnd(Track track, LinkedList<Fraction> posList, int fromIndex,
			EDecoration deco) {
		Fraction pos = null;
		for (int idx = fromIndex; idx < posList.size(); ++idx) {
			pos = posList.get(idx);
			Note note = track.getMelody().get(pos);
			if (note.getNoteDecorations().contains(deco)) {
				return pos;
			}
		}
		return pos;
	}

	/**
	 * Volume変化を更新する
	 *
	 * @param track
	 *            更新するTrack
	 * @param posList
	 *            音符の開始位置リスト
	 * @param fromPos
	 *            変化開始位置
	 * @param toPos
	 *            変化終了位置
	 * @param fromValue
	 *            変化開始時の値
	 * @param toValue
	 *            変化終了時の値
	 */
	private void updateVolumeChange(Track track, LinkedList<Fraction> posList, Fraction fromPos,
			Fraction toPos, int fromValue, int toValue) {
		// 決められた間隔に<V:>を挿入する

		int valDiff = toValue - fromValue;
		Fraction posDiff = Fraction.sub(toPos, fromPos);
		Fraction pos = fromPos.clone();
		Fraction diff = this.opt.getVolumeChangeLength();

		int prevValue = fromValue;

		while (pos.compareTo(toPos) < 0) {
			if (!track.getMelody().containsKey(pos)) {
				track.setEmptyNoteAt(pos);
			}
			double dd = (double) valDiff
					* Fraction.div(Fraction.sub(pos, fromPos), posDiff).getDouble();
			int val = fromValue + (int) (dd + 0.5);
			if (val != prevValue) {
				track.getMelody().get(pos).setVolume(val > MIN_VOLUME ? val : MIN_VOLUME);
			}
			prevValue = val;
			pos.add(diff);
		}
		if (prevValue != toValue) {
			track.getMelody().get(toPos).setVolume(toValue);
		}
	}

	/**
	 * Velocity変化を更新する
	 *
	 * @param track
	 *            更新するTrack
	 * @param posList
	 *            音符の開始位置リスト
	 * @param fromPos
	 *            変化開始位置
	 * @param toPos
	 *            変化終了位置
	 * @param fromValue
	 *            変化開始時の値
	 * @param toValue
	 *            変化終了時の値
	 */
	private void updateVelocityChange(Track track, LinkedList<Fraction> posList, Fraction fromPos,
			Fraction toPos, int fromValue, int toValue) {
		// *
		// 決められた間隔に<V:>を挿入する

		int valDiff = toValue - fromValue;
		Fraction posDiff = Fraction.sub(toPos, fromPos);
		Fraction pos = fromPos.clone();
		Fraction diff = this.opt.getVelocityChangeLength();

		int prevValue = fromValue;

		while (pos.compareTo(toPos) < 0) {
			if (!track.getMelody().containsKey(pos)) {
				track.setEmptyNoteAt(pos);
			}
			double dd = (double) valDiff
					* Fraction.div(Fraction.sub(pos, fromPos), posDiff).getDouble();
			int val = fromValue + (int) (dd + 0.5);
			if (val != prevValue) {
				track.getMelody().get(pos).setVelocity(val > MIN_VOLUME ? val : MIN_VOLUME);
			}
			prevValue = val;
			pos.add(diff);
		}
		if (prevValue != toValue) {
			track.getMelody().get(toPos).setVelocity(toValue);
		}
		/*
		 * / int diffValue = toValue - fromValue; Collection<Note> notes =
		 * track.getMelody().subMap(fromPos, toPos).values();
		 *
		 * int idx = 0; for (Note note : notes) { note.setVelocity(fromValue +
		 * (int) (0.5 + idx * diffValue / (double) notes.size())); if
		 * (note.getVelocity() < MIN_VELOCITY) { note.setVelocity(MIN_VELOCITY);
		 * } ++idx; } track.getMelody().get(toPos).setVelocity(toValue); //
		 */
	}

	/**
	 * テンポ変化を更新する
	 *
	 * @param track
	 *            更新するTrack
	 * @param posList
	 *            音符の開始位置リスト
	 * @param fromPos
	 *            変化開始位置
	 * @param toPos
	 *            変化終了位置
	 * @param fromValue
	 *            変化開始時の値
	 * @param toValue
	 *            変化終了時の値
	 */
	private void updateTempoChange(Track track, LinkedList<Fraction> posList, Fraction fromPos,
			Fraction toPos, int fromValue, int toValue) {
		int valDiff = toValue - fromValue;
		Fraction posDiff = Fraction.sub(toPos, fromPos);
		Fraction pos = fromPos.clone();
		Fraction diff = this.opt.getTempoChangeLength();

		int prevValue = fromValue;

		while (pos.compareTo(toPos) < 0) {
			if (!track.getMelody().containsKey(pos)) {
				track.setEmptyNoteAt(pos);
			}
			double dd = (double) valDiff
					* Fraction.div(Fraction.sub(pos, fromPos), posDiff).getDouble();
			int val = fromValue + (int) (dd + 0.5);
			if (val != prevValue) {
				Tempo tempo = new Tempo(TEMPO_BASE.clone(), (val > MIN_TEMPO ? val : MIN_TEMPO));
				track.getMelody().get(pos).addInformation("Q", new InformationFieldTempo(tempo));
			}
			prevValue = val;
			pos.add(diff);
		}
		if (prevValue != toValue) {
			Tempo tempo = new Tempo(TEMPO_BASE.clone(), toValue);
			track.getMelody().get(toPos).addInformation("Q", new InformationFieldTempo(tempo));
		}
	}

	/**
	 * Voice Overlayにより分割されているTrackを合成する
	 *
	 * @param score
	 *            合成するScore
	 */
	public void combineVoiceOverlay(Score score) {
		for (Tune tune : score.getTunes().values()) {
			for (Iterator<Entry<String, Track>> iter = tune.getTracks().entrySet().iterator(); iter.hasNext();) {
				Entry<String, Track> entry = iter.next();
				String trackId = entry.getKey();
				Track track = entry.getValue();
				String[] idOverlay = ParserTuneBody.getTrackIdAndOverlay(trackId);
				if (!idOverlay[1].equals("0")) {
					Track dstTrack = tune.getTracks().get(idOverlay[0]);
					dstTrack.combineWith(track);
					this.optimizeTrack(dstTrack, false);
					iter.remove();
				}
			}
		}

		for (TrackSrcDstID id : this.opt.getCombineList()) {
			Tune srcTune = score.getTunes().get(id.getSrcTuneId());
			if (srcTune == null) {
				continue;
			}
			Track srcTrack = srcTune.getTracks().get(id.getSrcTrackId());
			if (srcTrack == null) {
				continue;
			}
			Tune dstTune = score.getTunes().get(id.getDstTuneId());
			if (dstTune == null) {
				continue;
			}
			Track dstTrack = dstTune.getTracks().get(id.getDstTrackId());
			if (dstTrack == null) {
				continue;
			}

			dstTrack.combineWith(srcTrack);
			this.optimizeTrack(dstTrack, false);
		}
	}

	/**
	 * トラック間で情報をコピーする
	 *
	 * @param score
	 *            対象のScore
	 */
	public void copyInformation(Score score) {
		for (TrackSrcDstID id : this.opt.getCopyInformationList()) {
			Tune srcTune = score.getTunes().get(id.getSrcTuneId());
			if (srcTune == null) {
				continue;
			}
			Track srcTrack = srcTune.getTracks().get(id.getSrcTrackId());
			if (srcTrack == null) {
				continue;
			}
			Tune dstTune = score.getTunes().get(id.getDstTuneId());
			if (dstTune == null) {
				continue;
			}
			Track dstTrack = dstTune.getTracks().get(id.getDstTrackId());
			if (dstTrack == null) {
				continue;
			}

			dstTrack.copyInformationFrom(srcTrack);
			this.optimizeTrack(dstTrack, false);
		}
	}

	/**
	 * MOE-ABCに変換する
	 *
	 * @param score
	 *            変換するScore
	 * @throws IOException
	 */
	public void translate2MoeAbc(Score score) throws IOException {
		List<TrackID> list = this.opt.getOutList();
		if (list.size() == 0) {
			// 出力リストが指定されていなかった場合、全てのTrackを出力する
			for (Iterator<Entry<String, Tune>> itScore = score.getTunes().entrySet().iterator(); itScore.hasNext();) {
				Entry<String, Tune> eTune = itScore.next();
				String tuneId = eTune.getKey();
				Tune tune = eTune.getValue();

				for (Iterator<Entry<String, Track>> itTune = tune.getTracks().entrySet().iterator(); itTune.hasNext();) {
					Entry<String, Track> eTrack = itTune.next();
					String trackId = eTrack.getKey();

					list.add(new TrackID(tuneId, trackId));
				}
			}
		}

		for (TrackID id : list) {
			Tune tune = score.getTunes().get(id.getTuneId());
			if (tune == null) {
				continue;
			}

			Track track = tune.getTracks().get(id.getTrackId());
			if (track == null) {
				continue;
			}

			this.opt.setOutTrack(id.getTuneId(), id.getTrackId());

			LinkedList<AInformationField> field = tune.getInformation("T");
			this.print("%%Tune:" + id.getTuneId());
			if (field != null) {
				this.println(": " + ((InformationFieldTitle) field.getFirst()).getTitle());
			} else {
				this.println("");
			}

			this.println("%Track:" + id.getTrackId());
			this.print("{");
			this.track2MoeAbc(track);
			this.println("}");

			this.opt.resetOutTrack();
		}
	}

	/**
	 * TrackをMOE-ABCに変換す
	 *
	 * @param track
	 *            変換するTrack
	 * @throws IOException
	 */
	private void track2MoeAbc(Track track) throws IOException {
		String defaultKeyMode = "";

		track.clearTemporary();
		LinkedList<AInformationField> field;

		// キー(K:)情報
		field = track.getInformation("K");
		if (field == null || field.size() == 0) {
			this.key = new Key();
		} else {
			this.key = ((InformationFieldKey) field.getLast()).getKey().clone();
			if (!this.key.isNone()) {
				this.key.convertMode(defaultKeyMode);
				if (!this.opt.isKey2None()) {
					this.print("<K:" + this.key.getKeyString() + ">");
				}
			}
		}

		// テンポ(Q:)情報
		this.tempo = new Tempo(TEMPO_BASE, DEFAULT_TEMPO);
		field = track.getInformation("Q");
		if (field == null || field.size() == 0) {
		} else {
			Tempo newTempo = ((InformationFieldTempo) field.getLast()).getTempo();
			this.tempo = this.newTempo(newTempo, this.tempo);
		}

		// 長さ(L:)情報
		this.unitNoteLength = DEFAULT_LENGTH.clone();
		{
			Fraction newLength = track.getUnitNoteLength();
			// 拍子(M:)情報
			this.meterLength = track.getMeterLength();
			if (newLength == null) {
				newLength = InformationFieldLength.getDefaultLength(this.meterLength);
			}
			if (this.meterLength == null) {
				this.meterLength = InformationFieldMeter.getDefaultLength();
			}

			this.unitNoteLength = newUnitNoteLength(newLength, this.unitNoteLength);
		}

		// Program Id
		this.programId = null;

		int barCount = 0;
		this.barPos = Fraction.ZERO.clone();

		int preVelocity = DEFAULT_VELOCITY;
		int preVolume = DEFAULT_VOLUME;

		LinkedList<Entry<Fraction, Note>> list = new LinkedList<Entry<Fraction, Note>>(
				track.getMelody().entrySet());
		for (int idx = 0; idx < list.size(); ++idx) {
			// Trackの最後には何もしないNoteがあるはず！
			// 参照→abc.parser.parser.ParserTuneBody#parseEnd()
			Entry<Fraction, Note> nowEntry = list.get(idx);

			Note nowNote = nowEntry.getValue();

			// bar
			while (this.barPos.compareTo(this.meterLength) >= 0) {
				++barCount;
				if (this.opt.isBarOutput()) {
					this.print("|");
				}
				if (this.opt.getNewlineStep() > 0 && barCount % this.opt.getNewlineStep() == 0) {
					this.println("");
					if (this.opt.isBarComment()) {
						this.println("%" + (barCount + 1));
					}
				}
				this.barPos.sub(this.meterLength);
			}

			// information field
			field = nowNote.getInformation("K");
			if (field != null && field.size() > 0) {
				Key newKey = ((InformationFieldKey) field.getLast()).getKey();
				if (!newKey.isNone()) {
					newKey.convertMode(defaultKeyMode);
					if (!this.opt.isKey2None()
							&& !this.key.toBaseString().equals(newKey.toBaseString())) {
						this.print("<K:" + newKey.getKeyString() + ">");
					}
					this.key = newKey;
				}
			}

			field = nowNote.getInformation("Q");
			if (field != null && field.size() > 0) {
				Tempo newTempo = ((InformationFieldTempo) field.getLast()).getTempo();
				this.tempo = this.newTempo(newTempo, this.tempo);
			}

			field = nowNote.getInformation("L");
			if (field != null && field.size() > 0) {
				Fraction newLength = ((InformationFieldLength) field.getLast()).getLength();
				this.unitNoteLength = this.newUnitNoteLength(newLength, this.unitNoteLength);
			}

			field = nowNote.getInformation("M");
			if (field != null && field.size() > 0) {
				this.meterLength = ((InformationFieldMeter) field.getLast()).getMeter().getLength();
				this.barPos = Fraction.ZERO.clone();
			}

			// Program Id
			if (nowNote.getProgramId() != null) {
				if (this.programId != nowNote.getProgramId()) {
					this.programId = nowNote.getProgramId();
					this.print("<C:" + this.programId + ">");
				}
			}

			// 独自仕様フィールド(´▽｀*)
			Integer velocity = nowNote.getVelocity();
			if (velocity != null && velocity != preVelocity) {
				this.print("<v:" + (velocity == 100 ? "" : velocity) + ">");
				preVelocity = velocity;
			}

			Integer volume = nowNote.getVolume();
			if (volume != null && volume != preVolume) {
				this.print("<V:" + (volume == 100 ? "" : volume) + ">");
				preVolume = volume;
			}

			// 音符の追加
			LinkedList<NoteUnit> units = nowNote.getNotes();
			if (units == null || units.size() == 0) {
				// 空の音符
				Fraction length = nowNote.getRealLength();
				if (length == null || length.equals(Fraction.ZERO)) {
					// 何もなし
				} else {
					while (true) {
						if (length.compareTo(MAX_LENGTH) <= 0) {
							this.print("x"
									+ this.shortLength(Fraction.div(length, this.unitNoteLength)));
							break;
						} else {
							this.print("x");
							if (Fraction.sub(length, MAX_LENGTH).compareTo(MIN_LENGTH) < 0) {
								this.print(this.shortLength(Fraction.div(
										Fraction.sub(length, MIN_LENGTH), this.unitNoteLength)));
								length = MIN_LENGTH.clone();
							} else {
								this.print(this.shortLength(Fraction.div(
										MAX_LENGTH, this.unitNoteLength)));
								length.sub(MAX_LENGTH);
							}
						}
					}
				}
			} else if (units.size() == 1) {
				// 単音
				NoteUnit unit = units.getFirst();

				if (unit.getRealLength().equals(nowNote.getRealLength())) {
					// 単音の長さと次の音符までの長さが等しい場合
					this.print(this.noteUnit(unit));
				} else {
					String str;

					String noteUnit = this.noteUnit(unit);
					Fraction length = Fraction.div(nowNote.getRealLength(), this.unitNoteLength);

					str = "[" + noteUnit + "]" + this.shortLength(length, true);

					if (unit.getRealLength().compareTo(nowNote.getRealLength()) < 0) {
						str = "[" + noteUnit + "]";
						Fraction realLength = nowNote.getRealLength().clone();
						boolean isFirst = true;
						while (true) {
							if (realLength.compareTo(MAX_LENGTH) <= 0) {
								str += this.shortLength(
										Fraction.div(realLength, this.unitNoteLength), isFirst);
								break;
							} else {
								if (Fraction.sub(realLength, MAX_LENGTH).compareTo(MIN_LENGTH) < 0) {
									str += this.shortLength(Fraction.div(
											Fraction.sub(realLength, MIN_LENGTH),
											this.unitNoteLength), isFirst);
									realLength = MIN_LENGTH.clone();
								} else {
									str += this.shortLength(
											Fraction.div(MAX_LENGTH, this.unitNoteLength), isFirst);
									realLength.sub(MAX_LENGTH);
								}
								str += "z";
							}

							isFirst = false;
						}

						// 単音の長さが次の音符までの長さよりも短い場合
						// "Cz" と "[C]2" のうち短い表現を選択する (例の場合 "Cz" を選択)
						String str2;

						Fraction lengthRest = Fraction.sub(
								nowNote.getRealLength(), unit.getRealLength());

						if (lengthRest.compareTo(MIN_LENGTH) >= 0) {
							str2 = noteUnit;
							while (true) {
								if (lengthRest.compareTo(MAX_LENGTH) <= 0) {
									str2 += "z"
											+ this.shortLength(Fraction.div(
													lengthRest, this.unitNoteLength));
									break;
								} else {
									str2 += "z";
									if (Fraction.sub(lengthRest, MAX_LENGTH).compareTo(MIN_LENGTH) < 0) {
										str2 += this.shortLength(Fraction.div(
												Fraction.sub(lengthRest, MIN_LENGTH),
												this.unitNoteLength));
										lengthRest = MIN_LENGTH.clone();
									} else {
										str2 += this.shortLength(Fraction.div(
												MAX_LENGTH, this.unitNoteLength));
										lengthRest.sub(MAX_LENGTH);
									}
								}
							}
							if (str2.length() <= str.length()) {
								str = str2;
							}
						}
					}
					this.print(str);
				}
			} else {
				// 和音
				String str = null;

				// 和音の中の音符の最後の音を全通り試して、一番短い文字列になるようにする
				for (int i = units.size(); i > 0; --i) {
					// 和音の中身の文字列化
					StringBuffer sb = new StringBuffer();
					sb.append("[");
					for (int j = 0; j < units.size(); ++j) {
						sb.append(this.noteUnit(units.get(j)));
					}
					sb.append("]");

					if (units.getLast().getRealLength().equals(nowNote.getRealLength())) {
						// 和音最後の音の長さと、和音自体の長さが等しければ、和音文字列は完成
						if (str == null || str.length() > sb.length()) {
							str = sb.toString();
						}
					} else if (units.getLast().getRealLength().compareTo(nowNote.getRealLength()) < 0) {
						// 和音最後の音の長さが、和音自体の長さより短い場合、以下の2通りを試す
						// 1. 和音の後ろに休符を挿入
						// 2. 和音自体の長さを長くする

						// 1
						Fraction length = Fraction.sub(
								nowNote.getRealLength(), units.getLast().getRealLength());
						if (length.compareTo(MIN_LENGTH) >= 0) {
							String str2 = sb.toString();
							while (true) {
								if (length.compareTo(MAX_LENGTH) <= 0) {
									str2 += "z"
											+ this.shortLength(Fraction.div(
													length, this.unitNoteLength));
									break;
								} else {
									str2 += "z";
									if (Fraction.sub(length, MAX_LENGTH).compareTo(MIN_LENGTH) < 0) {
										str2 += this.shortLength(Fraction.div(
												Fraction.sub(length, MIN_LENGTH),
												this.unitNoteLength));
										length = MIN_LENGTH.clone();
									} else {
										this.shortLength(Fraction.div(
												MAX_LENGTH, this.unitNoteLength));
										length.sub(MAX_LENGTH);
									}
								}
							}
							if (str == null || str.length() > str2.length()) {
								str = str2;
							}
						}

						// 2
						String str2 = sb.toString();
						length = nowNote.getRealLength().clone();
						boolean isFirst = true;
						while (true) {
							if (length.compareTo(MAX_LENGTH) <= 0) {
								str2 += this.shortLength(
										Fraction.div(length, this.unitNoteLength), isFirst);
								break;
							} else {
								if (Fraction.sub(length, MAX_LENGTH).compareTo(MIN_LENGTH) < 0) {
									str2 += this.shortLength(
											Fraction.div(
													Fraction.sub(length, MIN_LENGTH),
													this.unitNoteLength), isFirst);
									length = MIN_LENGTH.clone();
								} else {
									str2 += this.shortLength(
											Fraction.div(MAX_LENGTH, this.unitNoteLength), isFirst);
									length.sub(MAX_LENGTH);
								}
								str2 += "z";
							}
						}
						if (str == null || str.length() > str2.length()) {
							str = str2;
						}
					} else {
						// 和音最後の音の長さが、和音自体の長さより長い場合、和音自体の長さを設定する
						Fraction length = Fraction.div(nowNote.getRealLength(), this.unitNoteLength);
						String str2 = sb.toString() + this.shortLength(length, true);
						if (str == null || str.length() > str2.length()) {
							str = str2;
						}
					}

					units.addLast(units.removeFirst());
				}

				this.print(str);
			}

			this.barPos.add(nowNote.getRealLength());
		}
	}

	/**
	 * 新しいL:を設定する
	 *
	 * @param newL
	 *            古いL:
	 * @param oldL
	 *            古いL:
	 * @return MOE-ABCの制限に沿った新しいL:
	 * @throws IOException
	 */
	private Fraction newUnitNoteLength(Fraction newL, Fraction oldL) throws IOException {
		Fraction unitNoteLength = newL.clone();
		// L: は1,1/2,1/4,1/8,1/16のみ許す
		if (unitNoteLength.getNumerator() != 1) {
			unitNoteLength.setDenominator(unitNoteLength.getDenominator()
					/ unitNoteLength.getNumerator());
		}
		switch ((int) unitNoteLength.getDenominator()) {
		case 1:
		case 2:
		case 4:
		case 8:
		case 16:
			break;
		default:
			unitNoteLength = new Fraction(1, 4);
			break;
		}

		if (oldL == null || !unitNoteLength.equals(oldL)) {
			if (unitNoteLength.getDenominator() == 1) {
				this.print("<L:1>");
			} else {
				this.print("<L:/" + unitNoteLength.getDenominator() + ">");
			}
		}

		return unitNoteLength;
	}

	/**
	 * 新しいQ:を設定する
	 *
	 * @param newQ
	 *            新しいQ:
	 * @param oldQ
	 *            古いQ:
	 * @return MOE-ABCに沿った新しいQ:
	 * @throws IOException
	 */
	public Tempo newTempo(Tempo newQ, Tempo oldQ) throws IOException {
		Tempo tempo = newQ.clone();
		tempo.convertBase(TEMPO_BASE);

		if (oldQ == null
				|| (tempo.getCount() != oldQ.getCount() || !tempo.getBase().equals(oldQ.getBase()))) {
			this.print("<Q:" + (tempo.getCount() == DEFAULT_TEMPO ? "" : tempo.getCount()) + ">");
		}

		return tempo;
	}

	/**
	 * 最短の長さ表現を取得する
	 *
	 * @param length
	 *            取得する長さ
	 * @return 最短の長さ表現
	 */
	private String shortLength(Fraction length) {
		return this.shortLength(length, false);
	}

	/**
	 * 最短の長さ表現を取得する
	 *
	 * @param length
	 *            取得する長さ
	 * @param is1string
	 *            長さ=1となっても"1"を返すかどうか
	 * @return 最短の長さ表現
	 */
	private String shortLength(Fraction length, boolean is1string) {
		if (length == null) {
			return null;
		}
		if (length.isOne()) {
			return is1string ? "1" : "";
		}
		if (length.isInt()) {
			return "" + length.getNumerator();
		}

		StringBuffer sb = new StringBuffer();

		if (length.getNumerator() != 1) {
			sb.append(length.getNumerator());
		}
		sb.append("/");
		if (length.getDenominator() != 2) {
			sb.append(length.getDenominator());
		}
		return sb.toString();
	}

	/**
	 * NoteUnitの文字列表現を取得する
	 *
	 * @param unit
	 *            対象のNoteUnit
	 * @return 文字列表現
	 */
	private String noteUnit(NoteUnit unit) {
		StringBuffer sb = new StringBuffer();

		EAccidental acci = unit.getAcci();
		String note = unit.getNote();
		int octave = unit.getOctave();
		Fraction length = Fraction.div(unit.getRealLength(), this.unitNoteLength);

		// 臨時記号
		if (acci != null) {
			if (this.opt.isKey2None()) {
				if (acci == EAccidental.NATURAL) {
					acci = null;
				}
			} else {
				if (acci == this.key.getAccidentalBase(note)) {
					acci = null;
				}
			}
		}
		if (acci != null) {
			switch (acci) {
			case DOUBLE_SHARP:
				sb.append("^^");
				break;
			case SHARP:
				sb.append("^");
				break;
			case NATURAL:
				sb.append("=");
				break;
			case FLAT:
				sb.append("_");
				break;
			case DOUBLE_FLAT:
				sb.append("__");
				break;
			default:
				break;
			}
		}

		// 音階
		sb.append(note);

		// オクターブ
		if (octave > 0) {
			sb.append(StringUtil.repeat("'", octave));
		} else if (octave < 0) {
			sb.append(StringUtil.repeat(",", -octave));
		}

		// 長さ
		sb.append(this.shortLength(length));

		// アクセント
		for (EDecoration deco : unit.getDecorations()) {
			if (deco == EDecoration.ACCENT) {
				sb.append(StringUtil.repeat("~", this.opt.getAccentCount()));
			}
		}
		if (this.barPos.equals(Fraction.ZERO)) {
			sb.append(StringUtil.repeat("~", this.opt.getBarAccentCount()));
		}

		// タイ
		if (unit.isWithTie() && "xz".indexOf(note) < 0) {
			sb.append("-");
		}

		return sb.toString();
	}

	/**
	 * タイを繋げる
	 *
	 * @param score
	 *            対象のScore
	 */
	public void combineTie(Score score) {
		for (Tune tune : score.getTunes().values()) {
			for (Track track : tune.getTracks().values()) {
				this.combineTieTrack(track);
			}
		}
	}

	/**
	 * タイを繋げる
	 *
	 * @param track
	 *            対象のTrack
	 */
	public void combineTieTrack(Track track) {
		for (Iterator<Entry<Fraction, Note>> iter = track.getMelody().descendingMap().entrySet().iterator(); iter.hasNext();) {
			Entry<Fraction, Note> entry = iter.next();
			Fraction pos = entry.getKey();
			Note note = entry.getValue();

			for (Iterator<NoteUnit> iterNoteUnit = note.getNotes().iterator(); iterNoteUnit.hasNext();) {
				NoteUnit unit = iterNoteUnit.next();

				if (unit.isWithTie()) {
					// タイがあれば
					Note nextNote = track.getMelody().get(Fraction.add(pos, unit.getRealLength()));
					if (nextNote != null) {
						for (Iterator<NoteUnit> it = nextNote.getNotes().iterator(); it.hasNext();) {
							NoteUnit u = it.next();
							if (unit.compareTo(u) == 0) {
								unit.setRealLength(unit.getRealLength().add(u.getRealLength()));
								unit.setWithTie(false);
								it.remove();
								break;
							}
						}
					}
				}
			}
		}
	}

	/**
	 * タイを付与する。音の途中でテンポが変わった時など、タイを付与して適切にテンポを変えられるようにする。
	 *
	 * @param score
	 *            対象のScore
	 */
	public void addTie(Score score) {
		for (Tune tune : score.getTunes().values()) {
			for (Track track : tune.getTracks().values()) {
				this.addTieTrack(track);
			}
		}
	}

	/**
	 * タイを付与する
	 *
	 * @param track
	 *            対象のTrack
	 */
	public void addTieTrack(Track track) {
		for (Iterator<Entry<Fraction, Note>> iter = track.getMelody().entrySet().iterator(); iter.hasNext();) {
			Entry<Fraction, Note> entry = iter.next();
			Fraction pos = entry.getKey();
			Note note = entry.getValue();

			for (Iterator<NoteUnit> iterNoteUnit = note.getNotes().iterator(); iterNoteUnit.hasNext();) {
				NoteUnit unit = iterNoteUnit.next();
				if (unit.isWithTie()) {
					// 既に処理済み？
					continue;
				}

				List<Entry<Fraction, Note>> entryList = new ArrayList<Entry<Fraction, Note>>(
						track.getMelody().subMap(
								pos, false, Fraction.add(pos, unit.getRealLength()), false).entrySet());
				for (int idx = 0; idx < entryList.size(); ++idx) {
					Note n = entryList.get(idx).getValue();

					// TODO タイで分割するフィールドの種類？
					if (n.getInformation("L") != null) {
						Fraction length = unit.getRealLength().clone();

						Note prevNote = null;
						Fraction prevPos = pos;
						for (int i = 0; i <= idx; ++i) {
							Fraction nowPos = entryList.get(i).getKey();
							Note nowNote = entryList.get(i).getValue();
							Fraction diff = Fraction.sub(nowPos, prevPos);

							if (prevNote == null) {
								unit.setWithTie(true);
								unit.setRealLength(diff);
							} else {
								NoteUnit u = unit.clone();
								u.setWithTie(true);
								u.setRealLength(diff);
								prevNote.addNote(u);
							}

							length.sub(diff);

							prevPos = nowPos;
							prevNote = nowNote;
						}

						if (prevNote != null) {
							NoteUnit u = unit.clone();
							u.setWithTie(false);
							u.setRealLength(length);
							prevNote.addNote(u);
						}

						break;
					}
				}
			}
		}
	}

	/**
	 * 全てのTrackに対しVolumeの最適化({@link Translator#optimizeVolumeTrack(Track)})を行う。
	 *
	 * @param score
	 *            最適化を行うScore
	 */
	public void optimizeVolume(Score score) {
		for (Tune tune : score.getTunes().values()) {
			for (Track track : tune.getTracks().values()) {
				this.optimizeVolumeTrack(track);
			}
		}
	}

	/**
	 * TrackのVolumeの最適化を行う。
	 *
	 * @param track
	 *            対象Track
	 */
	public void optimizeVolumeTrack(Track track) {
		Fraction maxPos = Fraction.ZERO;
		int volumeNow = DEFAULT_VOLUME; // 現在のVolume値
		int volumeSet = DEFAULT_VOLUME; // セットされているVolume値

		for (Iterator<Entry<Fraction, Note>> iter = track.getMelody().entrySet().iterator(); iter.hasNext();) {
			Entry<Fraction, Note> entry = iter.next();
			Fraction pos = entry.getKey();
			Note note = entry.getValue();

			if (note.getVolume() != null) {
				// Volume設定がある場合
				volumeNow = note.getVolume();
				if (note.getNotes().size() == 0 && maxPos.compareTo(pos) < 0) {
					note.setVolume(null);
				} else {
					volumeSet = note.getVolume();
				}
			} else {
				// Volume設定がない場合
				if (note.getNotes().size() > 0) {
					if (volumeNow != volumeSet) {
						note.setVolume(volumeNow);
						volumeSet = volumeNow;
					}
				}
			}

			for (NoteUnit unit : note.getNotes()) {
				Fraction lastPos = Fraction.add(pos, unit.getRealLength());
				if (maxPos.compareTo(lastPos) < 0) {
					maxPos = lastPos;
				}
			}
		}
	}

	/**
	 * 全てのTrackに対しVelocityの最適化({@link Translator#optimizeVelocityTrack(Track)}
	 * )を行う。
	 *
	 * @param score
	 *            最適化を行うScore
	 */
	public void optimizeVelocity(Score score) {
		for (Tune tune : score.getTunes().values()) {
			for (Track track : tune.getTracks().values()) {
				this.optimizeVelocityTrack(track);
			}
		}
	}

	/**
	 * TrackのVelocityの最適化を行う。
	 *
	 * @param track
	 *            対象Track
	 */
	public void optimizeVelocityTrack(Track track) {
		int velocityNow = DEFAULT_VELOCITY; // 現在のVelocity値
		int velocitySet = DEFAULT_VELOCITY; // セットされているVelocity値

		for (Iterator<Entry<Fraction, Note>> iter = track.getMelody().entrySet().iterator(); iter.hasNext();) {
			Entry<Fraction, Note> entry = iter.next();
			Note note = entry.getValue();

			if (note.getVelocity() != null) {
				// Velocity設定がある場合
				velocityNow = note.getVelocity();
				if (note.getNotes().size() == 0) {
					note.setVelocity(null);
				} else {
					velocitySet = note.getVelocity();
				}
			} else {
				// Velocity設定がない場合
				if (note.getNotes().size() > 0) {
					if (velocityNow != velocitySet) {
						note.setVelocity(velocityNow);
						velocitySet = velocityNow;
					}
				}
			}
		}
	}

	/********************************************************************************
	 * TODO MoE-ABCのバグ対応。対応されたら不要になるはず
	 ********************************************************************************/

	/**
	 * MoE-ABC側のバグ(テンポチェンジ)に対応する
	 *
	 * @param score
	 *            対応するScore
	 */
	public void fixTempoChange(Score score) {
		for (Tune tune : score.getTunes().values()) {
			for (Track track : tune.getTracks().values()) {
				this.fixTempoChangeTrack(track);
			}
		}
	}

	/**
	 * MoE-ABC側のバグ(テンポチェンジ)に対応する
	 *
	 * @param track
	 *            対応するTrack
	 */
	public void fixTempoChangeTrack(Track track) {
		track.clearTemporary();
		LinkedList<AInformationField> field;

		// テンポ(Q:)情報
		int tempo = DEFAULT_TEMPO;
		field = track.getInformation("Q");
		if (field == null || field.size() == 0) {
		} else {
			Tempo newTempo = ((InformationFieldTempo) field.getLast()).getTempo().clone();
			newTempo.convertBase(TEMPO_BASE);
			tempo = newTempo.getCount();
		}

		LinkedList<Entry<Fraction, Note>> list = new LinkedList<Entry<Fraction, Note>>(
				track.getMelody().entrySet());
		for (int idx = 0; idx < list.size(); ++idx) {
			Entry<Fraction, Note> nowEntry = list.get(idx);

			Fraction nowPos = nowEntry.getKey();
			Note nowNote = nowEntry.getValue();

			field = nowNote.getInformation("Q");
			if (field != null && field.size() > 0) {
				Tempo newTempo = ((InformationFieldTempo) field.getLast()).getTempo().clone();
				newTempo.convertBase(TEMPO_BASE);
				tempo = newTempo.getCount();
			}

			for (NoteUnit nowUnit : nowNote.getNotes()) {
				if (nowUnit.getRealLength().compareTo(nowNote.getRealLength()) > 0) {
					boolean isFound = false;
					Fraction length = Fraction.ZERO.clone();
					double dLength = 0.0;
					Fraction lastPos = Fraction.add(nowPos, nowUnit.getRealLength());
					for (Entry<Fraction, Note> entry : track.getMelody().subMap(
							nowPos, false, lastPos, false).descendingMap().entrySet()) {
						// 後ろから調べていくよ
						Fraction pos = entry.getKey();
						Note note = entry.getValue();

						field = note.getInformation("Q");
						if (field != null && field.size() > 0) {
							// テンポチェンジ発見
							isFound = true;
							Tempo newTempo = ((InformationFieldTempo) field.getLast()).getTempo().clone();
							newTempo.convertBase(TEMPO_BASE);

							length.add(Fraction.sub(lastPos, pos).mul(
									new Fraction(tempo, newTempo.getCount())));
							dLength += Fraction.sub(lastPos, pos).getDouble() * tempo
									/ newTempo.getCount();
							lastPos = pos;
						}
					}
					if (isFound) {
						// 最後はテンポ変化が無い部分
						length.add(Fraction.sub(lastPos, nowPos));
						dLength += Fraction.sub(lastPos, nowPos).getDouble();

						// TODO 分数の桁数オーバーフローへの暫定対応
						{
							int scale = 960;
							length = new Fraction((int) (dLength * scale), scale);
						}

						nowUnit.setRealLength(length);
					}
				}
			}
		}
	}
}
