package abc.util;

import util.MathUtil;

/**
 * 分数を表すクラス。負数を表す時は分子を負数とし、分母は必ず正数とする。
 *
 * @author CrapeMyrtle
 */
public class Fraction implements Comparable<Fraction>, Cloneable {

	private long	numerator;
	private long	denominator;

	public static final Fraction ZERO = new Fraction(0);

	/**
	 * コンストラクタ
	 */
	public Fraction() {
		this.numerator = 0;
		this.denominator = 1;
	}

	/**
	 * コンストラクタ<br />
	 * num/1で初期化する
	 *
	 * @param num
	 *            整数
	 */
	public Fraction(long num) {
		this.numerator = num;
		this.denominator = 1;
	}

	/**
	 * コンストラクタ<br />
	 * num/denで初期化する (負数の場合はnumを負数とする)
	 *
	 * @param num
	 *            分子
	 * @param den
	 *            分母
	 */
	public Fraction(long num, long den) {
		if (den < 0) {
			this.numerator = -num;
			this.denominator = -den;
		} else {
			this.numerator = num;
			this.denominator = den;
		}
		this.reduction();
	}

	/**
	 * 整数値をセットする
	 *
	 * @param num
	 *            セットする整数値
	 */
	public void setInt(long num) {
		this.numerator = num;
		this.denominator = 1;
	}

	/**
	 * 整数値であるかどうかを返す
	 *
	 * @return 整数値(=分母が1)ならtrue
	 */
	public boolean isInt() {
		return this.denominator == 1;
	}

	/**
	 * 1であるかどうかを返す
	 *
	 * @return 1(=1/1)ならtrue
	 */
	public boolean isOne() {
		return this.denominator == 1 && this.numerator == 1;
	}

	/**
	 * 分数の表す小数を返す
	 *
	 * @return 分数の表す小数
	 */
	public double getDouble() {
		return (double) this.numerator / (double) this.denominator;
	}

	/**
	 * 分数を表す文字列を返す。整数だった場合は分母を省略する。
	 *
	 * @return 分数を表す文字列
	 */
	public String toShortString() {
		return "" + this.numerator
				+ (this.denominator > 1 ? "/" + this.denominator : "");
	}

	/*
	 * (非 Javadoc)
	 *
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		return "" + this.numerator + "/" + this.denominator;
	}

	/**
	 * 分子を返す
	 *
	 * @return 分子
	 */
	public long getNumerator() {
		return numerator;
	}

	/**
	 * 分子を設定する
	 *
	 * @param numerator
	 *            設定する分子
	 */
	public void setNumerator(long numerator) {
		this.numerator = numerator;
	}

	/**
	 * 分母を返す
	 *
	 * @return 分母
	 */
	public long getDenominator() {
		return denominator;
	}

	/**
	 * 分母を設定する
	 *
	 * @param denominator
	 *            設定する分母
	 */
	public void setDenominator(long denominator) {
		this.denominator = denominator;
	}

	/**
	 * マイナス演算を行う(-1をかける)。オブジェクト自体に変更がかかり、変更結果を返り値として返す。
	 *
	 * @return マイナス演算結果の分数
	 */
	public Fraction minus() {
		this.numerator = -this.numerator;
		return this;
	}

	/**
	 * 逆数にする。オブジェクト自体に変更がかかり、変更結果を返り値として返す。
	 *
	 * @return 逆数演算結果の分数
	 */
	public Fraction inverse() {
		int sign = (this.numerator * this.denominator >= 0 ? 1 : -1);
		long temp = Math.abs(this.numerator);
		this.numerator = sign * Math.abs(this.denominator);
		this.denominator = temp;
		return this;
	}

	/**
	 * 分数を乗算する。乗算される側のオブジェクトを変更し、演算結果を返す。
	 *
	 * @param fra
	 *            乗算する分数
	 * @return 乗算結果のぶん数
	 */
	public Fraction mul(Fraction fra) {
		this.numerator *= fra.numerator;
		this.denominator *= fra.denominator;

		this.reduction();
		return this;
	}

	/**
	 * 分数を除算する。除算される側のオブジェクトを変更し、演算結果を返す。
	 *
	 * @param fra
	 *            除算する分数
	 * @return 除算結果のぶん数
	 */
	public Fraction div(Fraction fra) {
		this.numerator *= fra.denominator;
		this.denominator *= fra.numerator;

		this.reduction();
		return this;
	}

	/**
	 * 分数を加算する。加算される側のオブジェクトを変更し、演算結果を返す。
	 *
	 * @param fra
	 *            加算する分数
	 * @return 加算結果のぶん数
	 */
	public Fraction add(Fraction fra) {
		this.numerator = this.numerator * fra.denominator
				+ fra.numerator * this.denominator;
		this.denominator = this.denominator * fra.denominator;
		this.reduction();
		return this;
	}

	/**
	 * 分数を減算する。減算される側のオブジェクトを変更し、演算結果を返す。
	 *
	 * @param fra
	 *            減算する分数
	 * @return 減算結果のぶん数
	 */
	public Fraction sub(Fraction fra) {
		this.numerator = this.numerator * fra.denominator
				- fra.numerator * this.denominator;
		this.denominator = this.denominator * fra.denominator;
		this.reduction();
		return this;
	}

	/**
	 * マイナス演算を行う(-1をかける)。引数の分数を変更しない。
	 *
	 * @param fra
	 *            演算を行う分数
	 * @return マイナス演算結果の分数
	 */
	public static Fraction minus(Fraction fra) {
		return fra.clone().minus();
	}

	/**
	 * 逆数にする。引数の分数を変更しない。
	 *
	 * @param fra
	 *            逆数にする分数
	 * @return 逆数にした分数
	 */
	public static Fraction inverse(Fraction fra) {
		return fra.clone().inverse();
	}

	/**
	 * 分数どうしの乗算を行う。引数の分数を変更しない。
	 *
	 * @param fra1
	 *            乗算する分数
	 * @param fra2
	 *            乗算する分数
	 * @return 乗算結果のぶん数
	 */
	public static Fraction mul(Fraction fra1, Fraction fra2) {
		return fra1.clone().mul(fra2);
	}

	/**
	 * 分数どうしの除算を行う。引数の分数を変更しない。
	 *
	 * @param fra1
	 *            除算を行う分数
	 * @param fra2
	 *            除算を行う分数
	 * @return 除算結果の分数
	 */
	public static Fraction div(Fraction fra1, Fraction fra2) {
		return fra1.clone().div(fra2);
	}

	/**
	 * 分数どうしの加算を行う。引数の分数を変更しない。
	 *
	 * @param fra1
	 *            加算を行う分数
	 * @param fra2
	 *            加算を行う分数
	 * @return 加算結果の分数
	 */
	public static Fraction add(Fraction fra1, Fraction fra2) {
		return fra1.clone().add(fra2);
	}

	/**
	 * 分数どうしの減算を行う。引数の分数を変更しない。
	 *
	 * @param fra1
	 *            減算を行う分数
	 * @param fra2
	 *            減算を行う分数
	 * @return 減算結果の分数
	 */
	public static Fraction sub(Fraction fra1, Fraction fra2) {
		return fra1.clone().sub(fra2);
	}

	/**
	 * 通分を行う
	 *
	 * @return 通分結果の分数
	 */
	private void reduction() {
		long gcd = Math.abs(MathUtil.gcd(this.numerator, this.denominator));
		this.numerator /= gcd;
		this.denominator /= gcd;
	}

	/**
	 * 文字列から分数を解析する
	 *
	 * @param s
	 *            解析する文字列
	 * @return 解析結果の分数
	 */
	public static Fraction parseFraction(String s) {
		try {
		String[] a = s.split("/", 2);

		Fraction fraction = null;
		switch (a.length) {
		case 2:
			fraction = new Fraction(Integer.parseInt(a[0]),
					Integer.parseInt(a[1]));
			break;
		case 1:
			fraction = new Fraction(Integer.parseInt(a[0]));
			break;
		case 0:
			fraction = new Fraction();
			break;
		}

		return fraction;
		} catch (NumberFormatException e) {
			throw new NumberFormatException("For input string : \"" + s + "\"");
		}
	}

	/*
	 * (非 Javadoc)
	 *
	 * @see java.lang.Object#clone()
	 */
	@Override
	public Fraction clone() {
		return new Fraction(this.numerator, this.denominator);
	}

	/**
	 * 分数同士の値を比較する
	 * @param fra 比較する分数
	 * @return true: 分数同士の値が等しい
	 */
	public boolean equals(Fraction fra) {
		return this.numerator == fra.numerator && this.denominator == fra.denominator;
	}

	/* (非 Javadoc)
	 * @see java.lang.Comparable#compareTo(java.lang.Object)
	 */
	@Override
	public int compareTo(Fraction o) {
		long diff = Fraction.sub(this, o).getNumerator();
		return diff == 0 ? 0 : (diff > 0 ? 1 : -1);
	}

}
